(*
 * Matthew Stone
 * May 1993, Modified March 1994, September 1996
 * Revised May 1998 for SML-110.
 *
 * Lexical Analyser, for a prolog like language.
 * or, as I wrote it the first time around...
 *
 * Breaking up stuff into words 
 *)

signature LEX =
    sig
	val segment : string -> string list
	val atof : string -> real
    end;
    
structure Lex1: LEX =
struct

    exception error of string

    val comment = Char.contains "%"
    val newline = Char.contains "\n"
    val separator = Char.contains " \t\n"
    val tokenable = Char.contains "><#@*?:=~-"
    val magic = Char.contains ".,;!|()[]{}"

    fun segment s =
	let fun getword(w,[]) = [ implode (rev w) ]
	      | getword(w,c::rest) =
		if comment c then
		    implode(rev w) :: zap rest
		else
		    if separator c then
			implode(rev w) :: skipspace rest
		    else 
			if magic c then
			    implode(rev w) :: Char.toString c :: skipspace rest
			else
			    if tokenable c then
				implode(rev w) :: gettoken([c], rest)
			    else
				getword(c::w, rest)

	    and gettoken(w,[]) = [ implode (rev w) ]
		  | gettoken(w,c::rest) =
		    if comment c then
			implode(rev w) :: zap rest
		    else
			if separator c then
			    implode(rev w) :: skipspace rest
			else
			    if magic c then
				implode(rev w) :: Char.toString c :: skipspace rest
			    else
				if tokenable c then
				    gettoken(c::w, rest)
				else
				    implode(rev w) :: getword([c], rest)

	    and skipspace [] = []
	      | skipspace(c::rest) =
		if comment c then
		    zap rest
		else 
		    if separator c then
			skipspace rest
		    else 
			if magic c then
			    Char.toString c :: skipspace rest
			else 
			    if tokenable c then
				gettoken([c], rest)
			    else
				getword([c], rest)

	    and zap [] = []
	      | zap (c::rest) =
		if newline c then
		    skipspace rest
		else
		    zap rest
	in
	    skipspace (explode s)
	end

    fun atof str =
	let fun accum [] num denom inc = num / denom
	      | accum (c::s) num denom inc =
		if c = #"." then
		    accum s num denom (fn x => 10.0*x)
		else if c >= #"0" andalso c <= #"9" then
		    accum s (10.0*num + (real (ord c)) - 
			     (real (ord #"0")))
		          (inc denom) inc
		else
		    raise (error "atof")
	in
	    accum (explode str) 0.0 1.0 (fn x:real => x)
	end
	

end  (* Lex1 *)

signature IOA =
    sig
	exception error of string

	type reader
	type writer
	    
	val wrapInstream : TextIO.instream -> reader
	val wrapOutstream : TextIO.outstream -> writer
	val wrapSocket: 
	    (INetSock.inet, Socket.active Socket.stream) Socket.sock
	    -> reader * writer

	val getline: reader -> string 
	val getclause: reader -> string
	val getall: reader -> string 

	val put: writer * string -> unit
    end

structure SIOA : IOA =
    struct
	exception error of string

	(* these types are too specific, but 
	 * expanding for polymorphism would be a pain *)
	datatype reader = 
	    IS of TextIO.instream
	  | SKR of (INetSock.inet, Socket.active Socket.stream) Socket.sock

	datatype writer =
	    OS of TextIO.outstream
	  | SKW of (INetSock.inet, Socket.active Socket.stream) Socket.sock

	    
	fun wrapInstream s = IS(s)
	fun wrapOutstream s = OS(s)

	fun wrapSocket fd = (SKR(fd),  SKW(fd))

	fun next (IS instream) = 
	    if TextIO.endOfStream instream then "" 
	    else TextIO.inputN(instream,1)

	  | next (SKR (fd)) =
	    (let val v = Socket.recvVec(fd, 1) 
	     in if (0 = Word8Vector.length v) then
		 ""
		else
		    Byte.unpackStringVec (v,0,SOME 1)
	     end handle OS.SysErr(s,_) => raise error s)

	fun put (OS outstream, s) = TextIO.output(outstream,s)
	  | put (SKW fd, s) = 
	    let val vec = Byte.stringToBytes s
		val len = Word8Vector.length vec
		fun send i = Socket.sendVec (fd, {buf=vec, i=i, sz=NONE})
		fun put i = 
		    if (i < len) then 
			put(i + send i) 
		    else ()
	    in
		put 0
	    end handle OS.SysErr(s,_) => raise error s

	fun comment "%" = true
	  | comment _ = false

	fun newline ""   = true
	  | newline "\n" = true
	  | newline _    = false

	fun getto term reader =
	    let fun extend ("") = ""
		  | extend ("\\") =
		    let
			val nextc = next reader
		    in
			if nextc = "" then "\\"
			else (nextc ^ getstring ())
		    end
		  | extend c =
		    if term c then c 
		    else 
			if comment c then
			    (getto newline reader;
			     getstring ())
			else
			    c ^ (getstring())

		and getstring () = extend(next reader)
	    in
		getstring ()
	    end;

	fun dot "."  = true
	  | dot _   = false
	    
	fun theend "" = true
	  | theend _ = false

	val getline = getto newline

	val getclause = getto dot

	val getall = getto theend

    end
