(* 
 * Matthew Stone
 * Dialup 0.1 - 3/96
 *
 * Specification and implementation of a symbol table.
 * Symbols have fast equality and comparison testing,
 * and come with a hash function (also fast).
 * 
 * calling makeSymbol on equal strings twice returns
 * the the same symbol.
 *
 * revised  5/98 for sml-110
 *)

signature SYMBL =
    sig
	type symbol

	val sameSymbol : symbol * symbol -> bool
	val cmpSymbol : symbol * symbol -> order
	val makeSymbol : string -> symbol
	val genSymbol : string -> symbol
	val hashSymbol : symbol -> word
	val nameSymbol : symbol -> string

	structure HashSymbol : HASH_KEY 
	sharing type HashSymbol.hash_key = symbol
	    
	structure OrdSymbol : ORD_KEY
	sharing type OrdSymbol.ord_key = symbol

    end (* SYMBL *)

functor Symbol (structure HashTable : HASH_TABLE) : SYMBL =
    struct
	
	type symbol = (string * word) ref
	val sameSymbol = op = 
	fun cmpSymbol (ref(s1, h1) : symbol,
		       ref(s2, h2) : symbol) = 
	    if h1 < h2 then 
		LESS
	    else 
		if h1 > h2 then 
		    GREATER
		else String.compare(s1, s2)
		    
	fun hashSymbol (ref(_, hash)) = hash
	fun nameSymbol (ref(name, _)) = name
	    
	structure HashSymbol : HASH_KEY =
	    struct
		type hash_key = symbol
		val hashVal = hashSymbol
		val sameKey = sameSymbol
	    end
	
	structure OrdSymbol : ORD_KEY =
	    struct
		type ord_key = symbol
		val compare = cmpSymbol
	    end
	
	local
	    open HashTable
		
	    exception Miss
	    val myTable = mkTable (fn x => x, op =) (1000, Miss) : 
		(word, symbol) hash_table
		
	in
	    fun makeSymbol str =
		let 
		    val hash = HashString.hashString(str)
		in
		    lookup myTable hash
		    handle Miss => 
			let
			    val new = ref(str, hash)
			in
			    insert myTable (hash, new); new
			end
		end
	end
    
	local 
	    val sofar = ref 0
	in
	    fun genSymbol (prefix) =
		(sofar := !sofar + 1;
		 makeSymbol (prefix ^ (Int.toString (!sofar))))
	end
    
    end;
