(*
 *
 *)

functor TermCall (structure Term : TERM) : CALL =
    struct
	structure Term = Term
	structure Patheq = Term.Patheq

	datatype presupp =
	    Presupposed
	  | Asserted

	datatype pb =
	    V of presupp
	  | E of int

	type terminit = Term.init Term.IntMap.intmap

	datatype init = 
	    N of {term: terminit,
		  pre: int Term.IntMap.intmap}

	datatype call =
	    C of {pp : pb,
		  ct : Term.term,
		  fvs : Term.IntSet.intset,
		  init : init}

	datatype info =
	    I of {sys : Term.system,
		  pbind : pb Term.IntMap.intmap,
		  p : Term.Patheq.problem}

	fun mkInit () = N{term = Term.IntMap.empty (),
			  pre = Term.IntMap.empty() }

	local
	    val ct = ref 0
	in
	    fun newPPVar () = (ct := !ct + 1; !ct)
	end

	local
	    fun mkCt (pp, ct) =
		C{pp = pp,
		  ct = ct,
		  fvs = Term.freeVarsTerm(ct, Term.IntSet.empty),
		  init = mkInit ()}

	    fun readct (pp, str) =
		case Term.readTerm str of
		    (SOME ct, fin) =>
			(SOME (mkCt(pp,ct)),
			 fin)
		  | _ => (NONE, str)
	in
	    fun createCall (ct) =
		mkCt(E (newPPVar ()), ct)

	    fun readCall v ("p" :: ":" :: str) =
		readct (V Presupposed, str)

	      | readCall v ("a" :: ":" :: str) =
		readct (V Asserted, str)

	      | readCall v ("u" :: ":" :: str) =
		readct (E (newPPVar ()), str)

	      | readCall v str = 
		readct (E v, str)
	end 

	fun lookupPVar (E v, p) =
	    (case Term.IntMap.peek(p, v) of
		 NONE => E v
	       | SOME v' => E v')
	  | lookupPVar (v, _) = v

	fun printInit (N{term,...}) =
	    Term.IntMap.app 
	    (fn (i,Term.VI(x,_)) => (print (Int.toString i); print " is ";
				     print (Int.toString x); print "\n")
	  | (i,_) => (print (Int.toString i); print " is something weird\n"))
	    term

	fun checkGround (I{sys, pbind, p}) (C{pp, ct, init=N{term,pre},...}) =
	    let
		fun cg (V _) = true
		  | cg (E x) =
		    (case Term.IntMap.peek(pbind,x) of
			 NONE => false
		       | SOME n => cg n)
	    in
		cg (lookupPVar(pp, pre)) andalso
		Term.checkGround (ct, term, sys)
	    end

	fun checkKeyGround (I{sys,...}) (C{ct, init = N{term,...}, ...}) =
	    Term.checkGround(ct, term,sys)

	local
	    fun writep (V Presupposed) = "p:"
	      | writep (V Asserted) = "a:"
	      | writep (E _) = "u:"

	    fun pwritep (V Presupposed, _) = "p:"
	      | pwritep (V Asserted, _) = "a:"
	      | pwritep (E v, pb) = 
		case Term.IntMap.peek(pb,v) of 
		    NONE => "u:"
		  | SOME more => pwritep(more, pb)
	in
	    fun writeHead (C{ct,...}) =
		Term.writeHead ct

	    fun writeCall (C{pp, ct, ...}) = 
		(writep pp ^ Term.writeTerm ct)

	    val printCall = print o writeCall

	    fun writeprettyCall (I{sys, pbind, p})  (C{pp, ct, init=N{pre,term},...}) =
		(pwritep (lookupPVar(pp, pre), pbind) ^
		 Term.writeprettyTerm sys (SOME term, ct))

	    fun prettyprintCall i = print o (writeprettyCall i)
	end

	fun overlap (c as C{ct,...}, c2 as C{ct=ct',...}) =
	    let
		val result = Term.sameHead(ct, ct')
	    in
		result
	    end

	fun duplicates(c1 as C{pp=pp1, ct=ct1, init=N{term=t1,pre=p1},...},
		       i1 as I{sys=s1, pbind=pb1,...},
		       c2 as C{pp=pp2, ct=ct2, init=N{term=t2,pre=p2},...},
		       i2 as I{sys=s2, pbind=pb2,...}) =
	    let
		fun dp (V Presupposed, V Presupposed) = true
		  | dp (V Asserted, V Asserted) = true
		  | dp (E x, y) =
		    (case Term.IntMap.peek(pb1,x) of
			 NONE => true
		       | SOME z => dp(z, y))
		  | dp (x, E y) =
		    (case Term.IntMap.peek(pb2,y) of
			 NONE => true
		       | SOME z => dp(x, z))
		  | dp _ = false

		val result =
		    dp (lookupPVar (pp1, p1), lookupPVar(pp2, p2)) andalso
		    Term.deepMatch(ct1, t1, s1, ct2, t2, s2)
	    in
		result
	    end

	local
	    val collection = Term.Patheq.Modspec.newCollection ()
	    val _ = Term.Patheq.Modspec.compile (collection)
	    val problem = Term.Patheq.newProblem (collection)
	    val root = Term.Patheq.rootProblem (problem)

	    fun extendTinit (i, tinit) =
		case Term.IntMap.peek (tinit, i) of 
		    SOME _ => 
			(tinit)
		  | NONE => 
			let
			    val v = Term.newVar()
			in
			    Term.IntMap.insert(tinit, i, 
					       Term.VI(v, root))
			end

	    fun extendAfter model (i, tinit) =
		case Term.IntMap.peek (tinit, i) of 
		    SOME _ => tinit
		  | NONE => 
			(case Term.IntMap.peek (model, i) of
			     SOME entry =>
				 Term.IntMap.insert(tinit, i, entry)
			   | _ => 
				 Term.IntMap.insert(tinit, i, 
						    Term.VI(Term.newVar(), root)))

	    fun addConstraint (oldInit, newInit) (iv, Term.TI(ov, _, term), (sys, p)) =
		(let
		    val varToUpdate = 
			(case Term.IntMap.peek (oldInit, iv) of
			     NONE => ov
			   | SOME (Term.TI(v,_,_)) => v
			   | SOME (Term.VI(v,_)) => v
			   | SOME (Term.OI(v,_)) => v
			   | _ => ov)
		in
		    (Term.unify
		     (Term.updateVar(varToUpdate, term, newInit, root, 
				     p, sys)))
		end
	    handle x => (print "unification failed\n";
			 Term.printSystem print sys;
			 raise x))

	      | addConstraint (oldInit, newInit) (_, _, pair) = pair

	    fun unifyPP(E x, t, pbind) =
		(case Term.IntMap.peek(pbind, x) of
		     NONE => 
			 (t, Term.IntMap.insert(pbind, x, t))
		   | SOME t' => unifyPP(t', t, pbind))
	      | unifyPP (t, E y, pbind) =
		(case Term.IntMap.peek(pbind, y) of
		     NONE => 
			 (t, Term.IntMap.insert(pbind, y, t))
		   | SOME t' => unifyPP(t, t', pbind))
	      | unifyPP(V Asserted, V Asserted, pbind) = 
		(V Asserted, pbind)
	      | unifyPP(V Presupposed, V Presupposed, pbind) =
		(V Presupposed, pbind)
	      | unifyPP _ = raise Term.Clash

	    fun matchPP(pp1, p1, pp2, p2, pbind) =
		unifyPP(lookupPVar (pp1, p1),
			lookupPVar (pp2, p2),
			pbind)

	    fun copyCall extension (C{pp,ct,fvs,...}, N{term=tinit, pre=pinit}) =
		let
		    val newtinit = Term.IntSet.fold extension fvs tinit
		    val newpinit = 
			(case pp of 
			     E v => 
				 (case Term.IntMap.peek (pinit, v) of
				      SOME _ => pinit
				    | NONE => Term.IntMap.insert(pinit, v, 
								 newPPVar ()))
			   | _ => pinit)

		    val ni = N{term=newtinit, pre=newpinit}
		    val _ = ()			
		in
		    (C{pp=pp,
		       ct=ct,
		       fvs=fvs,
		       init = ni},
		     ni)
		end
	in
	    fun cacheTermApp (f:Term.term -> '1b) =
		let
		    val results = ref nil : (Term.term * '1b) list ref
		    val sys = Term.newSystem problem
		    val trans = Term.IntMap.empty() : Term.init Term.IntMap.intmap;
			
		    fun try t =
			case (List.find
			      (fn (x,r) => Term.deepMatch(x, trans, sys,
							  t, trans, sys))
			      (!results)) of
			    NONE => 
				let
				    val next = Clock.clockapp f "saved event" t
				in
				    results := (t,next) :: (!results);
				    next
				end
			  | SOME (_,r) => 
				r
		in
		    try
		end

	    fun mkInfo () =
		I{sys = Term.newSystem problem,
		  pbind = Term.IntMap.empty (),
		  p = problem}

	    fun printInfo (I{sys, pbind, p}) =
		(Term.printSystem print sys)

	    val newcopyCall = copyCall extendTinit

	    fun copycallAfter (c, i, N{term=model,...}) =
		copyCall (extendAfter model) (c, i)

	    fun match (C{pp=pp1, ct=ct1, init = N{term=i1,pre=p1},...},
		       C{pp=pp2, ct=ct2, init = N{term=i2,pre=p2},...},
		       I{sys, pbind, p}) =
		if Term.quickMatch(ct1, i1, ct2, i2, sys) then
		    let
			val (presup, pbind') = 
			    matchPP(pp1, p1, pp2, p2, pbind) 
			val (s', p') =
			    (Term.unify
			     (Term.updateSystem(ct1, i1, root,
						ct2, i2, root,
						p, sys)))
		    in 
			SOME (presup, I {sys = s', pbind = pbind', p= p'})
		    end
		handle Term.Clash => NONE
		     | Term.Cycle => NONE
		     | Term.Scope => NONE
		else
		    (NONE)
		    
	    fun setWithInit(I{sys, pbind, p}, N{term=oldinit,...}, terminit, pair) =
		let
		    val (sys', p') = (Term.IntMap.fold 
				      (addConstraint (oldinit, terminit))
				      terminit (sys, p))
		    val (pbind') = 
			case pair of
			    SOME (E x, value) =>
				(Term.IntMap.insert(pbind, x, V value))
			  | _ => pbind
		in
		    SOME (I{sys = sys', pbind = pbind', p = p'})
		end
	    handle Term.Clash => NONE
		 | Term.Cycle => NONE
		 | Term.Scope => NONE
	end
    
	local
	    fun makeNewTrans (sys, fvs, oldtrans, world) =
		let
		    fun gv (iv, ov, tr) =
			(case Term.readoutVar (sys, SOME ov) of
			      Term.Bound t => 
				 Term.IntMap.insert(tr, iv, 
						    Term.TI(Term.bestAlias(sys, ov),
							    world, t))

			    | _ => 
				 Term.IntMap.insert(tr, iv, 
						    Term.VI(Term.bestAlias (sys, ov), 
							    world)))

		    fun et (iv, Term.TI(ov, _, _), tr) =
			gv (iv, ov, tr)
		      | et (iv, Term.VI(ov, _), tr) =
			gv (iv, ov, tr)
		      | et (_, _, tr) = tr

		    fun lk (iv, tr) =
			case Term.IntMap.peek (oldtrans, iv) of
			    NONE => et(iv, Term.VI(iv, world), tr)
			  | SOME pair => et (iv, pair, tr)
		in
		    Term.IntSet.fold lk fvs (Term.IntMap.empty())
		end
	in
	    fun getInit (I{sys, pbind, p}, C{pp, ct, fvs, init=N{term,...}}, world) =
		makeNewTrans(sys, fvs, term, world)

	    fun getKeyInit(I{sys, pbind, p}, fvs, world) =
		makeNewTrans(sys, fvs, Term.IntMap.empty(), world)

	    fun readout (I{sys, pbind, p}, i) =
		Term.readoutVar(sys, SOME i)
	end

	fun matchkeys (I{sys,...}) (i,j) =
	    Term.matchVars sys (i,j)
    end
