
functor Constraint (structure Formula: FDB;
		    structure Call : CALL;
		    sharing Call.Term = Formula.Term) : CONSTRAINT =
    struct
	structure Formula = Formula
	structure Call = Call
	    
	open Formula

	datatype relation = 
	    R of {vars: int list,
		  tuples: Term.term list list}

	datatype domains =
	    D of {asgn : int Term.IntMap.intmap,
		  worksizes : int list,
		  cells : Term.term list Term.IntMap.intmap}

	datatype problem =
	    P of {domains: domains,
		  relations: relation list,
		  free: int list,
		  stuff: (Formula.Patheq.pathpt * 
			  (Formula.Term.init Formula.Term.IntMap.intmap  * 
			   Formula.formula -> 
			   unit Formula.proof list)),
		  domainf: Term.term -> Term.term list}

	local
	    val bogusModspec = Patheq.Modspec.newCollection ()

	    val bogusProblem = Patheq.newProblem (bogusModspec)
	    val bogusSystem = Term.newSystem (bogusProblem)

	    val bogusTrans = Term.IntMap.empty() : 
		Term.init Term.IntMap.intmap;

	    val queryable = [ "domain", "(", "X", ",", "Y", ")"]
	    val testvar = Term.getVar "X"
	    val resultvar = Term.getVar "Y"

	    val matchFormula = 
		case Formula.readFormula bogusModspec queryable of
		    (SOME t, nil) => t
		  | _ => raise (Fail "constraint.sml: couldn't read domain\n")

	in 
	    fun newProblem (publicworld, publicset) = 
		let
		    fun domain (testterm) =
			let
			    val transtest = Term.newVar ()
			    val transresult = Term.newVar ()
			    val tinit = 
				Term.IntMap.insert
				(Term.IntMap.insert
				 (Term.IntMap.empty(),
				  testvar, Term.TI(transtest, publicworld, testterm)),
				 resultvar, Term.VI(transresult, publicworld))

			    val info = publicset(tinit,matchFormula)

			    fun extract pf =
				let val i = makeNewTrans (pf, tinit, publicworld)
				in case Term.IntMap.peek (i, resultvar) of
				    SOME (Term.TI(_,_,t)) => SOME t
				  | _ => NONE
				end
			in
			    List.mapPartial extract info
			end
		in
		    P {domains= D{asgn=Term.IntMap.empty(),
				  worksizes=nil,
				  cells=Term.IntMap.empty()},
		       relations = nil,
		       free = nil,
		       stuff = (publicworld, publicset),
		       domainf = Call.cacheTermApp domain}
		end

	    fun contains (nil, _) = false
	      | contains (x :: more, elt) =
		Formula.Term.deepMatch(x, bogusTrans, bogusSystem,
				       elt, bogusTrans, bogusSystem)
		orelse contains(more, elt)

	    fun matches x elt =
		Formula.Term.deepMatch(x, bogusTrans, bogusSystem,
				       elt, bogusTrans, bogusSystem)

	end

	fun printProblem (P{domains = D{cells,...},...}) =
	    let
		fun pc (i:int, elts) =
		    (print "For ";
		     print (Int.toString i);
		     print ": ";
		     app (fn x => (Term.printTerm print x;
				   print " - ")) elts;
		     print "\n")
	    in
		Term.IntMap.app pc cells
	    end

	fun pruneDandR (R {vars, tuples}, (D{asgn, worksizes, cells}, change)) =
	    let
		fun check (nil, nil, guys) = rev guys
		  | check (fv :: morevars, value :: morevals, guys) = 
		    check (morevars, morevals,
			   (List.filter (matches value)
			    (Term.IntMap.find(cells, Term.IntMap.find(asgn, fv)))) ::
			   guys)
		  | check _ = [ nil ]

		fun add (nil, nil, newcells) = newcells
		  | add (fv :: morevars, values :: morevals, newcells) =
		    let
			val cellno = Term.IntMap.find(asgn, fv)
			val cell = 
			    case Term.IntMap.peek(newcells, cellno) of
				NONE => nil
			      | SOME l => l
		    in
			add(morevars, morevals,
			    case List.filter (fn x => not (List.exists (matches x) cell)) values of
				nil => newcells
			      | other => Term.IntMap.insert(newcells, cellno, other @ cell))
		    end
		  | add (_, _, newcells) = newcells

		fun restrict (i, newcell, (oldcells, change)) =
		    if length(newcell) < length(Term.IntMap.find(oldcells, i)) then
			(Term.IntMap.insert(oldcells, i, newcell),
			 true)
		    else
			(oldcells, change)

		fun update (nil, newcells, newtuples, change) =
		    let
			val (newdcells, dchange) = 
			    Term.IntMap.fold restrict newcells (cells, change)
		    in
			(R{vars=vars, tuples=newtuples},
			 (D{asgn=asgn, worksizes = worksizes, cells=newdcells},
			  dchange))
		    end

		  | update (tuple :: moretuples, newcells, newtuples, change) =
		    let
			val passes = check (vars, tuple, nil) 
		    in
			if not (List.exists null passes) then
			    update(moretuples, 
				   add (vars, passes, newcells),
				   tuple :: newtuples,
				   change)
			else
			    update(moretuples,
				   newcells,
				   newtuples,
				   true)
		    end
	    in
		update(tuples, Term.IntMap.empty(), nil, change)
	    end

	fun relax (p as P{domains, relations, free, stuff, domainf}) =
	    let
		fun loop(domain, relations) =
		    let
			val (newrelations, (newdomain, changed)) =
			    U.foldstate pruneDandR (relations, (domain, false))

			val _ =
			    if Debug.debugp "relax" then
				(print "One loop done\n";
				 printProblem(P{domains=newdomain, relations=newrelations, 
						free=nil, stuff= stuff, domainf=domainf});
				 print "\n")
			    else
				()
		    in
			if changed then
			    loop(newdomain, newrelations)
			else
			    (newdomain, newrelations)
		    end

		val _ =
		    if Debug.debugp "relax" then
			(print "Beginning relaxation\n";
			 printProblem(p);
			 print "\n")
		    else
			()

		val (newdomains, newrelations) = loop (domains, relations)
	    in
		P{domains = newdomains,
		  relations = newrelations,
		  free = free,
		  stuff = stuff,
		  domainf = domainf}
	    end

	local
	    fun buildRelation (fvo, fvt, infos, init, publicworld,
			       D{asgn,worksizes,cells}) =
		let 
		    fun buildTuple (pf, nil, _, info, tuple) = 
			[ rev (map (Term.subst info) tuple) ]
		      | buildTuple (pf, v :: fvs, tv :: tfvs, info, tuple) =
			(case Term.IntMap.peek (info, v) of
			    SOME(Term.TI(_,_,term)) => 
				buildTuple(pf,fvs, tfvs, info, term :: tuple)
			  | _ => 
				let 
				    val _ = 
					if Debug.debugp "warn" then
					    (print "Warning: free variable in relation\n")
					else
					    ()

				    val dom = Term.IntMap.find(cells, Term.IntMap.find(asgn, tv))
				in 
				    foldr (fn (x,l) => 
					   (buildTuple 
					    (pf, fvs, tfvs,
					     Term.IntMap.insert(info, 
								bestAlias(pf,tv), 
								Term.TI(tv,publicworld,x)),
					     x :: tuple) @ l))
				    nil dom
				end)
		      | buildTuple _ = nil

		    fun extract (pf,sofar) =
			let val i = makeNewTrans (pf, init, publicworld)
			in buildTuple (pf, fvo, fvt, i, nil) @ sofar
			end

		    val tuples = foldr extract nil infos

		    val _ =
			if Debug.debugp "constraint" then
			    (print "Tuples:\n";
			     app (fn x => (print "< ";
					   app (fn x => (Term.printTerm print x;

							 print " ")) x;
					   print ">\n"))
			     tuples;
			     print "\n")
			else
			    ()
		in
		    R{vars = fvt,
		      tuples = foldr extract nil infos}
		end

	in
	    fun addConstraint 
		(P{domains=D{asgn,worksizes,cells}, relations, free, 
		   stuff=stuff as (publicworld, publicset), domainf}, 
		 formula, gentinit, geninfo, 
		 spectinit, specinfo) =
		let
		    fun dom (specinfo, i) = 
			(case Call.readout(specinfo,i) of
			     Term.Bound testterm => domainf testterm
			   | _ => 
				 (print "(constraint) ACK: specific unbound\n";
				  nil))

		    val mk = Call.matchkeys geninfo

		    fun findmk j (i,_,NONE) =
			if mk(i,j) then SOME i
			else NONE
		      | findmk _ (_,_,x) = x

		    fun uniquify cells (j, l) =
			case Term.IntMap.fold (findmk j) cells NONE of
			    NONE => 
				if (List.exists (fn x => mk(j,x)) l) then 
				    l 
				else (case dom (specinfo, j) of
					  nil => l
					| [x] => l
					| _ => j :: l)
			      | _ => l

		    fun match (j, asgn, worksizes, cells) =
			(case Term.IntMap.fold (findmk j) cells NONE of
			     NONE => 
				 let
				     val start = dom (specinfo, j)
				 in
				     (Term.IntMap.insert(asgn, j, j),
				      (length start) :: worksizes,
				      Term.IntMap.insert(cells, j, start))
				 end
			   | SOME k => 
				 (Term.IntMap.insert(asgn, j, k),
				  worksizes,
				  cells))
			     
		    fun add (i, Term.VI(j, _), 
			     stuff as (fvo, fvt, fvni, (asgn, worksizes, cells))) =  
			if varFreeIn i formula then
			    (i :: fvo, j :: fvt, fvni, 
			     match(j, asgn, worksizes, cells))
			else 
			    (fvo, fvt, j :: fvni, (asgn, worksizes, cells))
		      | add (i, Term.TI(j, _, _),
			     stuff as (fvo, fvt, fvni, (asgn, worksizes, cells))) =
			(case Term.IntMap.fold (findmk j) cells NONE of
				NONE => stuff
			      | SOME k =>
				(i :: fvo, j :: fvt, fvni, 
				 (Term.IntMap.insert(asgn, j, k), worksizes, cells)))
		      | add (i, Term.OI(j, _), 
			     stuff as (fvo, fvt, fvni, (asgn, worksizes, cells))) = 
			if varFreeIn i formula then 
			    (i :: fvo, j :: fvt, fvni, 
			     match(j, asgn, worksizes, cells))
			else
			    (fvo, fvt, j :: fvni, 
			     (asgn, worksizes, cells))
		      | add (_, _, stuff) = stuff

		    val (fvo, fvt, newfree, (newasgn, newworksizes, newcells)) = 
			Term.IntMap.fold add gentinit 
			(nil, nil, free, (asgn, worksizes, cells))

		    val finfree =
			(foldr (uniquify newcells) nil newfree)

		    val _ =
			if Debug.debugp "constraint" then
			    (print "Adding constraint: ";
			     Formula.printFormula print formula;
			     print "\ngentinit is: \n";
			     Term.printTrans gentinit;
			     print "\n fvo = ";
			     app (fn x => (print (Int.toString x); print " ")) fvo;
			     print "\n fvt = ";
			     app (fn x => (print (Int.toString x); print " ")) fvt;
			     print "\n newfree = ";
			     app (fn x => (print (Int.toString x); print " ")) newfree;
			     print "\n free = ";
			     app (fn x => (print (Int.toString x); print " ")) finfree;
			     print "\n")
			else
			    ()

		    val newdomain = D{asgn=newasgn, 
				      worksizes= newworksizes,
				      cells=newcells}
		in
		    relax(P{domains = newdomain,
			    relations = 
			    buildRelation(fvo, fvt, publicset(gentinit, formula),
					  gentinit, publicworld, newdomain) 
			    :: relations,
			    free = finfree,
			    stuff = stuff,
			    domainf = domainf})
		end
	end		

	fun sizes (P{domains=D{cells,...},...}) =
	    Term.IntMap.fold (fn (_,cell,l) => (length cell) :: l) cells nil

	fun worksizes (P{domains=D{worksizes,...},...}) = worksizes

	fun numfree (P{free,...}) = length free

	val _ = 
	    app Debug.register
	    [("constraint", 
	      "Watch the bookkeeping of building up relations...\n", 
	      false),
	     ("relax",
	      "Watch the action of relaxation solver\n",
	      false),
	     ("warn",
	      "Print out notices when possibly bad things happen\n",
	      false)]
    end
