(*
 * feature.sml
 *
 * MS - 6/97 (after 7/94; 12/95)
 *
 * Implementation of feature structures
 *)

functor Features(structure ListSort : LIST_SORT;
		 structure MySymbol : SYMBL;
		 structure OrdSet : ORD_SET;
		 structure Dict : ORD_MAP;
		 sharing type MySymbol.symbol = OrdSet.item = Dict.Key.ord_key)
    : FEATURES =

    struct
	
	structure Symbol = MySymbol
	structure SymbolSet = OrdSet

	open Symbol

	datatype features = 
	    Top
	  | Atom of symbol
	  | Var of symbol
	  | Compound of (symbol * features) list

	type binding = features Dict.map

	val sortElts = 
	    (fn l => (ListSort.sort    
		      (fn ((a,_), (b,_)) => Symbol.cmpSymbol (a,b) = GREATER)
		      l))

	fun isVariableS s = Char.isUpper (String.sub (s, 0))
	    
	fun isVariable (Var _) = true
	  | isVariable _ = false

	fun isConstant (Atom _) = true
	  | isConstant _ = false

	fun topFeature () = Top

	fun makeVariable x = Var x
	fun genVariable () = Var (genSymbol "_")
	fun genConst () = Atom (genSymbol "*c")

	fun cmpList (nil,nil) = EQUAL
	  | cmpList (nil,_) = LESS
	  | cmpList (_,nil) = GREATER
	  | cmpList ((s,t)::more,(s',t')::more') =
	    (case cmpSymbol (s,s') of
		 EQUAL =>
		     (case cmpFeatures (t,t') of
			  EQUAL =>
			      cmpList (more, more')
			| result => 
			      result)
	       | result => 
		     result)
		 
	and cmpFeatures (Top, Top) = EQUAL
	  | cmpFeatures (Top, _) = LESS
	  | cmpFeatures (_, Top) = GREATER
	  | cmpFeatures (Atom s, Atom t) = cmpSymbol(s,t)
	  | cmpFeatures (Atom _, _) = LESS
	  | cmpFeatures (_, Atom _) = GREATER
	  | cmpFeatures (Compound l, Compound l') = cmpList(l,l')
	  | cmpFeatures (_, Compound _) = GREATER
	  | cmpFeatures (Compound _, _) = LESS
	  | cmpFeatures (Var s, Var t) = cmpSymbol(s,t)

	structure OrdTerm : ORD_KEY = 
	    struct
		type ord_key = features
		val compare = cmpFeatures
	    end

	local
	    exception ReadError

	    fun read_list (")"::more) sofar =
		(Compound (ListSort.sort 
			   (fn ((a,_),(b,_)) => 
			    Symbol.cmpSymbol (a, b) = GREATER)
			   sofar),
		 ")"::more)
	      | read_list ("]"::more) sofar =
		(Compound (ListSort.sort 
			   (fn ((a,_),(b,_)) => 
			    Symbol.cmpSymbol (a, b) = GREATER)
			   sofar),
		 "]"::more)
	      | read_list (";"::more) sofar =
		read_list more sofar
	      | read_list ("("::more) sofar =
		let
		    val (subterm, further) = read_list more sofar
		in
		    case (subterm, further) of
			(Compound sublist, ")"::rest) => read_list further sublist
		      | (Top, ")"::rest) => read_list further nil
		      | _ => raise ReadError
		end
	      | read_list (s::strs) sofar = 
		let 
		    val (subcat, more) = read_term strs
		in
		    read_list more ((Symbol.makeSymbol s, subcat)::sofar)
		end
	      | read_list nil sofar = raise ReadError
		
	    and read_term ("("::")"::more) = (Top, more)

	      | read_term ("("::more) =
		(case read_list more nil of
		     (t, ")"::stuff) => (t, stuff)
		   | _ => raise ReadError)

	      | read_term (s::more) = 
		(if isVariableS s then
		     Var (makeSymbol s)
		 else 
		     Atom (makeSymbol s),
		 more)

	      | read_term nil = raise ReadError
	in
	    fun readFeatures s = 
		let 
		    val (t, m) = read_term s
		in
		    (SOME t, m)
		end handle ReadError => (NONE, s)
	end

	fun emptyBinding () = Dict.empty : binding

	local
	    fun wf Top = "()"
	      | wf (Atom s) = (nameSymbol s)
	      | wf (Var s) = (nameSymbol s)
	      | wf (Compound l) = "(" ^ wprs l ^ ")"

	    and wprs (nil) = ""
	      | wprs [(feature, value)] =
		(nameSymbol feature) ^ " " ^ (wf value)
	      | wprs ((feature, value) :: rest) =
		(nameSymbol feature) ^ " " ^ (wf value) ^ "; " ^ (wprs rest)
	in
	    val writeFeatures = wf
	    val printFeatures = print o wf
	end

	fun writeprettyFeatures binding =
	    let
		fun writeFeature Top = "()"
		  | writeFeature (Atom s) = (nameSymbol s)
		  | writeFeature (Var s) =
		    (case Dict.find (binding, s) of
			 NONE => (nameSymbol s)
		       | SOME t => writeFeature t)
		  | writeFeature (Compound l) =
		    "(" ^ writePairs l ^ ")"

		and writePairs nil = ""
		  | writePairs [(feature, value)] =
		    (nameSymbol feature) ^ " " ^ (writeFeature value)
		  | writePairs ((feature, value) :: more) =
		    (nameSymbol feature) ^ " " ^ (writeFeature value) ^ "; " ^ 
		    (writePairs more)
	    in
		writeFeature
	    end

	fun prettyprintFeatures b = print o (writeprettyFeatures b)

	fun newcopyFeatures (Top, subst) = (Top, subst)
	  | newcopyFeatures (Atom s, subst) = (Atom s, subst)
	  | newcopyFeatures (Var s, subst) =
	    (case Dict.find (subst, s) of
		 NONE => 
		     let
			 val newvar = genVariable()
		     in
			 (newvar, Dict.insert(subst, s, newvar))
		     end
	       | SOME value =>
		     (value, subst))
	  | newcopyFeatures (Compound l, subst) =
	    let 
		val (copy, newsubst) = (foldr newcopyPair (nil,subst) l)
	    in
		(Compound copy, newsubst)
	    end

	and newcopyPair ((feature, value), (sofar, subst)) =
	    let 
		val (copy, newsubst) = newcopyFeatures (value, subst)
	    in
		((feature, copy) :: sofar, newsubst)
	    end

	(* no occurs check... *)

	local 
	    exception Failed

	    fun dereference (Var w, binding) =
		(case Dict.find (binding, w) of
		     NONE => Var w 
		   | SOME thing => dereference (thing,binding))
	      | dereference (t, b) = t

	    fun unify (Var s, t, binding) =
		bind(s, dereference (t,binding), binding)
	      | unify (t, Var s, binding) =
		bind(s, dereference (t,binding), binding)
	      | unify (Top, Top, binding) = 
		(Top, binding)
	      | unify (Top, s as Compound _, binding) =
		(s, binding)
	      | unify (s as Compound _, Top, binding) =
		(s, binding)
	      | unify (Atom s, Atom t, binding) =
		(if sameSymbol (s,t) then
		     (Atom t, binding)
		 else raise Failed)
	      | unify (Compound l1, Compound l2, binding) =
		let
		    val (l3, newbinding) = unifyLists (l1, l2, binding)
		in
		    (Compound l3, newbinding)
		end
	      | unify _ = raise Failed

	    and bind (s, Var w, binding) =
		if sameSymbol (s, w) then
		    (Var w, binding)
		else
		    bindterm(s, Var w, binding)
	      | bind (s, t, binding) =
		bindterm(s, t, binding)
		
	    and bindterm(s, t, binding) =
		(case Dict.find (binding, s) of
		     NONE =>
			 (Var s, Dict.insert(binding, s, t))
		   | SOME t2 =>
			 let
			     val (t3, newbinding) = unify (t2, t, binding)
			 in
			     (Var s, Dict.insert(newbinding, s, t3))
			 end)

	    and unifyLists (nil, l, binding) = (l, binding)
	      | unifyLists (l, nil, binding) = (l, binding)
	      | unifyLists (s as (sfeature, svalue) :: srest,
			    t as (tfeature, tvalue) :: trest,
			    binding) =
		(case cmpSymbol (sfeature, tfeature) of
		     LESS =>
			 mergePair((sfeature, svalue), srest, t, binding)
		   | GREATER => 
			 mergePair((tfeature, tvalue), s, trest, binding)
		   | EQUAL =>
			 let 
			     val (newvalue, newbinding) =
				 unify(svalue, tvalue, binding)
			 in
			     mergePair((sfeature, newvalue), srest, trest, newbinding)
			 end)
		     
	    and mergePair(h, s, t, binding) =
		let
		    val (rest, newbinding) =
			unifyLists(s, t, binding)
		in
		    (h :: rest, newbinding)
		end
	in
	    fun unifyFeatures (f1, f2, binding) =
		SOME (unify(f1, f2, binding))
		handle Failed => NONE
	end

	fun roughSubsumes (Var s, t, binding) =
	    (case Dict.find (binding, s) of
		 NONE => true
	       | SOME t2 => roughSubsumes(t2, t, binding))
	  | roughSubsumes (t, Var s, binding) =
	    (case Dict.find (binding, s) of
		 NONE => false
	       | SOME t2 => roughSubsumes(t, t2, binding))

	  | roughSubsumes (Top, _, _) = true
	  | roughSubsumes (_, Top, _) = false
	  | roughSubsumes (Atom s, Atom t, _) = sameSymbol(s,t) 
	  | roughSubsumes (Compound l1, Compound l2, binding) =
	    checklist(l1, l2, binding)
	  | roughSubsumes _ = false

	and checklist(nil, _, _) = true
	  | checklist(_, nil, _) = false
	  | checklist(s as (sfeature, svalue) :: srest,
		      t as (tfeature, tvalue) :: trest,
		      binding) =
	    (case cmpSymbol (sfeature, tfeature) of
		 LESS => false
	       | EQUAL => 
		     roughSubsumes(svalue, tvalue, binding) andalso
			 checklist(srest, trest, binding)
		   | GREATER => 
			 checklist(s, trest, binding))
    end
