(*
 * morph.sml
 * 
 * MS - 12/95 - first draft 
 *  updated quickly 6/97
 *
 * backwards morphological analyzer;
 * takes specifications of the form
 * 
 * <word> = begin
 *    <features> ~> <form> ;
 *    <features> ~> <form> ;
 * end.
 *
 * and performs appropriate reasoning
 *
 *)

functor Morph(structure Features : FEATURES) : MORPH =
    struct
	structure Features = Features
	open Features

	datatype spec = SPEC of {features: features, 
				 form:string }
	datatype entry = ENTRY of {lemma:string, 
				   forms:spec list }

	fun key (ENTRY{lemma,forms}) = lemma

	fun printSpec (SPEC{features, form}) =
	    (print "   ";
	     printFeatures features;
	     print " ~> ";
	     print form;
	     print " ;\n")

	fun printEntry (ENTRY{lemma, forms}) =
	    (print lemma;
	     print " = begin \n";
	     app printSpec forms;
	     print "end.\n")

	fun nabToSemiNormalize (";" :: more, sofar, _) =
	    SOME (String.concat(rev sofar), more)

	  | nabToSemiNormalize ("" :: more, sofar, state) =
	    nabToSemiNormalize(more, sofar, state)

	  | nabToSemiNormalize ("-" :: more, sofar, state) =
	    nabToSemiNormalize(more, "-" :: sofar, false)

	  | nabToSemiNormalize (s :: more, sofar, state) =
	    if Char.isPunct (String.sub(s,0)) then
		nabToSemiNormalize(more, s :: sofar, true)
	    else 
		if state then
		    nabToSemiNormalize(more, (" " ^ s):: sofar, true)
		else
		    nabToSemiNormalize(more, s :: sofar, true)

	  | nabToSemiNormalize _ = NONE
	    
	fun readSpec info =
	    (case (readFeatures info) of
		 (SOME features, "~>" :: rest) =>
		     (case nabToSemiNormalize(rest, nil, false) of
			  SOME(form, more) =>
			      (SOME (SPEC{features=features, 
					  form=form}), more)
			| _ => (NONE, info))
	       | _ => (NONE, info))

	local 
	    fun readSpecs (specs, "end"::"."::more) = SOME(specs, more)
	      | readSpecs (specs, info) =
		(case readSpec info of
		     (SOME spec, more) => 
			 readSpecs(spec::specs, more)
		   | _ => NONE)
	in
	    fun readEntry (arg as (word :: "=" :: "begin" :: specstr)) =
		(case readSpecs(nil, specstr) of
		     (SOME(specs, more)) =>
			 (SOME(ENTRY{lemma=word, forms=rev specs}), more)
		   | _ => (NONE, arg))
	      | readEntry arg = (NONE, arg)
	end

	fun match (instf, instb)  (ENTRY{lemma, forms}) =
	    let
		fun good nil = NONE
		  | good (SPEC{form,features}::more) = 
		    if roughSubsumes(features, instf, instb) then
			SOME form
		    else
			good more
	    in
		good forms
	    end
    end		    
		    
functor MorphDB(structure Morph:MORPH;
		structure PolyHashTable : HASH_TABLE) : MORPHDB =
    struct

	structure Morph = Morph

	type morphdb = (string,Morph.entry) PolyHashTable.hash_table

	exception Miss
	
	fun newdb () = 
	    PolyHashTable.mkTable(HashString.hashString,op =)
	    (1000, Miss) : morphdb

	fun addEntry table entry =
	    PolyHashTable.insert table (Morph.key entry, entry)
	    
	fun findForm (table:morphdb) (lemma, instf, instb) =
	    let
		val _ =
		    if Debug.debugp "morphs" then
			(print "\nlooking up form ";
			 print lemma;
			 print " for features ";
			 Morph.Features.prettyprintFeatures instb instf;
			 print "\n")
		    else
			()
	    in
		Morph.match (instf, instb) 
		(PolyHashTable.lookup table lemma)
		handle Miss => NONE
	    end

	fun dump table = 
	    PolyHashTable.app (Morph.printEntry) table

	val _ = Debug.register ("morphs",
				"Print out info from the morph db\n",
				false)
    end

