
functor Word (structure Call : CALL;
	      structure Features : FEATURES;
	      structure Tag : TAG;
	      structure Formula : FDB;
	      sharing Call.Term = Formula.Term;
	      sharing Tag.Features = Features
	      sharing type Tag.data = Call.call) : WORD =
    struct
	structure Call = Call
	structure Features = Features
	structure Tag = Tag
	structure Formula = Formula
		
	datatype word =
	    W of {name: string list,
		  basic: bool,
		  kind: Tag.kind,
		  parms: Call.call,
		  call: Call.call,
		  match: Features.features,
		  semantics: Formula.formula,
		  presup: Formula.formula,
		  pragmatics: Formula.formula,
		  trees: Call.call list}

	datatype lexicon = 
	    L of word list

	datatype wordMatch = 
	    M of {name: string list,
		  basic: bool,
		  parms: Call.call,
		  call: Call.call,
		  init: Call.init,
		  match: Features.features,
		  semantics: Formula.formula option,
		  presup: Formula.formula,
		  trees: Call.call list,
		  specinfo: Call.info,
		  oppo: Tag.opportunity}

	fun printName (W{name,...}) =
	    let
		fun pn nil = ()
		  | pn [x] = print x
		  | pn (h::t) = (print h;
				 print " ";
				 pn t)
	    in
		pn name
	    end

	(* printing *)

	fun printWord (W{name,basic,call,match,
			 semantics,presup,pragmatics,trees,...}) =
	    (app (fn x => (print x; print " ")) name;
	     if basic then print " (basic)" else ();
		 print ": ";
		 Call.printCall call;
		 Features.printFeatures match;
		 print " : ";
		 Formula.printFormula print semantics;
		 print " ; ";
		 Formula.printFormula print presup;
		 print " ; ";
		 Formula.printFormula print pragmatics;
		 print " ; ";
		 app (fn x => (Call.printCall x; print ", ")) trees;
		 print ".\n")

	fun printLexicon (L words) =
	    app printWord words

	fun printMatch (M{name,specinfo,trees,...}) =
	    (app (fn x => (print x; print " ")) name;
	     app (fn x => (Call.prettyprintCall specinfo x;
			   print ", ")) trees;
	     print "\n")

	(* reading *)

	fun newLexicon () = L nil

	datatype word_part =
	    NAME of string list
	  | BASIC of bool
	  | CALL of Call.call
	  | MATCH of Features.features
	  | FML of Formula.formula
	  | TREES of Call.call list

	local 
	    val substCall =
		(case Call.readCall 0 [ "alpha" ] of
		     (SOME c, _) => c
		   | _ => (print "(word): couldn't read substCall\n";
			   raise Match))

	    val adjoinCall =
		(case Call.readCall 0 [ "beta" ] of
		     (SOME c, _) => c
		   | _ => (print "(word): couldn't read adjoinCall\n";
			   raise Match))


	    fun getKind (call) = 
		if Call.overlap(call, substCall) then
		    Tag.ToSubst
		else
		    Tag.ToAdjoin
	in
	    fun wordFields (collection) =
		let
		    open FieldStr
		    fun checkDecl (call) =
			if  Call.overlap (call, substCall) orelse
			    Call.overlap (call, adjoinCall) then
			    SOME (CALL call)
			else
			    (print "Declaration must be alpha(...) or beta(...)\n";
			     NONE)
			    
		    fun checkTrees list =
			let
			    fun checkUniq (nil, sofar) = ()
			      | checkUniq (t::r, sofar) =
				if (List.exists (fn x=> Call.overlap(t, x)) sofar) then 
				    (print "WARNING: trees have non-unique names: ";
				     Call.printCall t;
				     print ".\n";
				     checkUniq(r, t::sofar))
				else
				    checkUniq(r, t::sofar)
			in
			    checkUniq (list, nil);
			    SOME (TREES list)
			end

		    val ppvar = Call.newPPVar ()
		    fun rcq s = Call.readCall ppvar s
		    val rfq = Formula.readFormula collection
		in
		    [F{name = "name",
		       read = wrapReader (SOME o NAME, readStrings)},
		     F{name = "basic",
		       read = wrapReader (SOME o BASIC, readBool)},
		     F{name = "decl",
		       read = wrapReader (checkDecl, rcq)},
		     F{name = "site",
		       read = wrapReader (SOME o CALL, rcq)},
		     F{name = "match",
		       read = wrapReader (SOME o MATCH, Features.readFeatures)},
		     F{name = "semantics",
		       read = wrapReader (SOME o FML, rfq)},
		     F{name = "presupposition",
		       read = wrapReader (SOME o FML, rfq)},
		     F{name = "pragmatics",
		       read = wrapReader (SOME o FML, rfq)},
		     F{name = "trees",
		       read = wrapReader (checkTrees, readListOf rcq)} ]
		end

	    val fieldInit = [ "w", "true", "alpha()", "np()", "()", "true", "true", "true", "" ]

	    fun readWord collection =
		let
		    val fields = wordFields collection
		    fun assemble([NAME s, BASIC b, CALL d, CALL c, MATCH f,
				  FML sem, FML presup, FML prag, TREES trees]) =
			SOME( W{name = s,
				basic = b,
				kind = getKind d,
				parms = d,
				call = c,
				match = f,
				semantics = sem,
				presup = presup,
				pragmatics = prag,
				trees = trees} )
		      | assemble _ =
			(print "(word) assembly: results don't match fields\n";
			 NONE)
		in
		    fn s => (case FieldStr.readstruct ("word", fields) s of
				 (FieldStr.GOT a, s') => (assemble a, s')
			       | (FieldStr.EMPTY, s') => (NONE, s')
			       | (FieldStr.ERROR e, s') => 
				     (print e; (NONE, s')))
				      
		end

	end (* local *)

	fun addWord (L wl, w) = L (w :: wl)

	fun matchWord ((publicworld, publicset),
		       (privateworld, privateset))
	    (oppo as Tag.OP{kind, address, match, data},
	     specinfo)
	    (W{name, kind = k2, parms, basic, call, match=features, 
	       semantics, presup, pragmatics, trees}) =
	    let
		exception AmbiguousPresups
		val (copy, copyinit) = Call.newcopyCall(call, Call.mkInit ())
		val (copyparms, midinit) = Call.newcopyCall(parms, copyinit)
		val (newtrees, newinit) = U.foldstate Call.newcopyCall (trees, midinit)

		fun uniquecxt (publicworld) (trans, f) = 
		    let
			val answers = publicset (trans, f)
		    in
			case answers of 
			    [one] => 			
				Formula.makeNewTrans 
				(one, trans, publicworld)
			  | _ => raise AmbiguousPresups
		    end

		fun fails (f, i) =
		    (case publicset(i, f) of
			 nil => true
		       | _ => false)

		fun check (presupvar, newinfo) =
		    let
			fun finfo (t,p) =
			    let
				val result = 
				    Call.setWithInit(newinfo, newinit, t,
						     case presupvar of
							 Call.E x => SOME(Call.E x, p)
						       | _ => NONE)

				val _ = 
				    (case result of 
					 SOME thing =>
					     (if not (Call.checkKeyGround thing copy) then
						  (print "ACK: finfo not ground\n";
						   print " presupvar is ";
						   print (case presupvar of
							      Call.E x => ("var " ^ (Int.toString x))
							    | _ => ("value"));
						   print " call is ";
						   Call.prettyprintCall thing copy;
						   print "\nyikes\n")
					      else ())
				       | _ => ())
			    in	
				result
			    end

			fun checkAsserted (tinit) =
			    let
				fun checke pf =
				    let
					val it = Formula.makeNewTrans(pf, tinit, publicworld)
					val uc = uniquecxt (publicworld)
					val i = uc (uc (it, presup), pragmatics)
				    in 
					if (fails (semantics, i)) then
					    case finfo (i, Call.Asserted) of
						NONE => NONE
					      | SOME fi =>
						    let
							val match =
							    M{name=name,
							      parms=copyparms,
							      call=copy,
							      init=newinit,
							      match=features,
							      basic = basic,
							      semantics = SOME (Formula.rename 
										(i,publicworld) 
										semantics),
							      presup = presup,
							      oppo = oppo,
							      specinfo = fi,
							      trees = newtrees}
						    in
							if not (Call.checkKeyGround fi copyparms) then
							    (print "ACK: checkAsserted not ground\n")
							else ();
							    SOME match
						    end
					else
					    NONE
				    end handle AmbiguousPresups => NONE

				val extensions = privateset(tinit, semantics)
				val result =
				    List.mapPartial checke extensions
			    in
				result
			    end
			fun checkPresupposed (tinit) =
			    let
				fun checke pf =
				    let
					val uc = uniquecxt (publicworld)
					val it = Formula.makeNewTrans(pf, tinit, publicworld)
					val i = uc(it,pragmatics)
				    in 
					case finfo(i, Call.Presupposed) of
					    NONE => NONE
					  | SOME fi =>
						let
						    val match =
							M{name=name,
							  parms=copyparms,
							  call=copy,
							  init=newinit,
							  match=features,
							  basic=basic,
							  semantics = NONE,
							  presup = semantics,
							  oppo = oppo,
							  specinfo = fi,
							  trees = newtrees}
						in
						    if not (Call.checkKeyGround fi copyparms) then
							(print "ACK: checkpresupposed not ground\n")
						    else 
							();
							SOME match
						end
				    end handle AmbiguousPresups => NONE

				val extensions = publicset(tinit, semantics)
				val result = 
				    (List.mapPartial checke extensions)
			    in
				result
			    end
		    in
			op@((case presupvar of
				 Call.V Call.Presupposed => nil
			       | _ => checkAsserted 
				     (Call.getInit(newinfo, copyparms, 
						   privateworld))),
			    (case presupvar of
				 Call.V Call.Asserted => nil
			       | _ => checkPresupposed 
				     (Call.getInit(newinfo, copyparms, 
						   publicworld))))
		    end
	    in
		if kind <> k2 then
		    (nil)
		else 
		    (case Call.match(data, copy, specinfo) of
			 SOME (presupvar, newinfo) =>
			     check(presupvar, newinfo)
		       | _ => 
			     (nil))
	    end

	fun specOrder test (L words) =
	    let
		open Formula
		fun asSpec (W{semantics=s1, presup=t1, pragmatics=p1,...}, 
			    W{semantics=s2, presup=t2, pragmatics=p2,...}) =
		    let
			fun t (f1,f2) =
			    case test (Term.IntMap.empty(), mkClosedImplic([f1],f2)) of
				nil => false
			      | _ => true
		    in
			t(p1, p2) andalso
			(not (t(p2, p1)) orelse (t (s1, s2) andalso t (t1,t2)))
		    end

		fun findleast (w, nil, left) = (w, rev left)
		  | findleast (w, w'::m, left) =
		    if asSpec(w', w) then
			findleast(w', m, w :: left)
		    else
			findleast(w, m, w' :: left)

		fun sort (nil, done) = rev done
		  | sort (w::more, done) =
		    let
			val (w', next) = findleast(w,more,nil)
		    in
			sort(next, w'::done)
		    end

		val result = sort(words, nil)
	    in
		L result
	    end

	fun matchAll infos call (L words) =
	    let
		val result = 
		    (foldr op@ nil (map (matchWord infos call) words))
	    in
		result
	    end

					  
	     

	local
	    fun killDuplicates (nil, this, sofar) = rev (this :: sofar)
	      | killDuplicates ((call, info) :: more,
			      (thiscall, thisinfo),
			      sofar) =
		if not (Call.overlap(call, thiscall)) then
		    killDuplicates(more, (thiscall, thisinfo),
				   (call, info) :: sofar)
		else
		    if Call.duplicates(call, info, thiscall, thisinfo) then
			(rev sofar) @ ((thiscall, thisinfo) :: more)
		    else
			killDuplicates(more, (thiscall, thisinfo),
				       (call, info) :: sofar)

	    fun prune info (tree, list) =
		if not (Call.checkKeyGround info tree) then
		    (print "ACK: tree is not ground\n";
		     Call.prettyprintCall info tree;
		     print "\n";
		     (tree, info) :: list)
		else
		    killDuplicates(list, (tree, info), nil)
	in
	    fun getCalls list = 
		let
		    fun gc (calls, nil) = calls
		      | gc (calls, M{trees, specinfo,...} :: more) = 
			gc (foldr (prune specinfo) calls trees,
			    more)

		    val result = gc (nil, list)

		    val _ =
			if Debug.debugp "word" then
			    (print "Calls before: ";
			     app printMatch list;
			     print "\nCalls after: ";
			     app (fn (c,i) => (Call.prettyprintCall i c;
					       print "\n")) result;
			     print "\n")
			else
			    ()
		in
		    result
		end
	end
    
	val _ = Debug.register("word", "Stuff about words", false);
    end
