(*
 *
 *)

signature ASSUME =
    sig
	structure Term : TERM

	type assumption

	val prettyprintAssumptions: (string -> unit) -> 
	    ((Term.Patheq.pathpt option -> string) option -> 
	     (Term.init Term.IntMap.intmap option * Term.term) -> unit) * 
	    (Term.Patheq.pathpt -> string) * 
	    assumption list -> unit

	val printAssumptions: (string -> unit) -> 
	    (Term.init Term.IntMap.intmap * string list -> unit) *
	    assumption list -> unit

	val matchAssumptions: int * Term.Patheq.Modspec.absmod * Term.Patheq.Modspec.absmod ->
	    Term.term * Term.init Term.IntMap.intmap * Term.Patheq.pathpt *
	    assumption list * Term.system * Term.Patheq.problem ->
	    (assumption list * Term.system * Term.Patheq.problem) list
    end

functor Assume(structure Term: TERM) : ASSUME =
    struct 

	structure Term = Term

	datatype assumption =
	    Assume of {trans: Term.init Term.IntMap.intmap,
		       var: int,
		       index: Term.Patheq.pathpt,
		       endPath: Term.Patheq.pathpt}
	    
	local
	    exception bad
	    val assumable = ["occur", "(", "X", ")"]
	    val actionVar = Term.getVar "X"
	    val matchTerm = case Term.readTerm assumable of
			     (SOME t, nil) => t
			   | _ => raise (Fail "Assume.sml: read term didn't accept assumable\n")
	in
	    fun makeAssumption (term, trans, index,
				knowledge, time, 
				maxass, assumptions,
				system, problem) =
		if length assumptions >= maxass then
		    nil
		else let
			 val (occurPath, optw) =
			     Term.Patheq.extendVars(problem,
						    Term.Patheq.rootProblem problem, 
						    Term.Patheq.rootProblem problem)

			 val (assumePath, aptw) =
			     Term.Patheq.extendVar(problem,
						   occurPath,
						   optw,
						   knowledge)
			     
			 val (endPath, _) =
			     Term.Patheq.extendVar(problem,
						   assumePath,
						   aptw,
						   knowledge)
			     
			 val p1 = 
			     case Term.Patheq.equate(problem, endPath, index) of
				 NONE => raise bad
			       | SOME p => p
			 val p2 =
			     case Term.Patheq.separateTop (p1, occurPath, assumePath) of
				 NONE => raise bad
			       | SOME p => p

			 val actName = Term.newVar ()
			 val actTrans = Term.IntMap.insert(Term.IntMap.empty(),
							   actionVar,
							   Term.VI(actName, occurPath))
			 val (newSystem, nextProblem) = 
			     Term.unify(Term.updateSystem(matchTerm, actTrans, endPath,
							  term, trans, index,
							  p2, system))
			     
			 val final = Term.Patheq.updateVtables nextProblem
		     in
			 case final of
			     NONE => (print "Modal scope violation.\n"; nil)
			   | SOME newProblem => 
				 [(Assume{trans = actTrans,
					  index = Term.Patheq.keepCopy assumePath,
					  var = actName,
					  endPath = Term.Patheq.keepCopy endPath} :: assumptions,
				   newSystem,
				   newProblem)]
		     end 
		 handle bad => nil
		      | Term.Clash => (if Debug.debugp "match" then
					   print "Assumption-Unification clashed.\n"
				       else (); nil)
		      | Term.Cycle => (if Debug.debugp "match" then
					   print "Occurs check violated.\n"
				       else (); nil)
		      | Term.Scope => (if Debug.debugp "match" then
					   print "No solution for scope constraints.\n"
				       else (); nil)
		      | Fail x => (print "Failure in make assumption!\n";
				   print ("Reason: " ^ x ^ "\n");
				   nil)
			    

	    fun testAssume(term, ttrans, tindex, 
			   knowledge, time,
			   Assume{trans, var, index, endPath}, assumptions,
			   system, problem) =
		let
		    val (newEndPath, _) =
			Term.Patheq.extendVar(problem, index, index, knowledge)
			
		    val p1 = 
			case Term.Patheq.equate(problem, newEndPath, tindex) of
			    NONE => raise bad
			  | SOME p => p

		    val (newSystem, newProblem) = 
			Term.unify(Term.updateSystem(matchTerm, trans, newEndPath,
						     term, ttrans, tindex,
						     p1, system))
		in
		    SOME(assumptions, newSystem, newProblem)
		end 
	    handle bad => NONE
		 | Term.Clash => (if Debug.debugp "match" then
				      print "Assumption-Unification clashed.\n"
				  else (); NONE)
		 | Term.Cycle => (if Debug.debugp "match" then
				      print "Occurs check violated.\n"
				  else (); NONE)
		 | Term.Scope => (if Debug.debugp "match" then
				      print "No solution for scope constraints.\n"
				  else (); NONE)
		 | Fail x => (print "Failure in test assumption!\n";
			      print ("Reason: " ^ x ^ "\n");
			      NONE)
		    
	    fun matchAssumptions(maxass, knowledge, time)
		(term, trans, tindex, 
		 assumptions, system, problem) =

		if Term.quickMatch(matchTerm, Term.IntMap.empty(), 
				   term, trans, system) then
		    (makeAssumption(term, trans, tindex, knowledge, time,
				    maxass, assumptions, system, problem)) @
		    (List.mapPartial
		     (fn x => testAssume(term, trans, tindex, 
					 knowledge, time,
					 x, assumptions, system, problem))
		     assumptions)
		else
		    nil

	    fun printAssumptions out =
		let
		    fun pa pt (Assume{trans,var,index,endPath}) =
			(out " K ";
			 pt (trans,assumable);
			 out " at ";
			 Term.Patheq.printPathpt out index;
			 out ";\n")
		in
		    (fn (pt, assumptions) =>
		     (out "Plan is:\n"; app (pa pt) assumptions))
		end

	    fun prettyprintAssumptions out =
		let
		    fun pa (pt,pp) (Assume{trans,var,index,endPath}) =
			(out " K ";
			 pt  (SOME (fn _ => pp index)) (SOME trans, matchTerm);
			 out ";\n")
		in
		    (fn (pt, pp, assumptions) =>
		     (out "Plan is:\n"; app (pa (pt, pp)) assumptions))
		end
	end
    end
