(*
 * Matthew Stone
 * Dialup 0.1 - 3/96
 * cterm-sig.sml
 *
 * Signature file for terms and unification.
 *)

signature TERM =
    sig
	structure IntSet : MYINTSET
	structure IntMap : MYINTMAP
	structure Patheq : PATHEQ

	type term
	type system

	datatype 'a desc =
	    Bound of 'a
	  | Free 
	  | Miss

	(* 
	 * translations of variables:
	 * VI(i,world) = logic variable with id i,
	 *               introduced at world.
	 * CI(i,term,world,args) =
	 *               herbrand function with 
	 *               head given by term, introduced
	 *               at world, and with arguments given
	 *               by the variables in ARGS;
	 *               (can be referenced by variable i).
	 * TI(i,world, term) = initialization for variable i
	 *              of copiable term t.
	 * (world is used to enforce an increasing-
	 * domain restriction).
	 *)
	datatype init = 
	    VI of int * Patheq.pathpt 
	  | OI of int * Patheq.pathpt 
	  | CI of int * term * Patheq.pathpt * int list
	  | TI of int * Patheq.pathpt * term


	val metaVar : term

	(* get a variable code for the passed string *)
	val getVar: string -> int

	(* gensym: make a fresh variable code *)
	val newVar: unit -> int

	(* gensym: make a fresh herbrand function symbol *)
	val newConst : unit -> term

	(*
	 * do the terms have the same head function symbol?
	 * if not, we can avoid costly unification between them 
	 *)
	val sameHead : term * term -> bool
	val quickMatch : term * init IntMap.intmap * 
	                 term * init IntMap.intmap * system  -> bool

	val deepMatch :  term * init IntMap.intmap * system *
	                 term * init IntMap.intmap * system  -> bool

	val checkGround : term * init IntMap.intmap * system -> bool
	val rename : init IntMap.intmap -> term -> term
	val subst : init IntMap.intmap -> term -> term
	val bindArg: term * int * Patheq.pathpt -> init IntMap.intmap

	(* I/O *)
	val readTerm: string list -> term option * string list

	(* Add to argument SET the codes of free vars in TERM *)
	val freeVarsTerm: term * IntSet.intset -> IntSet.intset

	(* make a new system of equations, working over problem *)
	val newSystem: Patheq.problem -> system

	(* describe system *)
	val writeHead : term -> string
	val writeTerm : term -> string
	val writeReifiedTerm : ((Patheq.pathpt -> string) * 
				(Patheq.pathpt option -> string) option) -> term -> string

	val printTerm: (string -> unit) -> term -> unit
	val readoutVar : system * int option -> term desc
	val printTrans: init IntMap.intmap -> unit

	val writeprettyTerm : system -> (init IntMap.intmap option * term) -> string
	val writeprettyVar : system -> int -> string
	val writeprettyReifiedTerm : (system * (Patheq.pathpt -> string) * 
				      (Patheq.pathpt option -> string) option) -> 
	    (init IntMap.intmap option * term) -> string
	val writeprettyReifiedVar : (system * (Patheq.pathpt -> string) * 
				     (Patheq.pathpt option -> string) option) -> 
	    int -> string

	val printSystem : (string -> unit) -> system -> unit

	(* build a system of equations including $ = t *)
	val makeSystem: term desc * Patheq.problem -> system

	(*
	 * given term1, translation1, and its world,
	 * and   term2, translation2, and its world,
	 * and   an updated constraint problem, 
	 * and   an old system of equations
	 *
	 * translate the terms and assemble variable counts
	 * and do other initialization to build an overal system.
	 *)
	val updateSystem: (term * init IntMap.intmap * Patheq.pathpt *
			   term * init IntMap.intmap * Patheq.pathpt * 
			   Patheq.problem * system -> system)

	val updateVar : (int * term * init IntMap.intmap * Patheq.pathpt *
			 Patheq.problem * system -> system)

	(*
	 * transform a system into solved form, reporting
	 * both new system and its associated tree constraint
	 * problem -- raising one of the passed exceptions
	 * in the case of failure. 
	 *)
	exception Clash
	exception Cycle
	exception Scope
	val unify: system -> system * Patheq.problem
	val collapseSystem: system -> system * Patheq.problem

	(* 
	 * matchTerms: does system say two terms are equal,
	 *   given (just) the unifications already done.
	 * precedes: does system rank the first var ahead
	 *   of the second var --- possible only if the 
	 *   value of the first does not contain that of
	 *   the second.
	 *)
	val matchTerms: system -> term * term -> bool
	val matchVars: system -> int * int -> bool
	val follows: system -> int * int -> bool
	val bestAlias: system * int -> int

    end

