(*
 * Matthew Stone 
 * Dialup 0.1 - 3/96
 *
 * Implementation of terms with unification. 
 * The unification algorithm follows Martelli & Montanari
 * (1982), but adds the machinery required to enforce
 * an increasing domain restriction on the values of
 * variables.
 *)

functor ConstrainedTerm (structure Symbol : SYMBL;
			 structure Patheq : PATHEQ;
			 structure IntMap : MYINTMAP;
			 structure IntSet : MYINTSET) : TERM =
    struct
	structure IntSet = IntSet
	structure IntMap = IntMap
	structure Patheq = Patheq

	(*
	 * types of function symbols: 
	 *  rigid - give the same value for args 
	 *      at all possible worlds.
	 *  varying - give possibly different
	 *      values for the same args at different
	 *      possible worlds.
	 *)
	datatype rigidity = RIGID | VARYING 

	datatype 'a desc = 
	    Bound of 'a
	  | Free
	  | Miss

	(*
	 * The term representation uses the 'multiterm'
	 * idea from Martelli & Montanari (p 273).
	 * This structure presents the usual term as
	 * a skeleton where sets of variables can appear
	 * as labels on the nodes.  Two consistent
	 * multiterms can be merged into a single
	 * multiterm --- doing such merges early makes
	 * detection of clashes faster.
	 *
	 * For each function application, we need to know
	 * the kind of function (RIGIDITY) and the WORLD variable
	 * representing the modal index where it is defined.
	 *)
	datatype term =
	    MEMPTY
	  | MAPP of {head: Symbol.symbol, 
		     rigid: rigidity,
		     world: Patheq.pathpt option,
		     args:multipair list}
	and multipair =
	    MPR of {least:int option,
		    vars:IntSet.intset, 
		    mt:term}

	(*
	 * For bound variables, an external module will 
	 * supply translations in context in INITS -- some 
	 * variables will be replaced by fresh logic variables 
	 * (with a particular index at which they must
	 * be defined): VI (variable code, index).
	 * Other variables will be replaced by skolem
	 * constants (CI(code, term, world, args)): TERM
	 * gives the head symbol, WORLD the world where
	 * the skolem term is invoked, and ARGS indices for
	 * the variables that will be the argument
	 * of the function, CODE a name for the application.
	 *)

	datatype init = 
	    VI of int * Patheq.pathpt
	  | CI of int * term * Patheq.pathpt * int list
	  | TI of int * Patheq.pathpt * term
	  | OI of int * Patheq.pathpt

	(*************************************************************
	 * 
	 * Interface routines 
	 *
	 *************************************************************)

	(*
	 * find free variables;
	 * this function is apparently never used;
	 * and may have to be changed to account for uses of
	 * MEMPTY as an anonymous predicate variable
	 *)
	fun freeVarsTerm (MEMPTY, sofar) = sofar
	  | freeVarsTerm (MAPP{args,...}, sofar) =
	    foldr freeVarsPair sofar args

	and freeVarsPair (MPR{least=SOME i, vars, mt}, sofar) =
	    freeVarsTerm (mt, IntSet.union(IntSet.add(vars, i), sofar))
	  | freeVarsPair (MPR{mt,...}, sofar) =
	    freeVarsTerm (mt, sofar)

	    
	(* reading *)
	local 
	    (*
	     * manage the codes which replace the names
	     * of variables
	     *)
	    val ct = ref 0
	    val vars = ref nil : (Symbol.symbol * int) list ref

	    val listNullSym = Symbol.makeSymbol "[]"
	    val listNull = MAPP{head = listNullSym, rigid = RIGID, 
				world = NONE, args = nil}
	    val listNullPair = MPR{least = NONE,
				   vars = IntSet.empty,
				   mt = listNull}

	    val listConsSym = Symbol.makeSymbol "."

	    fun luv (s, nil) = 
		(ct := (!ct) + 1;
		 vars := (s, !ct) :: (!vars);
		 !ct)
	      | luv (s, (t,i) :: rest) =
		if Symbol.sameSymbol(s,t) then i
		else luv(s, rest)

	    (* variables begin with CAPS,
	     * constants with smalls or #s *)
	    fun ivn (s) = Char.isUpper(String.sub(s,0))
	    fun icn (s) = Char.isLower(String.sub(s,0)) orelse
		          Char.isDigit(String.sub(s,0))

	    (* parens encode the character of a function
	     * curly means varying, round means rigid *)
	    fun paren ("(", false) = SOME (")",  RIGID)
	      | paren ("(", true) = SOME (")",  RIGID)
	      | paren ("{", _) = SOME ("}",  VARYING)
	      | paren _ = NONE

	    (* relt: read a variable or constant off the
	     * input string *)
	    fun relt ("$" :: more, false) = 
		(MPR{least=NONE, vars=IntSet.empty,
		     mt=MEMPTY},
		 more)

	      | relt (s :: more, nested) =
		if ivn (s) then 
		    let val i = luv(Symbol.makeSymbol s, !vars)
		    in (MPR{least=SOME i, 
			    vars=IntSet.empty,
			    mt=MEMPTY},
			more)
		    end
		else if icn (s) then
		    (MPR{least=NONE,
			 vars=IntSet.empty,
			 mt=MAPP{head=Symbol.makeSymbol s,
				 rigid=if nested then RIGID else VARYING,
				 world=NONE, args=nil}},
		     more)
		     else raise (Fail ("bad constant or variable: " ^ s))

	      | relt (nil, _) = raise (Fail "end of string")

	    (*
	     * rt: read a term (possibly recursive) off the
	     * input string.
	     *)
	    fun rt ("[" :: "]" :: more, true) = (listNullPair, more)

	      | rt ("[" :: more, true) =
		let
		    val start = rt (more, true)
		in
		    readList true start
		end

	      | rt (s :: ":" :: more, nested) =
		if ivn (s) then
		    let val i = luv(Symbol.makeSymbol s, !vars)
			val (MPR{least, vars, mt}, final) = 
			    rt (more, nested)
		    in
			case least of 
			    NONE => (MPR{least= SOME i,
					 vars= vars,
					 mt = mt},
				     final)
			  | SOME j => (MPR{least= least,
					   vars= IntSet.add(vars, i),
					   mt=mt},
				       final)
		    end
		else
		    raise (Fail ("colon must be preceded by var\n"))

	      | rt (s :: lparen :: more, nested) =
		(case paren (lparen, nested) of
		     NONE => relt(s :: lparen :: more, nested)
		   | SOME (close, rigid) => 
			 let val (args,rest) = ra(more,nil)
			 in
			     case rest of 
				 match :: final => 
				     if match = close then
					 (MPR{least=NONE,
					      vars=IntSet.empty,
					      mt=MAPP{head=Symbol.makeSymbol s,
						      rigid=rigid,
						      world=NONE,
						      args=args}}, 
					  final)
				     else
					 raise (Fail "no matching paren")
			       | _ => raise (Fail "no matching paren")
			 end)

	      | rt (s :: more, nested) = relt(s :: more, nested)
	      | rt (nil, _) = raise (Fail "end of string")

	    (*
	     * ra: read a list of arguments off the input
	     * stream (calling rt to read each one)...
	     *)
	    and ra ("," :: more, sofar) = ra(more, sofar)
	      | ra (")" :: more, sofar) = (rev sofar, ")"::more)
	      | ra ("}" :: more, sofar) = (rev sofar, "}"::more)
	      | ra ("]" :: more, sofar) = (rev sofar, "]"::more)
	      | ra (anything, sofar) =
		let val (first, rest) = rt(anything, true)
		in ra(rest, first::sofar)
		end

	    and readList true (head, "]" :: more) =
		(MPR{least=NONE,
		     vars=IntSet.empty,
		     mt= MAPP{head = listConsSym,
			      rigid = RIGID,
			      world = NONE,
			      args = [ head, listNullPair ]}},
		 more)
	      | readList false (head, "]" :: more) =
		(head, more)
	      | readList true (head, "," :: more) =
		let
		    val (rest, final) = readList true (rt (more,true))
		in
		    (MPR{least = NONE,
			 vars = IntSet.empty,
			 mt = MAPP{head = listConsSym,
				   rigid = RIGID,
				   world = NONE,
				   args = [ head, rest ]}},
		     final)
		end
	      | readList true (head, "|" :: more) =
		let
		    val (rest, final) = readList false (rt (more,true))
		in
		    (MPR{least = NONE,
			 vars = IntSet.empty,
			 mt = MAPP{head = listConsSym,
				   rigid = RIGID,
				   world = NONE,
				   args = [ head, rest ]}},
		     final)
		end
	      | readList _ _ = raise Fail "invalid list"
	in
	    fun getVar s = luv(Symbol.makeSymbol s, !vars)
	    fun newVar () = (ct := !ct + 1; !ct)

	    fun newConst () = MAPP{head=Symbol.genSymbol ("_f"),
				   rigid=VARYING,
				   world=NONE,
				   args=nil}

	    fun readTerm sl =
		let val (MPR{least, vars, mt}, rest) = rt (sl, false)
		in case least of 
		    NONE => (SOME mt, rest)
		  | SOME thing => (print "Top level can't be var\n";
				   (NONE, sl))
		end handle Fail s => (print s;
				      print "\n";
				      (NONE, sl))

	    val metaVar = MEMPTY
	end	    

	(****************************************************************
	 *
	 * UNIFICATION PROPER
	 *
	 ****************************************************************)

	datatype varrole = User | Service | Temporary
	datatype valrole = Existential | OtherRole

	datatype keep = K of varrole * valrole

	local 
	    fun b1 (User, _) = User
	      | b1 (_, User) = User
	      | b1 (Service, _) = Service
	      | b1 (_, Service) = Service
	      | b1 _ = Temporary

	    fun b2 (Existential, _) = Existential
	      | b2 (_, Existential) = Existential
	      | b2 _ = OtherRole
	in
	    fun better (K(r1,l1), K(r2,l2)) = 
		K (b1(r1,r2), b2(l1,l2))
	end
    
	(*
	 * an EQUATION corresponds to the multiequation type of
	 * Martelli & Montanari:
	 *             v1 = v2 = ... = vn = multiterm
	 * Each equation has an identification code, INDEX;
	 * a RANK indicating at what point it was solved in the
	 * last run of unification;
	 * a flag KEEP indicating whether any of the variables here
	 * may appear in subsequent unification terms (if not, we
	 * can presently discard the equation); a modal index var
	 * WORLD indicating where the term here is first defined;
	 * NVARS indicating the number of variables equated here;
	 * VARS is a list of those variables.  MT is the multiterm
	 * they are all equal to.
	 *)
	datatype equation =
	    E of {index: int,
		  rank: int,
		  keep: keep,
		  world: Patheq.pathpt,
		  nvars: int,
		  vars: int list,
		  mt: term}

	(*
	 * A COUNT gives a number of mentions for each variable
	 *)
	datatype count = 
	    C of int IntMap.intmap

	(*
	 * A BINDING has:
	 *    ASGN: indicating the equation which supplies the
	 *      value of each variable (by index).
	 *    CELLS: stores active equations by index
	 *    INCOUNT: number of references to each variable
	 *      in equations yet to be solved
	 *    OUTCOUNT: number of references to each variable
	 *      in equations that have been solved & removed,
	 *      but which will (in time) be added again.
	 *    NEXTCELL: smallest integer with which no equation
	 *      has yet been associated.
	 *)
	datatype binding = 
	    B of {asgn: int IntMap.intmap,
		  nextcell: int,
		  cells: equation IntMap.intmap,
		  incount: count,
		  outcount: count}

	(*
	 *  A SYSTEM has:
	 *    A binding, recording the current status of variables
	 *    	and TERM equations;
	 *    A problem, recording the current status of MODAL
	 *      indices and MODAL equations;
	 *    The list ZEROS of equations whose values are not
	 *      referred to in any (other) equations in BIND
	 *      (as reflected in the incount list).
	 *    A list SOLVED of equations already treated by the
	 *      unification algorithm.
	 *    The number of active equations, NEQ.
	 *    The key index ANON used to hold a 
	 *       anonymous single predicate (new 4/11)
	 *)
	datatype system =
	    S of {bind: binding,
		  problem: Patheq.problem,
		  zeros: equation list,
		  neq: int,
		  anon: int,
		  anonterm: term,
		  solved: equation list}

	(*
	 * Kinds of failure:
	 *   Clash: f = g
	 *   Cycle: X = f(X)
	 *   Scope: X = c in all x box p(x) => box all c p(c)
	 *)
	exception Clash
	exception Cycle
	exception Scope

	(*
	 * an abstraction for the counts of variables:
	 *   increment;
	 *   decrement;
	 *   getCount;
	 *   mergeCounts;
	 *   mergeDecCounts;
	 *)
	fun increment (C table, varcode) = 
	    (case IntMap.peek (table, varcode) of
		 NONE => C (IntMap.insert(table, varcode, 1))
	       | SOME j => C (IntMap.insert(table, varcode, j+1)))

	fun decrement (C table, varcode) =
	    (case IntMap.peek (table, varcode) of
		 NONE => (print "Decrement (newterm): negative # of occurrences!\n";
			  C (IntMap.insert(table, varcode, ~1)))
	       | SOME j => C (IntMap.insert(table, varcode, j-1)))

	fun getCount (C table, varcode) =
	    (case IntMap.peek (table, varcode) of NONE => 0 | SOME j => j)

	fun mergeCounts (c as C table, first, second) =
	    C(IntMap.insert(table, first, getCount(c, first) + getCount(c,second)))

	fun mergeDecCounts (c as C table, first, second) =
	    C(IntMap.insert(table, first, getCount(c, first) + getCount(c,second) - 1))
	    
	fun getAnon (S{bind=B{asgn,cells,...},anon,...}) =
	    (case IntMap.peek(asgn,anon) of
		 NONE => MEMPTY
	       | SOME cell => 
		     case IntMap.find(cells,cell) of
			 E{mt,...} => mt)

	(* avoid unification and matching by quick tests *)
	fun sameHead (MEMPTY, _) = true
	  | sameHead (_, MEMPTY) = true 
	  | sameHead (MAPP{head=h1,...}, MAPP{head=h2,...}) =
	    Symbol.sameSymbol(h1,h2)

	local 
	    fun match deep (t1, tr1, s1, MEMPTY, tr2, s2) = 
		(case getAnon s2 of
		     MEMPTY => true
		   | t2 => match deep (t1, tr1, s1, t2, tr2, s2))

	      | match deep (MEMPTY, tr1, s1, t2, tr2, s2) = 
		(case getAnon s1 of
		     MEMPTY => true
		   | t1 => match deep (t1, tr1, s1, t2, tr2, s2))

	      | match deep (MAPP{head=h1,args=a1,...}, trans1, 
			    S{bind=B{asgn=asgn1,cells=c1,...},...},
			    MAPP{head=h2,args=a2,...}, trans2, 
			    S{bind=B{asgn=asgn2,cells=c2,...},...}) =
		let 
		    fun termize(NONE, t, trans, asgn, cells) = t
		      | termize(SOME var, t, trans, asgn, cells) =
			(case IntMap.peek (trans,var) of
			     SOME (CI (_, t2, _, _)) => t2
			   | SOME (OI (nvar, _)) =>
				 (case IntMap.peek(asgn, nvar) of
				      NONE => t
				    | SOME j => 
					  let
					      val E{mt,...} = IntMap.find(cells,j)
					  in
					      mt
					  end)
			   | SOME (VI (nvar, _)) =>
				 (case IntMap.peek(asgn, nvar) of
				      NONE => t
				    | SOME j => 
					  let
					      val E{mt,...} = IntMap.find(cells,j)
					  in
					      mt
					  end)
			   | SOME (TI(nvar, world, nt)) => 
				 (case IntMap.peek(asgn, nvar) of
				      NONE => nt
				    | SOME j => 
					  let
					      val E{mt,...} = IntMap.find(cells,j)
					  in
					      mt
					  end)
			   | _ => t)

		    fun sh f (nil, nil) = true
		      | sh f (MPR{least=l1,mt=t1,...} :: m1, 
			      MPR{least=l2,mt=t2,...} :: m2) =
			let
			    val compterm1 = termize(l1, t1, trans1, asgn1, c1)
			    val compterm2 = termize(l2, t2, trans2, asgn2, c2)
			in 
			    f(compterm1, compterm2) andalso sh f (m1,m2)
			end
		      | sh _ _ = false

		    fun shallowm (t1, t2) = sameHead(t1, t2)
		    fun deepm (MEMPTY, _) = true
		      | deepm (_, MEMPTY) = true
		      | deepm (MAPP{head=h1, args=a1,...},
			       MAPP{head=h2, args=a2,...}) =
			Symbol.sameSymbol(h1, h2) andalso 
			sh deepm (a1, a2)
		in
		    Symbol.sameSymbol(h1, h2) andalso 
		    sh (if deep then deepm else shallowm) (a1,a2)
		end
	in
	    fun quickMatch (t1, r1, t2, r2, s) = match false (t1, r1, s, t2, r2, s)
	    val deepMatch = match true
	end

	fun rename translation =
	    let
		fun rn (MEMPTY) = MEMPTY
		  | rn (MAPP{head, rigid, world, args}) =
		    MAPP{head=head, rigid=rigid, world=world, 
			 args = map rna args}

		and rna (MPR{least, vars, mt}) =
		    MPR{least=(case least of 
				   NONE => NONE
				 | SOME guy =>
				       case IntMap.peek(translation,guy) of
					   NONE => NONE
					 | SOME(VI(v,_)) => SOME v
					 | SOME(OI(v,_)) => SOME v
					 | SOME (TI(v,_,_)) => SOME v
					 | SOME (CI(v,_,_,_)) => SOME v),
		    vars = IntSet.empty,
		    mt = rn mt}
	    in
		rn
	    end

	fun subst translation =
	    let
		fun rn (MEMPTY) = MEMPTY
		  | rn (MAPP{head, rigid, world, args}) =
		    MAPP{head=head, rigid=rigid, world=world, 
			 args = map rna args}

		and rna (MPR{least, vars, mt}) =
		    (case least of 
			 NONE => 
			     MPR{least = NONE,
				 vars = IntSet.empty,
				 mt = rn mt}
		       | SOME guy =>
			     case IntMap.peek(translation,guy) of
				 SOME (TI(_,_,t)) => 
				     MPR{least = NONE,
					 vars = IntSet.empty,
					 mt = rn t}
			       | x => 
				     MPR{least = (case x of
						      SOME(VI(v,_)) => SOME v
						    | SOME(OI(v,_)) => SOME v
						    | SOME (CI(v,_,_,_)) => SOME v
						    | _ => NONE),
					 vars = IntSet.empty,
					 mt = rn mt})
	    in
		rn
	    end

	fun bindArg(MEMPTY, var, world) = IntMap.empty() : init IntMap.intmap
	  | bindArg(MAPP{args,...}, var, world) = bindArgMpr(args,var,world)

	and bindArgMpr(nil, var,world) = IntMap.empty() : init IntMap.intmap
	  | bindArgMpr(MPR{least,mt=t as MAPP _,...} :: _,var,world) =
	    IntMap.insert(IntMap.empty(),
			  var,
			  TI(case least of NONE => var | SOME i => i, world, t))
	  | bindArgMpr(MPR{least=SOME i,...} :: _, var,world) =
	    IntMap.insert(IntMap.empty(),
			  var,
			  VI(i, world))
	  | bindArgMpr _ = IntMap.empty()

	(*
	 * merge: term * term * problem -> term * problem
	 *
	 * Given two multiterms and a set of modal constraints,
	 * collapse the multiterms into a common one, 
	 * equating the modal indices of corresponding terms.
	 * (Cf. Martelli and Montanari, p 273)
	 *)
	fun merge(MEMPTY, m2, problem) = (m2, problem)
	  | merge(m1, MEMPTY, problem) = (m1, problem)
	  | merge (MAPP{head=h1, rigid=r1, world=SOME w1, args=a1}, 
		   MAPP{head=h2, rigid=r2, world=SOME w2, args=a2},
		   problem) =
	    if Symbol.sameSymbol(h1, h2) andalso r1 = r2 then
		case Patheq.equate(problem, w1, w2) of
		    NONE => 
			(raise Scope)
		  | SOME newproblem => 
			let val (newargs, finalproblem) = 
			    mergepairs(a1, a2, newproblem)
			in (MAPP{head=h1, rigid=r1,
				 world=SOME w1, args=newargs},
			    finalproblem)
			end
	    else
		raise Clash
	  | merge _ = raise Clash

	and mergepairs(nil, nil, problem) = (nil, problem)
	  | mergepairs(MPR{least=l1, vars=v1, mt=mt1} :: r1,
		       MPR{least=l2, vars=v2, mt=mt2} :: r2,
		       problem) =
	    let val (newterm, newproblem) = merge(mt1, mt2, problem)
		val (newleast, newvars) =
		    (case (l1, l2) of
			 (SOME l1, SOME l2) => 
			     if l1 < l2 then 
				 (SOME l1,
				  IntSet.union(v1,IntSet.add(v2, l2)))
			     else
				 (SOME l2,
				  IntSet.union(v2,IntSet.add(v1, l1)))
		       | (SOME l1, NONE) =>
			     (SOME l1, v1)
		       | (NONE, SOME l2) => 
			     (SOME l2, v2)
		       | (NONE, NONE) =>
			     (NONE, v1))
		val (newpairs, finalproblem) = mergepairs(r1, r2, newproblem)
	    in
		(MPR{least = newleast, 
		     vars = newvars, 
		     mt = newterm} :: newpairs,
		 finalproblem)
	    end
	  | mergepairs _ = raise Clash

	datatype readout =
	    ReadBlank
	  | ReadVar of int
	  | ReadTerm of term

	local
	    fun wv (n:int) = "X" ^ (Int.toString n)
	    fun wvl nil = ""
	      | wvl [i] = wv i
	      | wvl (h::r) = (wv h) ^ ", " ^ (wvl r)

	    fun wopen RIGID = "("
	      | wopen VARYING = "{"
	    fun wclose RIGID = ")"
	      | wclose VARYING = "}"

	    fun wt MEMPTY = "$"
	      | wt (MAPP{head, args, rigid,...}) =
		(Symbol.nameSymbol head) ^
		(if null args then "" else
		     (wopen rigid) ^
		     (wprs args) ^
		     (wclose rigid))

	    and wprs nil = ""
	      | wprs [one] = wpr one
	      | wprs (one::more) = 
		(wpr one) ^ ", " ^ (wprs more)

	    and wpr (MPR{least=NONE, mt,...}) = wt mt
	      | wpr (MPR{least=SOME v, mt, vars}) =
		if IntSet.isEmpty vars then
		    case mt of
			MEMPTY => wv v
		      | _ => (wv v) ^ ": " ^ (wt mt)
		else (wv v) ^ ": " ^ (wt mt)
	in
	    val writeTerm = wt
	    fun writeHead MEMPTY = "$"
	      | writeHead (MAPP{head,...}) = (Symbol.nameSymbol head)
	end

	fun writeReifiedTerm (wpl,starttop) =
	    let
		val dotop =
		    case starttop of 
			NONE => (fn _ => "") 
		      | SOME thing => thing

		fun wv (n:int) = "X" ^ (Int.toString n)
		fun wvl nil = ""
		  | wvl [i] = wv i
		  | wvl (h::r) = (wv h) ^ ", " ^ (wvl r)

		fun wtp (MEMPTY,_) = "$"
		  | wtp (MAPP{head, args, rigid, world}, top) =
		    let
			val noargs = ((null args) andalso (rigid = RIGID)
				      andalso (not top))
			val argstr = wprsp args
			val worldstr = (case (rigid, top, world) of
					    (_, true, world) => dotop world
					  | (_, _, NONE) => ""
					  | (RIGID, false, _) => ""
					  | (_, _, SOME w) => (wpl w))
		    in
			(Symbol.nameSymbol head) ^
			(if noargs then "" else
			    ("(" ^
			     argstr ^ 
			     (case (argstr, worldstr) of
				  ("", _) => ""
				| (_, "") => ""
				| _ => ", ")
				  ^ worldstr ^ ")"))
		    end

		and wprsp nil = ""
		  | wprsp [one] = wprp one
		  | wprsp (one::more) =
		    (wprp one) ^ ", " ^ (wprsp more)

		and wprp (MPR{least=NONE, mt,...}) = wtp (mt,false)
		  | wprp (MPR{least=SOME v, mt, vars}) =
		    if IntSet.isEmpty vars then
			case mt of
			    MEMPTY => wv v
			  | _ => (wv v) ^ ": " ^ (wtp (mt,false))
		    else (wv v) ^ ": " ^ (wtp (mt,false))
	    in
		fn x => wtp(x,case starttop of NONE => false | _ => true)
	    end

	fun printTerm out = out o writeTerm

	fun printTrans tr =
	    let
		fun printEntry (iv, VI(ov, _)) =
		    (print (Int.toString iv);
		     print " <- VI ";
		     print (Int.toString ov);
		     print "\n")
		  | printEntry (iv, OI(ov, _)) =
		    (print (Int.toString iv);
		     print " <- OI ";
		     print (Int.toString ov);
		     print "\n")
		  | printEntry (iv, CI(ov, t, _, _)) =
		    (print (Int.toString iv);
		     print " <- CI ";
		     print (Int.toString ov);
		     print ":";
		     printTerm print t;
		     print "\n")
		  | printEntry (iv, TI(ov, _, t)) =
		    (print (Int.toString iv);
		     print " <- TI ";
		     print (Int.toString ov);
		     print ":";
		     printTerm print t;
		     print "\n")
	    in
		IntMap.app printEntry tr
	    end

	(* extracting terms from bindings *)
	fun readout (reo, useworlds, s as S{bind = B{asgn, cells,...},...}) =
	    let
		fun wso (least, vars, OtherRole, mt, subs, varsseen, transopt) =
		    let
			val (newmt, newvars) = rt(mt, varsseen, transopt)
		    in
			(MPR{least=least,
			     vars=vars,
			     mt=newmt} :: subs,
			 newvars)
		    end
		  | wso (least, vars, Existential, mt, subs, varsseen, transopt) =
		    if reo then
			(MPR{least=least,
			     vars=vars,
			     mt=MEMPTY} :: subs,
			 varsseen)
		    else
			wso (least, vars, OtherRole, mt, subs, varsseen, transopt)

		and rt (MAPP{head,rigid,world,args}, varsout, transopt) =
		    let
			val (newargs, newvars) =
			    foldl (ra transopt)  (nil, varsout) args
		    in
			(MAPP{head=head,
			      rigid=rigid,
			      world=if useworlds then world else NONE,
			      args=rev newargs},
			 newvars)
		    end
		  | rt (MEMPTY, varsout, transopt) = (MEMPTY, varsout)

		and ra transopt (MPR{least=NONE, vars, mt}, (l,varsout)) =
		    wso(NONE, IntSet.empty, OtherRole, mt, l, varsout, transopt)

		  | ra NONE (MPR{least=SOME i, vars, mt}, (l, varsout)) =
		    (re (IntMap.find(cells,IntMap.find(asgn,i)), l, varsout)
		     handle IntMap.NotFound =>
			 (wso(SOME i, vars, OtherRole, mt, l, varsout, NONE)))
			
		  | ra (SOME trans) 
		    (MPR{least=SOME i, vars, mt}, (l, varsout)) =
		    let 
			val (newvar, switched) =
			    (case IntMap.peek(trans, i) of
				 NONE => (i, false)
			       | SOME (VI(j,_)) => (j, true)
			       | SOME (OI(j,_)) => (j, true)
			       | SOME (TI(j,_,_)) => (j, true)
			       | SOME (CI(j,_,_,_)) => (j, true))
		    in
			if switched then
			    ra NONE (MPR{least=SOME newvar, 
					 vars=vars, mt=mt}, (l, varsout))
			else
			    wso(SOME i, vars, OtherRole, mt, l, varsout, SOME trans)
		    end

		and re (E{vars=nil,mt,...}, l, varsout) =
		    wso(NONE, IntSet.empty, OtherRole, mt, l, varsout, NONE)

		  | re (E{vars=(t::_),keep=K(_,role),mt,...}, l, varsout) = 
		    if IntSet.member(varsout, t) then
			(MPR{least=SOME t,
			     vars=IntSet.empty,
			     mt=MEMPTY} :: l,
			 varsout)
		    else 
			(wso(SOME t, IntSet.empty, role, mt, l, 
			     IntSet.add(varsout, t), NONE))
			
	    in
		(fn (transopt,i) => #1 (rt(i,IntSet.empty,transopt)),
		 fn i => 
		 let 
		     val E{mt,keep=K(_,role),vars,...} = IntMap.find(cells, i)
		 in
		     case (reo,role) of
			 (True, Existential) => 
			     (case vars of 
				  nil => ReadTerm MEMPTY
				| t::_ => ReadVar t)
		       | _ => 
			     (case (#1 (rt(mt,IntSet.empty,NONE)),vars) of
				  (MEMPTY, t::_) => ReadVar t
				| (nt, _) => ReadTerm nt)
		 end handle IntMap.NotFound => ReadBlank)
	    end

	fun myreadoutVar (reo,useworld) (s as S{bind=B{asgn,...}, anon, ...}, NONE) =
	    (case IntMap.peek(asgn,anon) of
		 NONE => ReadBlank
	       | SOME i => #2 (readout (reo,useworld,s)) i)
	  | myreadoutVar (reo,useworld) (s as S{bind=B{asgn,...},...}, SOME j) =
	    (case IntMap.peek(asgn,j) of
		 NONE => ReadBlank
	       | SOME i => #2 (readout (reo,useworld,s)) i)

	fun readoutVar x = 
	    (case myreadoutVar (false,false) x of
		 ReadBlank => Miss
	       | ReadVar i => Free
	       | ReadTerm t => Bound t)

	fun writeprettyTerm system =
	    writeTerm o (#1 (readout (true, false, system)))

	fun writeprettyReifiedTerm (system, wpl, starttop) =
	    (writeReifiedTerm (wpl,starttop)) o (#1 (readout (true, true, system)))

	fun writeprettyVar system i =
	    case myreadoutVar (true,false) (system, SOME i) of
		ReadBlank => ("X" ^ (Int.toString i))
	      | ReadVar j => ("X" ^ (Int.toString j))
	      | ReadTerm t => (writeTerm t)

	fun writeprettyReifiedVar (system, wpl, starttop) i =
	    case myreadoutVar (true,true) (system, SOME i) of
		ReadBlank => ("X" ^ (Int.toString i))
	      | ReadVar j => ("X" ^ (Int.toString j))
	      | ReadTerm t => (writeReifiedTerm (wpl, starttop) t)

	(* printing *)
	fun printEquation (out, pt) (E{keep=K(_,role), vars, mt, world, ...}) =
	    let
		fun printVar (n:int) = (out ("X" ^ (Int.toString n)))
		fun pv nil = ()
		  | pv (h::r) = (printVar h;
				 out " = ";
				 pv r)

	    in
		pv vars;
		(case role of Existential => out "*!*" | _ => ());
		pt (NONE, mt);
		out "\n"
	    end

	fun printSystem out (s as S{bind=B{cells,incount,outcount,...},...}) =
	    let 
		val pt = (printTerm out) o (#1 (readout (false,false,s)))
		val pe = printEquation (out, pt)
	    in
		(IntMap.app (fn (i,j) => (pe j)) cells)
	    end

	fun checkGround (MEMPTY, tr1, s) =
	    (case getAnon s of
		 MEMPTY => false
	       | t1 => checkGround(t1, tr1, s))

	  | checkGround (m as MAPP{head=h1, args=a1,...}, trans, 
			 s as S{bind=B{asgn,cells,...},...}) =
	    let
		fun getv (NONE,guy) = (SOME guy,NONE)
		  | getv (SOME trans,guy) =
		    (case IntMap.peek (trans,guy) of
			 SOME (CI(v,_,_,_)) => (SOME v,NONE)
		       | SOME (OI(v,_)) => (SOME v,NONE)
		       | SOME (VI(v,_)) => (SOME v,NONE)
		       | SOME (TI (v,_,nt)) => (SOME v,SOME nt)
		       | NONE => (NONE,NONE))

		fun gr (nil,transopt,sofar) = SOME sofar
		  | gr (MPR{least=NONE, mt, vars} :: more, transopt, sofar) =
		    (case cg (mt, transopt, sofar) of
			 NONE => 
			     (NONE)
		       | SOME vars => 
			     (gr(more, transopt, vars)))

		  | gr (MPR{least=SOME v, mt, vars}::more, transopt, sofar) =
		    (case getv (transopt,v) of
			 (NONE,_) => 
			     if IntSet.member(sofar, v) then 
				 gr(more, transopt, sofar)
			     else
				 (case cg (mt, transopt, sofar) of
				      NONE => 
					  (NONE)
				    | SOME vars => 
					  (gr(more, transopt, IntSet.add(vars, v))))
		       | (SOME guy, thing)  => 
			     if IntSet.member(sofar, guy) then
				 gr(more, transopt, sofar)
			     else
				 let 
				     val ft = (case IntMap.peek(asgn, guy) of
						   NONE => (case thing of
								NONE => mt
							      | SOME nt => nt)
						 | SOME j => 
						       let
							   val E{mt,...} = IntMap.find(cells,j)
						       in
							   mt
						       end)
				 in
				     case cg (ft,NONE,sofar) of
					 NONE =>
					     (NONE)
				       | SOME vars => 
					     (gr(more, transopt, IntSet.add(vars,guy)))
				 end)

		and cg (MEMPTY, _, _) = NONE
		  | cg (MAPP{head, args,...},t,s) =
		    gr (args,t,s)
	    in
		case gr (a1,SOME trans,IntSet.empty) of NONE => false | _ => true
	    end

	(*
	 * Translation and addition of new unifications.
	 *)
	local
	    (* inner should be closer to the root than outer in problem *)
	    fun relate (problem, inner, outer) =
		(case Patheq.makeAncestor(problem, inner, outer) of
		     NONE => 
			 (raise Scope)
		   | SOME p => p)

	    (*
	     * addVar: int * pathelt * pathelt * 
	     *         (binding * int * problem * (int * int list) * term) ->
	     *         (binding * int * problem * (int * int list) * term) ->
	     *
	     * Called when a variable i is encountered, associated by initial
	     * translation with world WORLD, in a term whose index is 
	     * PARENTWORLD, in a context summarized by the long string
	     * of arguments.
	     *
	     * If necessary, adds a new equation to store variable i,
	     * makes a new index (closer to the root than WORLD).
	     *
	     * Then ups the number of times var is referenced, ensures
	     * that PARENTWORLD extends the modal index of the var,
	     * and makes record of the presence of the var on this
	     * term in the context.
	     *)
	    fun addVar (i, world, parentworld, status,
			(B{asgn, nextcell, cells, incount, outcount},
			 eqnct, problem, (least, vars), term)) =
		case IntMap.peek(asgn, i) of
		    NONE => 
			(case Patheq.newAncestor(problem, world) of
			     NONE => 
				 (raise Scope)
			   | SOME (newproblem, newworld) => 
				 (B{asgn = IntMap.insert(asgn, i, nextcell),
				    nextcell = nextcell + 1,
				    incount = increment(incount, nextcell),
				    outcount = outcount,
				    cells = IntMap.insert(cells, nextcell,
							  E{index=nextcell,
							    rank=0,
							    world=newworld,
							    keep=status,
							    nvars=1,
							    vars=[i],
							    mt= MEMPTY})},
				 eqnct+1,
				 relate(newproblem, newworld, parentworld),
				 (case least of 
				      NONE => (SOME i, vars) 
				    | SOME j => (SOME j, IntSet.add(vars,i))),
				 term))
		  | SOME j =>
			(B{asgn = asgn,
			   nextcell = nextcell,
			   incount = increment(incount, j),
			   outcount = outcount,
			   cells = cells},
			 eqnct,
			 (relate(problem, 
				 (fn E{world,...} => world) (IntMap.find(cells,j)),
				 parentworld) 
			  handle IntMap.NotFound => 
			      raise (Fail ("addVar (cterm): mismatch between " ^
					   "cells and asgn on " ^  (Int.toString j) ^
					   "."))),
			 (case least of 
			      NONE => (SOME i, vars) 
			    | SOME j => (SOME j, IntSet.add(vars,i))),
			 term)
			

	    (* 
	     * fix: init IntMap.intmap * Patheq.pathelt ->
	     *      term * Patheq.pathelt * binding * equation IntMap.intmap * 
	     *       Patheq.problem ->
	     *      binding * equation IntMap.intmap * Patheq.problem * term
	     *
	     * Given translation of variable and current world:
	     *    Take a term (whose parent is defined at pathelt) and
	     *    perform the appropriate translation on it (updating the
	     *    binding and equations if new free variables are 
	     *    introduced, and updating the problem to introduce the new
	     *    possible worlds at which subterms will be defined).
	     *)
	    fun fix (translation, currentworld) =
		let
		    (* 
		     * addSkfv: int * term * pathpt * int list * data -> data
		     *         make sure that we've added the needed variable
		     *         for this skolem function instance, and
		     *         return a reference to it.
		     *    much like the function addVar in code and behavior.
		     *)
		    fun addSkfv (name, head, world, args,
				 (b as B{asgn, nextcell, cells, incount, outcount},
				  eqnct, problem, (least,vars), term)) =
			case IntMap.peek(asgn, name) of
			    NONE => 
				let 
				    val (B{asgn, nextcell, cells, incount, outcount}, 
					 newe, p', (_, newvars), _) =
					foldr (translateAndAdd world) 
					(b, eqnct, problem, (SOME ~1, IntSet.empty), NONE) args
				    val goodterm = (MAPP{head = head,
							 world = SOME world,
							 rigid = VARYING,
							 args = (IntSet.fold
								 (fn (x,l) =>
								  (MPR{least = SOME x,
								       vars = IntSet.empty,
								       mt = MEMPTY} :: l))
								 newvars
								 nil)})

				    val (adjustedterm, newproblem) = 
					case term of
					    NONE => (goodterm, p')
					  | SOME oldterm =>
						merge(oldterm, goodterm, p')
				in
				    (B{asgn = IntMap.insert(asgn, name, nextcell),
				       nextcell = nextcell+1,
				       incount = increment(incount, nextcell),
				       outcount = outcount,
				       cells = IntMap.insert(cells, nextcell,
							     E{index = nextcell,
							       rank = 0,
							       world = world,
							       keep = K(Service,Existential),
							       nvars = 1,
							       vars = [name],
							       mt = adjustedterm})},
				    newe+1,
				    newproblem,
				    (case least of
					 NONE => (SOME name, vars)
				       | SOME j => (SOME j, IntSet.add(vars, name))), 
				     term)
				end
			  | SOME j => 
				(B{asgn = asgn,
				   nextcell = nextcell,
				   incount = increment(incount, j),
				   outcount = outcount,
				   cells = cells},
				 eqnct, problem, 
				 (case least of
				      NONE => (SOME name, vars)
				    | SOME v => (SOME v, IntSet.add(vars,name))), term)
				
		    (*
		     * translateAndAdd: pathelt -> int * data -> data
		     *    Build a term corresponding to variable <int>.  
		     *    If <int> is translated as a variable, we call
		     *    addVar to link in an appropriate reference to
		     *    the name of the variable.  If <int> is translated
		     *    as a constant with specified args, we recursively
		     *    call translateAndAdd on the args, and build a
		     *    varying function symbol.  (According to the 
		     *    syntax, we may have specified a term to go with
		     *    this variable, so we may have to merge those terms
		     *    together now...)   If <int> is not translated, 
		     *    treat it as a wide scope existential quantifier.
		     *)
		    and translateAndAdd parent (var, data as (b,e,problem,(l,v),term)) =
			(case IntMap.peek(translation, var) of
			     SOME (VI (new, world)) => 
				 (addVar (new, world, parent, K(Service,OtherRole), data))

			   | SOME (OI (old, oldworld)) => 
				 (case IntMap.peek(translation, old) of
				      SOME (VI (new, world)) =>
					  if new = var then
					      (addVar (new, world, parent, K(Service,OtherRole), data))
					  else
					      (addVar (old, oldworld, parent, K(Service,OtherRole), data))
				    | _ =>  addVar (old, oldworld, parent, K(Service,OtherRole), data))
						   
			   | SOME (CI (nameid, MAPP{head,...}, world, args)) =>
				 (addSkfv(nameid, head, world, args, data))

			   | SOME (TI (nameid, world, term)) => 
				 let
				     val (b', e', p', (l',v'), t) =
					 addVar (nameid, world,
						 parent, K(User,OtherRole), data)
				     val (newb, newe, newp, newt) =
					 updateAndCountTerm(term, parent, b', e', p')

				     val (finalterm, finalproblem) =
					 (case t of 
					      NONE => (newt, newp)
					    | SOME t' => merge(t', newt, newp))
				 in
				     (newb, newe, finalproblem, (l',v'), 
				      SOME finalterm)
				 end

			   | _ => 
				 (addVar (var, Patheq.rootProblem problem, 
					  parent, K(User,OtherRole), data)))

		    (*
		     * updateAndCount:
		     *   take a term and translate the vars in it recursively.
		     *   At EMPTY nodes, there's nothing to do;
		     *   At APP nodes, you must introduce a new path variable
		     *    indicating the node at which the function-application
		     *    is defined (its behavior depends on whether the symbol is
		     *    RIGID or VARYING) and you link this variable in with
		     *    the variable for its parent term.
		     *   For ARGUMENTS, you call translate and add on any 
		     *    variable, and call update and count recursively on
		     *    any term.
		     *)
		    and updateAndCountTerm (MEMPTY, ancinc, bind, eqs, problem) = 
			(bind, eqs, problem, MEMPTY)
		      | updateAndCountTerm (MAPP{head, rigid, world, args}, 
					    ancinc, bind, eqs, problem) =
			let 
			    val (newproblem, newworld) = 
				(case (rigid, args) of 
				     (VARYING, _) => (problem, currentworld)
				   | (RIGID, nil) => (problem, Patheq.rootProblem problem)
				   | (RIGID, _) => 
					 (case Patheq.newAncestor(problem, ancinc) of
					      NONE => 
						  (raise Scope) 
					    | SOME x => x))

			    val (newb, neweqs, finalproblem, newargs) = 
				updateAndCountPairs(args, newworld, bind, 
						    eqs, newproblem,  nil)
			in (newb,
			    neweqs,
			    finalproblem,
			    MAPP{head = head,
				 rigid = rigid,
				 world = SOME newworld,
				 args = newargs})
			end

		    and updateAndCountPairs (nil, world, bind, eqs, problem, args) =
			(bind, eqs, problem, rev args)
		      | updateAndCountPairs(MPR{least=NONE, vars, mt} :: more, 
					    world, bind, eqs, problem, sofar) =
			let 
			    val (newbind, neweqs, newproblem, newterm) =
				updateAndCountTerm(mt, world, bind, eqs, problem)
			in
			    updateAndCountPairs(more, world, newbind, neweqs, 
						newproblem, 
						MPR{least=NONE, vars=vars, mt=newterm} :: 
						sofar)
			end
		      | updateAndCountPairs(MPR{least=SOME i, vars, mt} :: more,
					    world, bind, eqs, problem, sofar) =
			let
			    val thing as (b, e, p, (newleast, newvars), newterm) =
				IntSet.fold (translateAndAdd world) vars
				(translateAndAdd world 
				 (i, (bind, eqs, problem, (NONE, IntSet.empty), NONE)))

			    val (newbind, neweqs, p'', newmt) =
				updateAndCountTerm(mt, world, b, e, p)

			    val (finalterm, finalproblem) =
				case newterm of  
				    NONE => (newmt, p'')
				  | SOME newguy => merge(newmt, newguy, p'')
			in
			    updateAndCountPairs(more, world, newbind, neweqs, finalproblem,
						MPR{least= newleast,
						    vars = newvars,
						    mt = finalterm} :: sofar)
			end
		in
		    updateAndCountTerm
		end

	    fun realizeUpdate(newCell, newIndex, newVar, 
			      B{asgn, nextcell, incount, outcount, cells},
			      newProblem, feqct, 
			      S{zeros, anon, anonterm, solved, ...}) =
		let
		    val newcells = IntMap.insert(cells, newIndex, newCell)
		    val newzeros = (IntMap.fold (fn (i,x,z) =>
						 if 0 = getCount(incount, i) then x::z else z)
				    newcells nil)
		in
		    S{bind = B{asgn = IntMap.insert(asgn, newVar, newIndex),
			       nextcell = nextcell + 1,
			       incount = incount,
			       outcount = outcount,
			       cells = newcells},
		      zeros = newzeros,
		      neq = feqct,
		      solved = solved,
		      anon = anon,
		      anonterm = anonterm,
		      problem = newProblem}
		end

	in 
	    (*
	     * updateSystem: 
	     *  Use the FIX function to translate t1 according to trans1 at w1
	     *   (and t2 according to trans2 at w2) and to update the counts of
	     *   occurrences of variables in the system.
	     *  Then add a dummy variable R and add the equation R = t1 = t2 
	     *   to the system;  recalculate the ZERO equations that are on
	     *   top of the partial order by occurrences -- and which can 
	     *   therefore be eliminated on the Martelli & Montanari scheme.
	     *)
	    fun myUpdateSystem _ (MEMPTY, trans1, w1,
				  t2, trans2, w2,
				  problem, s as S{bind=b as B{asgn,...}, zeros, neq,
						  solved, anon, anonterm, ...}) =
		(case (anonterm, IntMap.peek(asgn, anon)) of
		     (MAPP _, NONE) =>
			 myUpdateSystem (K(User,OtherRole))
			 (anonterm, trans1, w1,
			  t2, trans2, w2, problem, s)
		   | (MEMPTY, NONE) =>
			 let
			     val (newb as B{nextcell,...}, neweqct, p', nt2) =
				 fix (trans2, w2) (t2, w2, b, neq, problem)

			     val (newterm, newproblem) = merge(MEMPTY, nt2, p')
			 in
			     realizeUpdate(E{index = nextcell,
					     rank = 0,
					     nvars = 1,
					     keep = K(User,OtherRole),
					     world = w1,
					     vars = [anon],
					     mt = newterm},
					   nextcell, anon, newb, newproblem, neweqct+1, s)
			 end
		   | (_, SOME cell) => 
			 let
			     val (b as B{nextcell,cells,...}, neweqct, p', nt2) =
				 fix (trans2, w2) (t2, w2, b, neq, problem)

			     val E{index,rank,nvars,keep,world,vars,mt} =
				 IntMap.find(cells,cell)

			     val (newterm, newproblem) = merge(mt, nt2, p')
			 in
			     realizeUpdate(E{index=index, 
					     rank=rank,
					     nvars=nvars,
					     world=w1,
					     keep=keep,
					     vars=vars,
					     mt=newterm},
					   index, anon, b, newproblem, neweqct, s)
			 end)
		    
	      | myUpdateSystem keep (t1, trans1, w1, t2, trans2, w2, problem,
				     s as S{bind=b, zeros, neq, solved, anon, anonterm, ...}) =
		let 
		    val (b', e', p', nt1) = 
			fix (trans1, w1) (t1, w1, b, neq, problem)

		    val (b as B{nextcell, ...}, 
			 neweqct, p'', nt2) = 
			fix (trans2, w2) (t2, w2, b', e', p')
			
		    val r = newVar ()

		    val (newterm, newproblem) = merge(nt1, nt2, p'')
		in
		    realizeUpdate(E{index = nextcell,
				    rank = 0,
				    nvars = 1,
				    keep = keep,
				    world = w1,
				    vars = [r],
				    mt = newterm},
				  nextcell, r, b, newproblem, neweqct+1, s)
		end

	    val updateSystem = myUpdateSystem (K(Temporary,OtherRole)) 

	    fun updateVar (var, t2, trans2, w2, problem,
			   s as S{bind=b, zeros, neq, solved, anon, anonterm,...}) =
		let
		    val (b as B{asgn, cells, nextcell, ...}, neweqct, newp, nt2) =
			fix (trans2, w2) (t2, w2, b, neq, problem)
		in 
		    let
			val E{index, rank, keep, world, nvars, vars, mt} = 
			    IntMap.find(cells, IntMap.find(asgn, var))
			
			val (newterm, newproblem) = merge(mt, nt2, newp)
		    in
			realizeUpdate(E{index = index,
					rank = 0,
					nvars = nvars,
					keep = keep,
					world = world,
					vars = vars,
					mt = newterm},
				      index, var, b, newproblem, neweqct, s)
		    end
		handle IntMap.NotFound => 
		    realizeUpdate(E{index = nextcell,
				    rank = 0,
				    nvars = 1,
				    keep = K(User, OtherRole),
				    world = w2,
				    vars = [var],
				    mt = nt2},
				  nextcell, var, b, newp, neweqct+1, s)
		end
	end

	(*
	 * Make a new system with no equations.
	 *)
	fun urNewSystem(problem, term) =
	    S{bind = B{asgn=IntMap.empty(),
		       nextcell=0,
		       incount=C(IntMap.empty()),
		       outcount=C(IntMap.empty()),
		       cells=IntMap.empty()},
	      zeros = nil, 
	      neq = 0, 
	      solved = nil,
	      anon = newVar (),
	      anonterm = term,
	      problem = problem}

	fun newSystem(problem) =
	    S{bind = B{asgn=IntMap.empty(),
		       nextcell=0,
		       incount=C(IntMap.empty()),
		       outcount=C(IntMap.empty()),
		       cells=IntMap.empty()},
	      zeros = nil, 
	      neq = 0, 
	      solved = nil,
	      anon = newVar (),
	      anonterm = MEMPTY,
	      problem = problem}

	(*
	 * Make a system that consists entirely of the
	 * equation t1 = t2 -- e.g., for debugging purposes.
	 *)
	fun makeSystem(Miss, problem) = urNewSystem(problem, MEMPTY)
	  | makeSystem(Free, problem) = urNewSystem(problem, MEMPTY)
	  | makeSystem(Bound t1, problem) = urNewSystem(problem, t1)

	(*
	 * follows: system -> int * int -> bool
	 *    returns the relative order in which the two vars
	 *    were most recently removed from the system.
	 *    By the M&M invariant, if x follows y 
	 *    (was removed from the system later) then
	 *    y cannot appear in the value of x.
	 *)
	fun follows (S{bind = B{asgn,cells,...},...}) =
	    let
		fun rank i =
		    let
			val E{rank,mt,...} =
			    IntMap.find(cells,
					IntMap.find(asgn,i))
		    in
			(rank,mt)
		    end
	    in
		fn (x,y) => (let val (rx,mtx) = (rank x) 
				 val (ry,mty) = (rank y)
			     in
				 rx > ry
			     end				     
				 handle IntMap.NotFound => true)
	    end

	fun bestAlias (S{bind = B{asgn, cells,...},...}, x) =
	    let
		val E{vars,...} = IntMap.find(cells,
					      IntMap.find(asgn,x))
	    in
		case vars of nil => x | (t::_) => t
	    end handle IntMap.NotFound => x

	(*
	 * Determine whether two terms are identical,
	 * according to the bindings in the system.
	 *)
	fun matchTermsPB (problem, bind as B{asgn,cells,...}) =
	    let
		fun mt (MAPP{head=h1,rigid=r1,world=SOME w1,args=a1},
			MAPP{head=h2,rigid=r2,world=SOME w2,args=a2}) =
		    if (Symbol.sameSymbol(h1,h2) andalso
			r1 = r2 andalso
			Patheq.samePathpt(problem,w1,w2)) then
			ma(a1,a2)
		    else
			false
		  | mt _ = false

		and ma (nil, nil) = true
		  | ma (MPR{least=l1,mt=t1,...} :: r1,
			MPR{least=l2,mt=t2,...} :: r2) =
		    (case (l1,l2) of 
			 (NONE, NONE) => mt(t1,t2)
		       | (SOME v1, SOME v2) => 
			     IntMap.peek(asgn,v1) = IntMap.peek(asgn,v2)
		       | (SOME v1, NONE) =>
			     let
				 val E{mt=t1',...} = 
				     IntMap.find(cells,
						 IntMap.find(asgn,v1))
				     handle IntMap.NotFound =>
					 raise (Fail ("matchTerms (cterm) " ^
						      "v1 evil mismatch\n"))
			     in
				 mt(t1',t2)
			     end 
		       | (NONE, SOME v2) =>
			     let
				 val E{mt=t2',...} =
				     IntMap.find(cells,
						 IntMap.find(asgn,v2))
				     handle IntMap.NotFound =>
					 raise (Fail ("matchTerms (cterm) " ^
						      "v2 evil mismatch\n"))
			     in
				 mt(t1,t2')
			     end) andalso ma(r1,r2)
		  | ma _ = false
	    in
		mt
	    end

	fun matchTerms (S{bind,problem,...}) = matchTermsPB(problem,bind)

	fun matchVars (S{bind=B{asgn,cells,...},...}) (v1, v2) =
	    (case IntMap.peek(asgn,v1) of 
		 NONE => v1 = v2
	       | x => x = IntMap.peek(asgn,v2))

	local 
	    (*
	     * Reduce computes the common part and the frontier of
	     * a multiequation, as in Martelli & Montanari (pp 273-275).
	     *
	     * Here we have reduce: binding -> 
	     *     term * equation list * count -> term * equation list * count
	     *
	     * Input is a term, an initial frontier of equations, and an
	     * initial "outcount" of occurrences of variables in 
	     * equations posted as solved.  Output is the common part 
	     * of the term, the overall frontier of equations, and
	     * a modified outcount which adds counts for the occurrences of
	     * variables in the common part term.
	     *)
	    fun reduce asgn =
		let
		    fun red (MEMPTY, frontier, outcount) = 
			(MEMPTY, frontier, outcount)

		      | red (MAPP{head, rigid, world, args},
			     frontier, outcount) =
			let 
			    val (newargs, newfrontier, newoutcount) = 
				redpairs(args, nil, frontier, outcount) 
			in 
			    (MAPP{head = head, 
				  rigid = rigid,
				  world = world, 
				  args = newargs}, 
			     newfrontier,
			     newoutcount)
			end

		    and redpairs(nil, args, frontier, outcount) = 
			(rev args, frontier, outcount)

		      | redpairs(MPR{least=NONE, vars, mt} :: rest, 
				 args, frontier, outcount) =
			let 
			    val (newmt, newfrontier, newoutcount) = 
				red(mt, frontier, outcount)
			in 
			    redpairs(rest, 
				     MPR{least=NONE, vars=vars, mt=newmt} :: args,
				     newfrontier,
				     newoutcount)
			end
		      | redpairs((pair as MPR{least=SOME i, vars, mt}) :: rest, 
				 args, frontier, outcount) =
			redpairs(rest, 
				 MPR{least=SOME i, vars=IntSet.empty, mt=MEMPTY} :: args, 
				 pair :: frontier,
				 increment(outcount, IntMap.find(asgn, i)))
			handle IntMap.NotFound =>
			    raise (Fail ("redpairs (newterm): no cell for var " ^
					 (Int.toString i)))
		in
		    red
		end;

	    (*
	     * mergeEquations: equation * equation * bool * int * 
	     *                 Patheq.problem * binding ->
	     *        int * equation * int * Patheq.problem * binding
	     *
	     * Implementation of the MergeMultEq procedure in Martelli & Montanari's
	     * implementation of compact (p. 277).  
	     *
	     * <e1> and <e2> are the two equations to be merged:
	     * (bool indicates whether a reference to one equation is
	     *  being discharged here).
	     * Merging requires deciding which equation to move and which
	     * to save (the index of one we save is returned as the first <int>).
	     * It requires constructing a new representation of the equation
	     * (which is returned).  We update the count of unsolved equations
	     * (the <int> arg and second <int> result), we update the
	     * Patheq.problem to record equalities of domains of definitions
	     * of merged terms, and we update the binding to reflect accurate
	     * cell mappings for variables defined by the overall equations and 
	     * occurrence counts for variables mentioned in the merged term.
	     *)
	    fun mergeEquations(e1 as E{nvars=n1,...}, 
			       e2 as E{nvars=n2,...}, 
			       dischargeRef, remainingEqs,  problem, 
			       B{asgn, nextcell, incount, outcount, cells}) =
		let
		    (* update the association between vars and equations *)
		    fun reset (nil, index, asgn, final) = (asgn, final)
		      | reset (v1 :: more, index, asgn, final) =
			reset (more, index, 
			       IntMap.insert(asgn, v1, index), v1 :: final)

		    (* reset, merge, build new equation and binding *)
		    fun msm(e1 as E{index=i1, rank = r1, world = w1, keep=k1, nvars=n1, vars=v1, mt=t1},
			    e2 as E{index=i2, rank = r2, world = w2, keep=k2, nvars=n2, vars=v2, mt=t2})=
			let val (newasgn, newvars) = reset(v1, i2, asgn, v2)
			    val (newterm, newproblem) =
				merge(t1, t2, 
				      case Patheq.equate(problem, w1, w2) of
					  NONE => 
					      (raise Scope)
					| SOME p => p)
			    val newe = E{index = i2,
					 world = w1,
					 rank = r1,
					 nvars = n1+n2,
					 keep = better(k1,k2),
					 vars = newvars,
					 mt = newterm}
			in (i2,
			    newe,
			    remainingEqs - 1,
			    newproblem,
			    B{asgn = newasgn, 
			      nextcell = nextcell,
			      incount = ((if dischargeRef then 
					      mergeDecCounts 
					  else mergeCounts) (incount, i2, i1)),
			      outcount = mergeCounts(outcount, i2, i1),
			      cells = IntMap.insert(#1(IntMap.remove(cells, i1)),
						    i2, newe)})
			    handle IntMap.NotFound =>
				raise Fail ("msm (term.sml): cells has no "
					    ^ (Int.toString i1) ^ ".")
					   
			end
		in
		    (* pick the update that involves the least work *)
		    if n1 < n2 then msm(e1, e2) else msm(e2, e1)
		end;

	    (*
	     * Update the passed binding to indicate that equation
	     * index is referred to once fewer.
	     *)
	    fun dec(index, B{asgn, nextcell, incount, outcount, cells}) =
		B{asgn = asgn, 
		  nextcell = nextcell,
		  incount = decrement(incount, index),
		  outcount = outcount,
		  cells = cells}

	    (*
	     * Add term t1 into equation E, if possible,
	     * adding the modal constraint that the domain of definition 
	     * of t1 matches the domains of definition of E's other
	     * constituents, and recording that this equation is
	     * referred to once fewer.
	     *)
	    fun decm(E{index, rank, world = world, nvars, vars, mt, keep}, 
		     B{asgn, nextcell, incount, outcount, cells}, 
		     remainingEqs, problem, t1) =
		let
		    val (newterm, newproblem) = 
			case (mt, t1) of
			    (MEMPTY, MEMPTY) => (MEMPTY, problem)
			  | (MEMPTY, MAPP{world=SOME w2,...}) => 
				(t1, (case Patheq.equate(problem, world, w2) of
					  NONE => 
					      (raise Scope) 
					| SOME p => p))
			  | _ => merge(mt, t1, problem)

		    val newe = E{index = index,
				 rank = rank,
				 world = world,
				 keep = keep,
				 nvars = nvars,
				 vars = vars,
				 mt = newterm}
		in
		    (index,
		     newe,
		     remainingEqs,
		     newproblem,
		     B{asgn=asgn, nextcell=nextcell, 
		       incount = decrement(incount, index),
		       outcount = outcount, 
		       cells=IntMap.insert(cells, index, newe)})
		end;
		
	    (*
	     * compact: frontier * binding * equation list * int * Patheq.problem ->
	     *     binding * equation list * int * Patheq.problem
	     *
	     * Compact follows the implementation of compact from Martelli &
	     * Montanari (described on p 275, given on p 277): "performs 
	     * compactification by repeatedly merging multiequations".
	     *
	     * Compact applies to elements on the frontier: pairs of vars
	     * and terms.  First, the multiequation for the pair's key LEAST 
	     * var is retrieved, and it is merged with the pair's term T1
	     * using the DECM function.  Then the multiequations for the 
	     * remaining variables are merged by looping over the pair's VARS
	     * list and calling MERGE EQUATIONS.
	     *
	     * As the frontier is being removed from the system, some
	     * decrementing of incounts goes on.
	     *
	     * If the resulting new equation is now mentioned nowhere else,
	     * it is added to the zeros list, and then the rest of the
	     * frontier is considered.
	     *)
	    fun compact (MPR{least=SOME i, vars, mt=t1} :: rest, 
			 b as B{asgn, cells,...}, 
			 zeros, remainingEqs, topproblem) =
		(let 
		     fun one (var, (i2, e2, eqct, problem,
				    b as B{asgn, nextcell, incount, outcount, cells})) = 
			 let val i1 = 
			     IntMap.find(asgn, var) 
			     handle IntMap.NotFound =>
				 raise Fail ("one (term.sml): asgn did not " ^
					     "contain " ^ (Int.toString var) ^
					     ".")

			 in if i1 = i2 then (i2, e2, eqct, problem, dec(i2, b))
			    else mergeEquations(IntMap.find(cells, i1), e2, true,
						eqct, problem, b)
				handle IntMap.NotFound =>
				    raise Fail ("one (term.sml): cells did not "
						^ "contain " ^ (Int.toString i1) ^
						".")
			 end 

		     val (newindex, newe, newEqnCt, newproblem, newb as B{incount,...}) =
			 (IntSet.fold one vars 
			  (decm(IntMap.find(cells, IntMap.find(asgn, i)), 
				b, remainingEqs, topproblem, t1)))
			 handle IntMap.NotFound =>
			     raise Fail ("compact (term.sml): asgn did not " 
					 ^ "contain " ^ (Int.toString i) ^
					 "or cells the result.")
		 in
		     if 0 = getCount(incount, newindex) then
			 (compact(rest, newb, newe :: zeros, newEqnCt, newproblem))
		     else
			 compact(rest, newb, zeros, newEqnCt, newproblem)
		 end)
	      | compact (nil, b, zeros, ct, problem) = (b, zeros, ct, problem) 
	      | compact _ = raise Clash

	    (*
	     * dumpOccurs: binding -> unit
	     *   spew dirt when the occurs check is raised.
	     *)
	    fun dumpOccurs (B{asgn, nextcell, incount, outcount, cells}) =
		if (Debug.debugp "occurs") then
		    (print "Occurs check triggered!\n";
		     IntMap.app (fn (_,e as E{index,...}) =>
				 if 0 = getCount(incount, index) then ()
				 else (print ("Bad guy (" ^ 
					      (Int.toString (getCount(incount, index)))
					      ^ "):");
				       printEquation (print, 
						      fn (_,x) => 
						      printTerm print x) e)) cells)
		else ()
	in
	    (*
	     * unify is the main loop:
	     *    it repeatedly selects a free equation from the zeros list
	     *    and reduces it (to find the commonpart and a frontier of
	     *    new equations) then compacts the frontier to propagate
	     *    these equalities and marks the equation as solved (if
	     *    we care about the value of the variable referred to here).
	     *
	     * (Of course) follows the unify main loop given in Martelli &
	     * Montanari p. 275.
	     *)
	    fun unify (s) =
		let 
		    fun u (s as S{bind=bind as B{asgn, nextcell, incount, outcount, cells},
				  zeros, neq, anon, anonterm, solved, problem}, ct) =
			let
			    fun work (E{index, rank, nvars, vars, mt, 
					keep=keep as K(varrole,_), world}, z) = 
				let val _ = ()
				    val (common, frontier, newoutcount) = 
					reduce asgn (mt, nil, outcount)
				    val (newb, newzeros, neweqct, newproblem) = 
					compact(frontier, 
						B{asgn = asgn, nextcell = nextcell,
						  incount = incount, 
						  outcount= if varrole<>Temporary then 
						  newoutcount else outcount,
						  cells=cells}, 
						z, neq - 1, problem)

				    val newsolved =
					if varrole<>Temporary then
					    E{index = index, rank = ct, world = world, 
					      nvars = nvars, vars = vars, 
					      keep = keep, mt = common} :: solved
					else solved
				in 
				    S{bind = newb,
				      zeros = newzeros,
				      neq = neweqct,
				      problem = newproblem,
				      anon = anon,
				      anonterm = anonterm,
				      solved = newsolved}
				end
			    handle IntMap.NotFound =>
				raise Fail ("work (term.sml): cells has no "
					    ^ (Int.toString index) ^ ".")
			in
			    if neq = 0 then 
				(let
				     val (newcells, newzeros) =
					 foldr (fn (x as E{index,...},(c,z)) =>
					       (IntMap.insert(c, index, x),
						if 0 = getCount(outcount, index) then x::z else z))
					 (IntMap.empty(), nil)
					 solved
				 in
				     (S{bind= B{asgn= asgn, 
						nextcell= nextcell, 
						incount= outcount,
						outcount= C(IntMap.empty()),
						cells = newcells},
					zeros = newzeros, 
					neq = length solved,
					solved = nil,
					anon = anon,
					anonterm = anonterm,
					problem = problem},
				     problem)
				 end)
			    else (case zeros of 
				      nil => (dumpOccurs bind;
					      raise Cycle)
				    | (z1::rest) => u (work (z1, rest), ct+1))
			end
		in
		    u(s,0)
		end

	    local
		fun collapse (c,p,b, e1 as E{mt=t1,...}, 
			      (e2 as E{mt=t2,...}) :: more, sofar) =
		    if matchTermsPB (p,b) (t1,t2) then
			let val (_, e', c', p', b') = 
			    mergeEquations (e1, e2, false, c, p, b)
			in  (c', p', b', U.rappend sofar (e' :: more))
			end
		    else collapse(c,p,b, e1, more, e2 :: sofar)
		  | collapse (c,p,b,e,nil,sofar) = (c,p,b,U.rappend sofar [e])

		fun collapseloop(c,p,b, e1 :: more, done) =
		    let val (c',p',b',done') = collapse(c,p,b,e1,done,nil)
		    in collapseloop(c', p', b', more, done')
		    end
		  | collapseloop(c,p,b,nil,done) = (c,p,b,done)
	    in
		fun collapseSystem(S{bind=bind as B{cells,...}, 
				     zeros, anon, anonterm, neq, solved, problem}) =
		    let val start = 
			(ListMergeSort.sort
			 (fn (E{rank=r1,...}, E{rank=r2,...}) => r1 < r2)
			 (map #2 (IntMap.listItems(cells))))
			val (newct, newproblem, newbind as B{outcount,...}, neweqs) =
			    collapseloop(neq, problem, bind, start, nil)
			val newzeros =
			    foldr (fn (x as E{index,...},z) =>
				  if 0 = getCount(outcount,index) then x :: z else z)
			    nil neweqs
		    in
			(S{bind=newbind,
			   zeros=newzeros,
			   neq=newct,
			   solved=nil,
			   anon = anon,
			   anonterm = anonterm,
			   problem=newproblem},
			 problem)
		    end
	    end
	end

	val _ =
	    app Debug.register
	    [("occurs",
	      "Dump instances of occur check violation\n",
	      false)];
    end


