(*
 * Matthew Stone
 * Dialup 0.2 - 1/98
 * fdb.sml
 *
 * Implementation of uniform proof search for a fragment
 * of first order modal logic.
 *)

functor FDBFo (structure Cterm : TERM;
		structure Symbol : SYMBL) : FDB =
    struct

	structure Term = Cterm
	structure Patheq = Cterm.Patheq
	structure Modspec = Patheq.Modspec

	(*
	 * Abstract syntax for formulas:
	 *)

	datatype formula =
	    Atomic of Cterm.term
	  | Equals of Cterm.term * Cterm.term
	  | Box of Modspec.absmod * formula
	  | And of formula * formula
	  | Or of formula list
	  | Imp of formula * formula
	  | All of int * Cterm.term * formula
	  | Some of int * Cterm.term * formula

	(*
	 * readFormula: collection -> string list ->
	 *              string list * formula option
	 * Reads a formula according to the following
	 * syntactic convention:
	 *
	 * formula ::= disjunction 
         *           | disjunction '->' implication
	 * disjunction ::= conjunction 
         *           | conjunction ';' disjunction
	 * conjunction ::= guy
	 *           | guy ',' conjunction
	 * guy ::=     '*' var guy [forall]
	 *           | '?' var guy [exists]
	 *           | box guy [necessity]
	 *           | term [atomic formula]
	 *           | term = term [atomic equality]
	 *           | '(' formula ')'
	 * 
	 * Implementation is straightforward and uses
	 * functions to parse each of the constituents
	 * defined above.
	 *)

	fun readFormula (collection) =
	    let
		exception error
		fun readGuy (nil, _) = raise error
		  | readGuy ("(" :: stuff, boxes) =
		    (case readImps stuff of
			 (f, ")"::rest) =>
			     (foldl Box f boxes, rest)
		       | _ => raise error)
		  | readGuy ("*" :: name :: stuff, boxes) =
		    let val variable = Cterm.getVar(name)
			val skolemf = Cterm.newConst ()
			val (next, more) = readGuy(stuff, nil) 
		    in
			(foldl Box (All (variable, skolemf, next)) boxes,
			 more)
		    end
		  | readGuy ("?" :: name :: stuff, boxes) =
		    let
			val variable = Cterm.getVar(name)
			val skolemf = Cterm.newConst ()
			val (next, more) = readGuy(stuff, nil)
		    in
			(foldl Box (Some (variable, skolemf, next)) boxes,
			 more)
		    end
		  | readGuy (name  :: more, boxes) = 
		    (case Modspec.lookup(collection, name) of
			 NONE => 
			     (case Cterm.readTerm(name::more) of
				  (NONE,_) => raise error
				| (SOME term, "=" :: further) =>
				      (case Cterm.readTerm(further) of
					   (NONE,_) => raise error
					 | (SOME term2, final) =>
					       (foldl Box (Equals (term, term2)) boxes,
						final))
				| (SOME term, final) => 
				      (foldl Box 
				       (Atomic term) boxes,
				       final))
		       | SOME i => readGuy (more, i :: boxes))

		and readConjuncts stuff =
		    case readGuy (stuff, nil) of
			(f, ","::more) => 
			    let val (g, rest) = readConjuncts more
			    in (And(f, g), rest)
			    end
		      | otherwise => otherwise

		and readDisjuncts (stuff, sofar) =
		    case readConjuncts stuff of
			(f, ";"::more) => 
			    readDisjuncts (more, f::sofar)
		      | (f, otherwise) => 
			    (case sofar of 
				 nil => (f, otherwise)
			       | l => (Or(rev (f :: l)), otherwise))

		and readImps stuff =
		    case readDisjuncts (stuff, nil) of
			(f, "->" :: more) =>
			    let val (g, rest) = readImps more
			    in (Imp(f,g), rest)
			    end
		      | otherwise => otherwise
	    in
		fn x => (let val (r, s) = readImps x
			 in (SOME r, s)
			 end handle error => (NONE, x))
	    end

	fun readAtom strl =
	    let
		val (t, r) = Cterm.readTerm strl
	    in
		case t of
		    NONE => (NONE, strl)
		  | SOME guy => (SOME (Atomic guy), r)
	    end

	fun breakFormulas formulas =
	    let
		fun bf (nil, sofar) = rev sofar
		  | bf (And (p, q) :: more, sofar) =
		    bf (more, p :: q :: sofar)
		  | bf (p :: more, sofar) =
		    bf (more, p :: sofar)
	    in
		bf (formulas, nil)
	    end

	local
	    (* box,quantifiers = 0, and = 1, or = 2, imp = 3 *)
	    fun inparens(i:int,j,s) =
		if i < j then 
		    "(" ^ s ^ ")" 
		else s

	    fun wfwt wt =
		let
		    fun wf (Atomic t, _) = wt t
		      | wf (Equals (t1,t2), _) =
			(wt t1) ^ " = " ^ (wt t2)
		      | wf (Some (v,_,p), i) =
			inparens(i, 0, "?X" ^ (Int.toString v) ^ " " ^ wf(p,0))
		      | wf (All (v,_,p), i) =
			inparens(i, 0, "*X" ^ (Int.toString v) ^ " " ^ wf(p,0))
		      | wf (Box (m, p), i) =
			inparens(i, 0, (Modspec.writeModality m) ^ " " ^ wf(p,0))
		      | wf (And (p, q), i) =
			inparens(i, 1, wf(p,1) ^ ", " ^ wf(q,1))
		      | wf (Or nil, i) = ""
		      | wf (Or [f], i) = wf(f, 2)
		      | wf (Or (h :: t :: fl), i) =
			inparens(i, 2, wf(h,2) ^ " ; " ^ wf(Or(t::fl),2))
		      | wf (Imp (p, q), i) =
			inparens(i, 3, wf(p,3) ^ " -> " ^ wf(q,3))
		in
		    wf
		end
	in
	    fun writeFormula f = wfwt (Cterm.writeTerm) (f, 4)
	    fun printFormula out = out o writeFormula 

	    fun prettyprintFormula out system (tr, f) =
		out (wfwt (fn t => Cterm.writeprettyTerm system (SOME tr, t)) (f, 4))
	end

	local
	    (* box,quantifiers = 0, and = 1, or = 2, imp = 3 *)
	    fun inparens(i:int,j,s) =
		if i < j then 
		    "(" ^ s ^ ")" 
		else s

	    fun wps nil _ = ""
	      | wps [x] _ = x
	      | wps (x::r) u = x ^ ", " ^ (wps r u)

	    fun update (m, params, vname) =
		let
		    val depth = Modspec.depth m
		    val oldv = List.nth (params, depth)
		    val mname = Modspec.writeModality m
		    fun replace (i, k, first::params, vname) =
			if i = k then
			    vname :: params
			else 
			    first :: (replace(i+1,k,params,vname))
		      | replace _ = nil
		in
		    (mname ^ "(" ^ oldv ^ ", " ^ vname ^ ")",
		     replace (0, depth, params, vname))
		end

	    fun wfwt wt =
		let
		    fun wf (Atomic t, params, _) = wt (SOME (wps params)) t
		      | wf (Equals (t1,t2), params, _) =
			(wt (SOME (wps params)) t1) ^ " = " ^ 
			(wt (SOME (wps params)) t2)
		      | wf (Some (v,_,p), params, i) =
			inparens(i, 0, "?X" ^ (Int.toString v) ^ " " ^ 
				 wf(p,params,0))
		      | wf (All (v,_,p), params, i) =
			inparens(i, 0, "*X" ^ (Int.toString v) ^ " " ^ 
				 wf(p,params,0))
		      | wf (Box (m, p), params, i) =
			let 
			    val v = Cterm.newVar()
			    val vname = "*X" ^ (Int.toString v)
			    val (antecedent, newparams) = 
				update(m, params, vname)
			in
			    inparens(i, 3, vname ^ " " ^ "(" ^
				     antecedent ^ " -> " ^ 
				     wf(p, newparams, 3) ^ ")")
			end
		      | wf (And (p, q), params, i) =
			inparens(i, 1, wf(p,params,1) ^ ", " ^ wf(q,params,1))
		      | wf (Or nil, params, i) = ""
		      | wf (Or [f], params, i) = wf(f, params, 2)
		      | wf (Or (h :: t :: fl), params, i) =
			inparens(i, 2, wf(h,params, 2) ^ " ; " ^ wf(Or(t::fl),params,2))
		      | wf (Imp (p, q), params, i) =
			inparens(i, 3, wf(p,params,3) ^ " -> " ^ wf(q,params,3))
		in
		    wf
		end
	in
	    fun prettyprintReifiedFormula out (system, problem) (tr, f, world) =
		out (wfwt (fn f => 
			   fn t => (Cterm.writeprettyReifiedTerm 
				    (system, Cterm.Patheq.writeprettyPathpt(problem,false,true), f)
				    (SOME tr, t)))
		     (f, Cterm.Patheq.splitWriteprettyPathpt(problem,false) world, 4))
	end

	fun varFreeIn var =
	    let
		fun vfi (Atomic term) =
		    Term.IntSet.member(Term.freeVarsTerm(term, Term.IntSet.empty), 
				       var)
		  | vfi (Equals (t1, t2)) =
		    Term.IntSet.member(Term.freeVarsTerm(t1, Term.IntSet.empty), var)
		    orelse
		    Term.IntSet.member(Term.freeVarsTerm(t2, Term.IntSet.empty), var)
		  | vfi (Box (_, f)) = vfi f
		  | vfi (And (f1, f2)) = vfi f1 orelse vfi f2
		  | vfi (Or list) = List.exists vfi list
		  | vfi (Imp (f1, f2)) = vfi f1 orelse vfi f2
		  | vfi (All (i, _, f)) =
		    if var = i then false
		    else vfi f
		  | vfi (Some (i, _, f)) =
		    if var = i then false
		    else vfi f
	    in
		vfi
	    end
	    
	fun freeVars f =
	    let
		fun fvc (Atomic term, i) = Term.freeVarsTerm(term, i)
		  | fvc (Equals (t1, t2), i) =
		    Term.freeVarsTerm(t1, Term.freeVarsTerm(t2, i))
		  | fvc (Box (_, f),i) = fvc(f,i)
		  | fvc (And (f1, f2), i) = fvc(f1, fvc(f2, i))
		  | fvc (Or list, i) = foldr fvc i list
		  | fvc (Imp (f1, f2), i) = fvc(f1, fvc(f2, i))
		  | fvc (All (x, _, f), i) =
		    let val i' = fvc(f,i) in
			Term.IntSet.delete(i', x)
			handle Term.IntSet.NotFound => i'
		    end
		  | fvc (Some (x, _, f), i) =
		    let val i' = fvc(f,i) in
			Term.IntSet.delete(i', x)
			handle Term.IntSet.NotFound => i'
		    end
	    in
		fvc (f, Term.IntSet.empty)
	    end

	local
	    fun close f = 
		Term.IntSet.fold 
		(fn (i,f') => Some(i,Cterm.newConst (),f')) 
		(freeVars f)
		f

	in
	    fun compose start =
		let
		    fun mi (Atomic c) = start
		      | mi (Box(m,f)) = Box(m, mi f)
		      | mi (Or fl) = Or (map mi fl)
		      | mi (And (p,q)) = And (mi p, mi q)
		      | mi (Imp (p,q)) = Imp(p, mi q)
		      | mi (All (x,t,f)) = All(x, t, mi f)
		      | mi (Some (x,t,f)) = Some(x, t, mi f)
		      | mi (Equals e) = Equals e
		in
		    mi
		end 

	    fun mkClosedImplic (nil, c) = close c 
	      | mkClosedImplic (a :: r, c) = 
		Imp(close (foldl And a r), close c)

	    fun mkImplic (nil, c) = c 
	      | mkImplic (a :: r, c) = 
		Imp(foldl And a r, c)

	    (*
	     * L, Start[], Assume[], Prove[] -> 
	     *       Start[Prove[Assume[&L] -> Prove[]]]
	     *)
	    fun substImplic (nil, start, assume, prove) = 
		compose prove start

	      | substImplic (a::r, start, assume, prove) =
		let
		    val conj = foldl And a r
		in
		    compose (compose (Imp (compose conj assume, 
					   prove))
			 prove) 
		    start
		end
	    
	    val emptyCxt = Atomic (Cterm.metaVar)
	end

	(***********************************************************
	 *
	 * PROOF SEARCH
	 *   DATA STRUCTURES
	 *
	 ***********************************************************)

	(*
	 * translation:
	 *   describes how bound variables are to be replaced 
	 *   with logic variables or herbrand constants when
	 *   we reach atomic formulas.
	 *)
	type translation = Cterm.init Cterm.IntMap.intmap

	fun rename (trans, world) =
	    let
		fun rn (trans, Atomic term) = (Atomic (Term.rename trans term))
		  | rn (trans, Equals (t1,t2)) = (Equals 
						  (Term.rename trans t1, 
						   Term.rename trans t2))
		  | rn (trans, Box (i, f)) = (Box (i, rn (trans, f)))
		  | rn (trans, And (f1, f2)) = (And (rn (trans, f1), rn (trans, f2)))
		  | rn (trans, Or list) = (Or (map (fn x => rn(trans, x)) list))
		  | rn (trans, Imp (f1, f2)) = (Imp (rn (trans, f1), rn (trans, f2)))
		  | rn (trans, All (i, s, f)) = 
			(All (i, s,
			      rn (Term.IntMap.insert(trans, i, Term.VI(i,world)), f)))
		  | rn (trans, Some (i, s, f)) =
			(Some (i, s,
				 rn (Term.IntMap.insert(trans, i, Term.VI(i,world)), f)))
	    in
		fn y => rn (trans, y)
	    end

	fun subst (trans,world) =
	    let
		fun rn (trans, Atomic term) = (Atomic (Term.subst trans term))
		  | rn (trans, Equals (t1,t2)) = (Equals 
						  (Term.subst trans t1, 
						   Term.subst trans t2))
		  | rn (trans, Box (i, f)) = (Box (i, rn (trans, f)))
		  | rn (trans, And (f1, f2)) = (And (rn (trans, f1), rn (trans, f2)))
		  | rn (trans, Or list) = (Or (map (fn x => rn(trans, x)) list))
		  | rn (trans, Imp (f1, f2)) = (Imp (rn (trans, f1), rn (trans, f2)))
		  | rn (trans, All (i, s, f)) = 
			(All (i, s,
			      rn (Term.IntMap.insert(trans, i, Term.VI(i,world)), f)))
		  | rn (trans, Some (i, s, f)) =
			(Some (i, s,
				 rn (Term.IntMap.insert(trans, i, Term.VI(i,world)), f)))
	    in
		fn y => rn (trans, y)
	    end

	(*
	 *  TRANSLATION: how bound variables must be replaced
	 *  SKOLEMARGS: list of logic variables introduced in
	 *    this scope on which herbrand functions will depend
	 *  WORLD: the modal index at which GOAL must be proven
	 *)

	datatype formuse =
	    FR of {formula: formula,
		   translation: translation,
		   skolemargs: int list,
		   world: Patheq.pathpt}

	(*
	 * right:
	 *    describes a goal formula -- a formula on the right
	 *    of a sequent.  With fields:
	 *  GOAL: formula that must be proved
	 *  ASSUME: local clauses (in addition to global program
	 *    clauses) introduced in proving GOAL
	 *  ANCESTORS: a list of modal indices on which WORLD
	 *    depends (and to which we can rule out the 
	 *    possibility of restarts).
	 *  HISTORY: a stack of previous goals that may be 
	 *    returned to in case of restarts.
	 *)
	datatype right =
	    R of {goal: formuse,
		  assume: left list,
		  ancestors: Patheq.pathpt list,
		  history: history list}
	(*
	 * left:
	 *   describes a program formula -- a formula on the
	 *   left of a sequent.  With fields:
	 *  HEAD: atomic term that the formula can be used
	 *   to prove ((a & b) is copied and listed separately
	 *   under a and b.
	 *  RULE: the full formula
	 *  ID: code identifying the assumption (for purposes
	 *   of recording cancellations -- when assumptions are
	 *   used within a block).
	 *  OHID: short for 'order of head in disjunct'
	 *   list of integer variables indicating the position 
	 *   of the head in (nested) original disjunctions
	 *)
	and left =
	    L of {head: Cterm.term,
		  rule: formuse,
		  ancestors: Patheq.pathpt list,
		  id: int,
		  disjunct: bool,
		  ohid: int list}

	(*
	 * history:
	 *   describes a possible goal to which proof search
	 *   could be restarted.
	 * GOAL: the goal to restart
	 * TRANSLATION, ASSUME, WORLD, ANCESTORS: as with right
	 * ID: an identifier for the goal
	 *)
	and history =
	    H of {goal: formula,
		  translation: translation,
		  assume: left list,
		  world: Patheq.pathpt,
		  ancestors: Patheq.pathpt list,
		  id: int}
	    
	(*
	 * restart:
	 *   describes the need to restart proof search because
	 *   of the use of a disjunctive premise.
	 * DISJUNCT: if we match (a or b), a matches the current
	 *   goal, and b is stored as the DISJUNCT, which must
	 *   be assumed for the course of the restart block.
	 * TRANSLATION, SKOLEMARGS, ANCESTORS, HISTORY:
	 *   as wtih right
	 * INDEX: the possible world at which DISJUNCT is to be
	 *   assumed.
	 * ID: assumption from which the disjunct was derived.
	 * SWAPFORBID: whether former mustuse becomes
	 *    forbidden in this restart.
	 * DEPTH: nesting of this restart in other restarts.
	 *)
	datatype restart =
	    RS of {disjunct: formula,
		   translation: translation,
		   skolemargs: int list,
		   index: Patheq.pathpt,
		   ancestors: Patheq.pathpt list,
		   history: history list,
		   id: int,
		   depth: int,
		   swapforbid: bool}

	(*
	 * state:
	 *   describes a particular point in proof search.
	 * GOALS: gives the right formulas that must be 
	 *   established as part of the current block
	 * RESTARTS: gives the restarts that must be 
	 *   undertaken in subsequent blocks.
	 * UNI: gives the substitution of terms for 
	 *   variables generated by the proof so far.
	 * PROBLEM: represents the constraints on the
	 *   scopes (or possible worlds model) for the
	 *   proof so far.
	 * MUSTUSE: possible single disjunct id we're working on 
	 *   canceling. 
	 * SOURCE: id of disjunctive clause from which 
	 *   mustuse is derived.
	 * FORBIDDEN: possible single disjunct id we can't cancel
	 *   now (if we're to keep answers different).
	 * CONTRIBUTIONS: restart goals that all uses of
	 *   mustuse contribute to.
	 * DEPTH: depth of restarts to reach current
	 *   goals
	 * MAXDEPTH: when do you give up?
	 * DATA: user-defined info (for abduction, etc.)
	 *)
	datatype 'a state = 
	    S of {goals: right list,
		  restarts: restart list,
		  uni: Cterm.system,
		  problem: Patheq.problem,
		  mustuse: int option,
		  source: int option,
		  forbidden: int option,
		  contributions: history list option,
		  maxworld: Patheq.pathpt,
		  assumptions: (formuse * Patheq.pathpt) list,
		  depth: int,
		  maxdepth: int,
		  data: 'a}

	datatype 'a proof =
	    PF of {uni: Cterm.system, 
		   assumptions: (formuse * Patheq.pathpt) list,
		   problem: Patheq.problem, 
		   data:'a}

	datatype 'a more =
	    ALTS of 'a state list

	(***********************************************************
	 *
	 * PROOF SEARCH
	 *   FUNCTIONS
	 *
	 ***********************************************************)

	(*
	 * two debugging helper functions:
	 *   printBack: to describe a backchaining attempt
	 *   printAfter: to describe the result of a 
	 *     successful backchaining
	 *)

	fun printBack(g, h, left, right) =
	    (if Debug.debugp "match" then
		 (print "Matching ";
		  Cterm.printTerm (print) g;
		  print " at ";
		  Patheq.printPathpt (print) right;
		  print "against ";
		  Cterm.printTerm (print) h;
		  print " at ";
		  Patheq.printPathpt (print) left;
		  print "\n")
	     else ();
	     if Debug.debugp "trace" then
		 (Debug.getline();
		  ())
	     else
		 ())

	fun printAfter(system, problem) =
	    (if Debug.debugp "spew" then
		 (print "Modality:\n";
		  Patheq.printProblem (print) problem;
		  print "Unification ";
		  Cterm.printSystem (print) system)
	     else
		 ();
	     if Debug.debugp "trace" then
		 (Debug.getline();
		  ())
	     else
		 ())

	fun intersectHistories(nil, h) = h
	  | intersectHistories(h, nil) = h
	  | intersectHistories((h1 as H{id,...}) :: r1, (h2 as H{id=id2,...}) :: r2) =
	    if id = id2 then
		h1 :: intersectHistories(r1, r2)
	    else 
		if id < id2 then
		    intersectHistories(h1 :: r1, r2)
		else
		    intersectHistories(r1, h2 :: r2)

	(*
	 * backchain:
	 *    (hopefully) straightforward implementation of 
	 *    goal-directed application of left rules on
	 *    sequents.
	 * Each left formula that possibly matches is an
	 *    argument to the function TRIGGER, which applies
	 *    all necessary left sequent rules to expose
	 *    the 'head' of the left formula for matching 
	 *    against the current goal.
	 * The different cases of the LEFT unwrap the 
	 *    different connectives.
	 *)
	fun backchain (R{goal=FR{formula=Atomic rightatom, 
				 translation = righttrans, 
				 skolemargs = rightskargs, 
				 world = rightworld}, 
			 assume, ancestors, history},
		       S{goals, restarts, uni, problem, depth, maxdepth,
			 mustuse, source, maxworld,
			 forbidden, contributions, assumptions,
			 data},
		       othergoals, newmaxworld, alternatives, axioms) =
	    let
		(* 
		 * see if the modal and first-order data
		 * of an atomic formula can be unified 
		 * with G.  If so, ste up a new state;
		 * If not (optionally) explain why.
		 *)
		fun axiom(FR{formula=Atomic leftatom, 
			     translation=lefttrans,
			     skolemargs = leftskargs, 
			     world = leftworld},
			  id, newgoals, newrestarts, newassumptions) =
		    (printBack(rightatom, leftatom, leftworld, rightworld);
		     case Patheq.equate(problem, leftworld, rightworld) of
			 NONE => NONE
		       | SOME p' => 
			     (let
				  val temprightworld = Patheq.keepCopy rightworld
				  val (newsystem, uproblem)  =
				      Cterm.unify
				      (Cterm.updateSystem
				       (rightatom, righttrans, 
					temprightworld,
					leftatom, lefttrans,
					Patheq.keepCopy leftworld,
					p', uni))
				  val final = Patheq.updateVtables uproblem
				  val newcontributions =
				      case mustuse of
					  NONE => NONE
					| SOME guy => 
					      if id = guy then
						  (case contributions of 
						       NONE => SOME history
						     | SOME guys =>
							   SOME (intersectHistories(guys,
										    history)))
					      else
						  contributions
			      in
				  case final of 
				      NONE => 
					  (print "Modal scope violation.\n";
					   NONE)
				    | SOME newproblem => 
					  (printAfter(newsystem, newproblem);
					   SOME(S{goals = (rev newgoals) @ othergoals,
						  restarts = newrestarts,
						  problem = newproblem,
						  uni = newsystem,
						  forbidden = forbidden,
						  mustuse = mustuse,
						  source = source,
						  depth = depth,
						  maxdepth = maxdepth,
						  maxworld = newmaxworld,
						  contributions = newcontributions,
						  assumptions = 
						  (foldr (fn (x,l) => (x,maxworld)::l)
						   assumptions newassumptions),
						  data = data}))
			      end
			  handle Cterm.Clash =>
			      (if Debug.debugp "match" then
				   print "Unification clashed.\n"
			       else ();
				   NONE)
			       | Cterm.Cycle =>
				     (if Debug.debugp "match" then
					  print "Occurs check violated.\n"
				      else ();
					  NONE)
			       | Cterm.Scope =>
				     (if Debug.debugp "match" then
					  print "No solution for scope constraints.\n"
				      else ();
					  NONE)))
		  | axiom _ = raise Match

		fun trigger (formuse, id, leftancestors, ohid, newassumptions) =
		    let
			fun left (f as FR{formula=Atomic h,...}, 
				  newgoals, newrestarts, newassumptions, _) =
			    (axiom (f, id, newgoals, newrestarts, newassumptions))

			  (* restrictions on the language *)
			  | left (f as FR{formula=Equals _,...}, _, _, _, _) = raise Match
			  | left (f as FR{formula=Or nil,...}, _, _, _, _) = raise Match
			  | left (f as FR{formula=Or _,...}, _, _, _, nil) = raise Match

			  (* We've rewritten the formula: we know And(p,q) -> p *)
			  | left (FR{formula=And (p,q), translation, skolemargs, world},
				  newgoals, newrestarts, newassumptions, ohid) =
			    left (FR{formula=p, translation=translation, 
				     skolemargs=skolemargs, world=world},
				  newgoals, newrestarts, newassumptions, ohid)
			    
			  (* For Or(p::r) match p and setup restarts for r cases *)
			  | left (FR{formula=Or (p::r), translation, skolemargs, world}, 
				  newgoals, newrestarts, newassumptions, t :: ohid) =
			    let
				val nfr = FR{formula=p, translation=translation, 
					     skolemargs = skolemargs, world = world}
			    in
				left (nfr,
				      newgoals,
				      #2 (foldr (fn (q,(ctt,l)) => 
						(ctt-1,
						 RS{disjunct=q, 
						    index=Patheq.keepCopy(world),
						    translation = translation, 
						    skolemargs = skolemargs,
						    ancestors=leftancestors, 
						    history=history, 
						    depth = depth+1,
						    id=id,
						    swapforbid = ctt > 0} :: l)) 
				      (t,newrestarts) r),
				      nfr :: newassumptions,
				      ohid)
			    end

			  (* For Box(i,p) extend the path and look at p *)
			  | left (FR{formula=Box (i,p), translation, skolemargs, world},
				  newgoals, newrestarts, newassumptions, ohid) =
			    let
				val (newpath, _) = 
				    Patheq.extendVar(problem, world, world, i)
			    in 
				left (FR{formula=p, translation=translation,
					 skolemargs=skolemargs, world = newpath},
				      newgoals, newrestarts, newassumptions, ohid)
			    end

			  (* For Imp(p,q) follow q and make a goal of p *)
			  | left (FR{formula=Imp(p,q), translation, skolemargs, world},
				  newgoals, newrestarts, newassumptions, ohid) =
			    left(FR{formula=q, translation=translation, 
				    skolemargs=skolemargs, world=world},
				 R{goal=FR{formula=p, 
					   world=Patheq.keepCopy(world), 
					   translation=translation, 
					   skolemargs=skolemargs}, 
				   assume=assume, 
				   ancestors=leftancestors, history=history} :: 
				 newgoals, newrestarts, newassumptions, ohid)

			  (* For All(i,_,p) translate xi to a new logic var *)
			  | left (FR{formula=All(i, _, p), translation, skolemargs, world},
				  newgoals,  newrestarts, newassumptions, ohid) =
			    let 
				val var = Cterm.newVar()
				val lwc = Patheq.keepCopy(world)
			    in 
				left(FR{formula=p, 
					world = world,
					skolemargs = var :: skolemargs, 
					translation = Cterm.IntMap.insert
					(Cterm.IntMap.insert (translation, i, 
							      Cterm.VI(var, lwc)),
					 var, Cterm.OI(i, lwc))},
				newgoals, newrestarts, newassumptions, ohid)
			    end

			  (* For Some(i,f,p) translate xi to f(...) and do p *)
			  | left (FR{formula=Some(i, f, p), translation, skolemargs, world},
				  newgoals, newrestarts, newassumptions, ohid) =
			    left(FR{formula=p, 
				    skolemargs= skolemargs,
				    world = world,
				    translation=Cterm.IntMap.insert
				    (translation, i, Cterm.CI(Cterm.newVar(), f, 
							    Patheq.keepCopy(world),
							    skolemargs))},
			    newgoals, newrestarts, newassumptions, ohid)
			    
		    in
			left (formuse, nil, restarts, newassumptions, ohid)
		    end

		fun zeros nil = true
		  | zeros (0::x) = zeros x
		  | zeros _ = false

		(* trigger all the clauses whose heads match *)
		fun match (nil, alts, true) = alts
		  | match (nil, alts, false) =
		    (if Debug.debugp "empties" then
			 (print ("No clauses with same head for ");
			  Cterm.printTerm (print) rightatom;
			  print "\n")
		     else ();
		     alts)
		  | match (L{head, id, ohid, ancestors, disjunct, 
			     rule=rule as FR{translation=lefttrans,...}} :: more,
			   alts, b) =
		    let
			val applies = 
			    (case forbidden of
				 SOME guy => guy <> id
			       | NONE => true) andalso
				 (case (mustuse,source) of
				      (SOME _,SOME guy) => 
					  zeros ohid orelse guy <= id
				    | _ => zeros ohid) andalso
				      Cterm.sameHead(head, rightatom)
		    in
			if applies then
			    if Cterm.quickMatch(head, lefttrans, 
						rightatom, righttrans, 
						uni) then
				case trigger(rule, id, ancestors, ohid,
					     if disjunct then [rule] else nil) of
				    NONE => match(more, alts, true)
				  | SOME opt => match(more, opt :: alts, true)
			    else
				match(more, alts, true)
			else
			    match(more, alts, b)
		    end
	    in
		match(axioms @ (rev assume), alternatives, false)
	    end
	  | backchain _ = raise (Fail "Backchaining on non-atomic goal")

	local
	    (* 
	     * HEADS: get all the atomic formulas that a rule can
	     * be used to establish and swap conjunctions to get
	     * easily to each.
	     *)
	    fun heads (f as Atomic s, l) = (s, nil, f) :: l 
	      | heads (And (p, q), l) = heads(q, heads (p, l))
	      | heads (Or nil, _) = raise Match
	      | heads (Or (p :: r), l) = 
		let
		    fun wraprest(n,nil,prev,all) = all
		      | wraprest(n,p::r,prev,sofar) =
			wraprest(n+1,r,p::prev,
				 (foldr (fn ((s,h,f),l) => (s,n::h,Or(f :: (U.rappend prev r)))::l)
				  sofar (heads (p,nil))))
		in
		    wraprest(1, r, [p],
			     (foldr (fn ((s,h,f),l) => (s,0::h,Or(f :: r))::l) 
			      l (heads (p,nil))))
		end

	      | heads (All(i, f, p), l) =
		(foldr (fn ((s,h,p),l) => (s, h, All(i,f,p))::l) 
		 l (heads (p,nil)))
	      | heads (Some(i, f, p), l) =
		(foldr (fn ((s,h,p),l) => (s,h,Some(i,f,p))::l) 
		 l (heads (p,nil)))
	      | heads (Imp (p, q), l) =
		(foldr (fn ((s,h,f),l) => (s,h,Imp(p,f))::l) 
		 l (heads (q,nil)))
	      | heads (Box (i,p), l) =
		(foldr (fn ((s,h,f),l) => (s,h,Box(i,f)) :: l)
		 l (heads (p,nil)))
	      | heads (Equals _,_) = raise Match

	    val ct = ref 0
	in
	    fun makeId () =
		(ct := !ct + 1; !ct)

	    (*
	     * MAKELEFT: take a formula and details about the context and
	     *    compile it so that it can be backchained on at the right
	     *    times...
	     *)
	    fun makeLeft (formula, trans, skolemargs, world, 
			  disjunct, ancestors, seq) =
		let
		    val id = makeId ()
		in
		    (foldl (fn ((head, ohid, form), l) => 
			     L{head = head,
			       rule = FR{formula=form, 
					 translation = trans,
					 skolemargs = skolemargs,
					 world = world},
			       ancestors = ancestors,
			       disjunct = disjunct,
			       id = !ct,
			       ohid = ohid} :: l)
		    seq (heads (formula, nil)), 
		    id)
		end

	    fun embedded(j, nil) = false
	      | embedded(j, h::rest) = List.exists (fn x => x=j) rest
	end

	(*
	 * doRestart:
	 *   called when the current goal stack is empty but
	 *   the state has restart goals to attend to.
	 * The top restart goal is popped off, and for each
	 *   goal in the history list that the new case for
	 *   analysis might contribute to proving (as enforced
	 *   by imposing a modal matching constraint) we generate a 
	 *   new possibility.
	 *)
	fun doRestart(S{mustuse, contributions, assumptions,
			maxdepth, problem, uni, data,...},
		      RS{disjunct, translation, skolemargs, index, 
			 ancestors=norestarts, history, 
			 id = disjid, swapforbid, depth},
		      restarts, alternatives) =
	    let
		val newhistories = 
		    if swapforbid then
			case contributions of 
			    NONE => nil
			  | SOME guys => intersectHistories(guys, history)
		    else
			history
		val newforbid = 
		    if swapforbid then
			mustuse
		    else
			NONE

		fun trigger (nil, alts) = alts
		  | trigger (h as (H{goal, translation=rtrans, 
				     assume, world, ancestors, id} :: rest), 
			     alts) =
		    if embedded(world, norestarts) then 
			alts
		    else case Patheq.makeMatchedAncestor(problem, index, world) of
			NONE => trigger(rest, alts)
		      | SOME newproblem =>
			    let 
				val (newseq, idct) = 
				    makeLeft (disjunct, translation, skolemargs, index, 
					      true, norestarts, assume)

				fun mesg () =
				    if Debug.debugp "restart" then
					(print "Restarting goal ";
					 prettyprintFormula (print) uni (rtrans, goal);
					 print " at depth ";
					 print (Int.toString depth);
					 print " (limited to ";
					 print (Int.toString maxdepth);
					 print ") with new assumption ";
					 prettyprintFormula (print) uni (translation, disjunct);
					 if swapforbid then
					     print " and last out\n"
					 else print " and last in\n")
				    else
					()
			    in
				mesg();
				trigger(rest, S{goals=[R{goal=FR{formula = goal,
								 translation =  rtrans,
								 world = world,
								 skolemargs = nil},
							 ancestors = ancestors,
							 history = H{goal=goal,
								     translation = rtrans,
								     assume = newseq,
								     world=world,
								     ancestors=ancestors,
								     id= makeId()} :: rest,
							 assume = newseq}],
						uni = uni,
						depth = depth,
						maxdepth = maxdepth,
						maxworld = world,
						restarts=restarts,
						problem=newproblem,
						mustuse= SOME idct,
						source= SOME disjid,
						forbidden = newforbid,
						contributions = NONE,
						assumptions = assumptions,
						data=data} :: alts)
			    end
	    in
		trigger (newhistories, alternatives)
	    end

	local 
	    fun cancelblock (NONE, _) = 
		(true)
	      | cancelblock (_, SOME list) = 
		(true)
	      | cancelblock _ = 
		(false)

	    fun incdepth (S{goals,restarts,problem,uni,depth,maxdepth,
			    mustuse,source,forbidden,contributions,
			    assumptions,data,maxworld}) =
		S{goals = goals,
		  restarts = restarts,
		  problem = problem,
		  uni = uni,
		  depth = depth,
		  maxdepth = maxdepth + 1,
		  mustuse = mustuse,
		  source = source,
		  forbidden = forbidden,
		  contributions = contributions,
		  maxworld = maxworld,
		  assumptions = assumptions,
		  data = data}
	in
	    (*
	     * restart: 
	     *    manage completing a case.
	     * Maybe we found a proof, maybe we didn't apply
	     *    a case we needed to consider, or maybe we should
	     *    continue with further case analysis.
	     *)
	    fun restart (s as S{restarts, depth, maxdepth, uni, data, problem,
				mustuse, contributions, assumptions, ...},
			 alternatives) =
		if cancelblock(mustuse, contributions) then 
		    (case restarts of
			 nil =>
			     (SOME (PF{data=data, 
				       problem=problem, 
				       assumptions = assumptions,
				       uni=uni}, 
				    ALTS alternatives),
			      alternatives)
		       | (top :: rest) =>
			     if depth = maxdepth then
				 ((if Debug.debugp "restart" then
				       print "Block postponed for exceeding depth\n"
				   else ());
				       (NONE, alternatives @ [incdepth s]))
			     else
				 (NONE,
				  doRestart(s, top, rest, alternatives)))
		else
		    ((if Debug.debugp "restart" then
			  print "Block rejected for lack of cancellation\n"
		      else ());
			  (NONE, alternatives))
	end

	datatype 'a resolution =
	    Continue of 'a state list
	  | Backchain of right * right list * Patheq.pathpt * 'a state list * 'a state
	  | Solved of 'a proof * 'a more
	  | Failed

	local
	    (* abstract updating of state *)
	    fun asgngoals(S{goals, restarts, problem, uni, depth, maxdepth,
			    mustuse, source, maxworld,
			    forbidden, contributions, assumptions, data},
			  newgoals) =
		S{goals=newgoals, restarts=restarts, problem=problem, uni=uni, 
		  mustuse = mustuse, source = source, forbidden=forbidden, 
		  contributions = contributions, assumptions = assumptions, data=data, 
		  depth = depth, maxdepth = maxdepth, maxworld = maxworld}

	    (* handle goals of equality by calling unifier *)
	    fun equality(S{goals, restarts, problem, uni, depth, maxdepth,
			   mustuse, source, maxworld,
			   forbidden, contributions, assumptions, data},
			 newgoals, translation, world, t1, t2, newmaxworld, alts) = 
		(let
		     val w' = Patheq.keepCopy world
		 in
		     printBack(t1, t2, world, world);
		     case Patheq.equate(problem, w', world) of
			 NONE => alts
		       | SOME p' => 
			     let 
				 val (newsystem, newproblem) =
				     Cterm.unify(Cterm.updateSystem(t1, translation, w',
								    t2, translation, w',
								    p', uni)) 
			 in
			     printAfter(newsystem, newproblem);
			     S{goals=newgoals, restarts=restarts,
			       problem=newproblem, mustuse=mustuse, 
			       depth=depth, maxdepth = maxdepth, maxworld = newmaxworld,
			       uni=newsystem, forbidden=forbidden, source=source,
			       contributions=contributions, assumptions=assumptions,
			       data=data} :: alts
			 end
		 end
		     handle Cterm.Clash => (print "Unification clashed.\n"; alts)
			  | Cterm.Cycle => (print "Occurs check violated.\n"; alts)
			  | Cterm.Scope => (print "No solution for scope constraints.\n"; 
					    alts))

	    (* try user-defined alternative strategies, eg. abduction *)
	    fun special (f, S{goals, restarts, problem, uni, depth, maxdepth,
			      mustuse,source, maxworld,
			      forbidden,contributions,assumptions,data},
			 topgoal, othergoals, translation, world, newmaxworld, alts) =
		(let
		     val specials = f (topgoal, translation, world, data, uni, problem)
		 in
		     foldr (fn ((data, uni, problem), l) => 
			   S{goals=othergoals, restarts=restarts, depth = depth,
			     problem=problem, mustuse=mustuse, source=source,
			     uni=uni, forbidden=forbidden, maxdepth = maxdepth, 
			     contributions=contributions, assumptions=assumptions,
			     data=data, maxworld = newmaxworld} :: l)
		     alts specials
		 end)
		
	in
	    fun chain (specialf, axioms) 
		(goal as R{goal=FR{formula=Atomic g, translation,world,...},...}, 
		 goals, maxworld, newalts, s) =
		special(specialf, s, g, goals, translation, world, maxworld,
			backchain(goal, s, goals, maxworld, newalts, axioms))
	      | chain _ _ = nil

	    (*
	     * rewrite:
	     *   process the current state.
	     *   maybe it's empty: in that case we've failed.
	     *   maybe it's done: in that case we should look at restarts
	     *   otherwise it has a top formula that we should analyze
	     *      and backtrack for.  The analysis uses the STRIP 
	     *      function to implement the right sequent rules.
	     *)
	    fun rewrite (nil) = 
		Failed

	      | rewrite ((s as S{goals=nil,...}) :: alts) =
		(case restart (s, alts) of
		     (SOME answer, _) => Solved answer
		   | (NONE, newalts) => Continue newalts)

	      | rewrite ((s as S{goals=(first as R{goal=FR{world,...},...})::more,maxworld,...})
			 :: alts) =
		let
		    (* at atomic goals, backchain *)
		    fun strip (goal as R{goal=FR{formula=Atomic g, world,...},...},
			       goals, maxworld, newalts) =
			Backchain (goal, goals, maxworld, newalts, s)
			
		      (* for equality, invoke the unifier *)
		      | strip (goal as R{goal=FR{formula=Equals(t1,t2),translation,world,...},...},
			       goals, maxworld, newalts) =
			Continue (equality(s, goals, translation, world, 
					   t1, t2, maxworld, newalts))

		      (* for conjunction: make two new goals and work first *)
		      | strip (R{goal=FR{formula=And(p,q),translation,skolemargs,world},
				 assume,ancestors,history,...},
			       goals, maxworld, newalts) =
			strip(R{goal=FR{formula=p, translation=translation, 
					skolemargs=skolemargs, world=world},
				assume=assume, 
				ancestors=ancestors, history=history},
			      R{goal=FR{formula=q, translation=translation, 
					skolemargs=skolemargs, world=Patheq.keepCopy(world)},
				assume=assume, 
				ancestors=ancestors, history=history}::goals,
			      maxworld, newalts)

		      (* for disjunction: try first but remember second *)
		      | strip (R{goal=FR{formula=Or (p :: r),translation, skolemargs,world},
				 assume,ancestors,history,...},
			       goals, maxworld, newalts) =
			strip(R{goal=FR{formula=p, translation=translation, 
					skolemargs=skolemargs, world=world},
				assume=assume,
				ancestors=ancestors, history=history},
			      goals, 
			      maxworld, 
			      foldr (fn (q,l) =>
				    asgngoals(s, 
					      R{goal=FR{formula=q,
							translation=translation, 
							skolemargs=skolemargs, world=world},
						assume=assume,
						ancestors=ancestors, 
						history=history} :: goals) :: l)
			      newalts r)

		      | strip (R{goal=FR{formula=Or _,...},...}, _, _, _) = raise Match

		      (* for box, create a new scope and work inside it *)
		      | strip (R{goal=FR{formula=Box(i,p),translation,skolemargs,world},
				 assume,ancestors,history,...},
			       goals, maxworld, newalts) =
			let val newworld = Patheq.extendConst(world, i)
			    val (newancestors, newhistory, newskargs) =
				if Modspec.isFunctional i then
				    (ancestors, history, skolemargs)
				else
				    let val copy = Patheq.keepCopy(newworld)
				    in (copy :: ancestors,
					H{goal=p, assume=assume,
					  translation=translation, world=copy,
					  ancestors= copy::ancestors, id= makeId()} :: history,
					nil)
				    end
			in
			    strip(R{goal=FR{formula=p, 
					    world = newworld,
					    translation = translation,
					    skolemargs = nil},
				    assume=assume, 
				    ancestors=newancestors,
				    history=newhistory},
				  goals, hd ancestors, newalts)
			end

		      (* for arrow, add assumption and look at consequent *)
		      | strip (R{goal=FR{formula=Imp(p,q),translation,skolemargs,world},
				 assume,ancestors,history,...},
			       goals, maxworld, newalts) =
			let val news = #1(makeLeft(p,translation,skolemargs,
						   Patheq.keepCopy world, false,
						   ancestors,assume))
			    val next = R{goal=FR{formula=q, world=world,
						 translation = translation,
						 skolemargs = skolemargs},
					 assume=news, 
					 ancestors = ancestors, 
					 history=history}
			in strip(next, goals, maxworld, newalts)
			end

		      (* for universal: introduce skolem function *)
		      | strip (R{goal=FR{formula=All(i, f, p), translation, skolemargs,world},
				 assume, ancestors, history,...},
			       goals, maxworld, newalts)  =
			strip(R{goal=FR{formula=p, world=world, 
					skolemargs=skolemargs, 
					translation=Cterm.IntMap.insert
					(translation, i, 
					 Cterm.CI(Cterm.newVar(), f, 
						  Patheq.keepCopy(world), skolemargs))},
				assume=assume, ancestors=ancestors,
				history=history}, 
			      goals, maxworld, newalts)

		      (* for existential: introduce logic variable *)
		      | strip (R{goal=FR{formula=Some(i, _, p), translation, skolemargs, world},
				 assume, ancestors, history,...},
			       goals, maxworld, newalts) =
			let val var = Cterm.newVar()
			in strip(R{goal=FR{formula=p, world=world,
					   skolemargs= var::skolemargs,
					   translation=Cterm.IntMap.insert
					   (translation, i, Cterm.VI(var, Patheq.keepCopy(world)))},
				   assume=assume, 
				   ancestors=ancestors,
				   history=history}, 
			    goals, maxworld, newalts)
			end
		in
		    strip(first, more, maxworld, alts)
		end
	end

	type database = Modspec.collection * left list * Patheq.problem

	local 
	    open Modspec
	in
	    (* 
	     * check:
	     *    make sure syntactic restrictions are observed.
	     *    these include:
	     *       no ways to prove equality.
	     *       no right implications except at scope boundary.
	     *       no hard sequences of left boxes.
	     *)
	    fun check (collection, isgoal) =
		let
		    fun checkLeft (Atomic _, _) = true
		      | checkLeft (Equals _, _) = false
		      | checkLeft (Box (i,f), state) = 
			extendSafely(collection, state, i) andalso
			checkLeft(f, state)
		      | checkLeft (And (p, q), state) =
			checkLeft (p, state) andalso checkLeft(q, state)
		      | checkLeft (Or l, state) =
			List.all (fn x => checkLeft (x, state)) l
		      | checkLeft (Imp(p,q), state) =
			checkLeft(q, state) andalso checkRight(p, false)
		      | checkLeft (All(_, _, p), state) = 
			checkLeft (p, state)
		      | checkLeft (Some(_, _, p), state) = 
			checkLeft (p, state)

		    and checkRight (Atomic _, _) = true
		      | checkRight (Equals _, _) = true
		      | checkRight (Box (i,p), s) = 
			(case querySafely(collection,i,s) of
			     NONE => false 
			   | SOME b => checkRight (p, b))
		      | checkRight (And (p,q), _) = 
			checkRight (p, false) andalso checkRight (q, false)
		      | checkRight (Or l, _) = 
			List.all (fn x => checkRight (x, false)) l
		      | checkRight (Imp(p,q), false) = false
		      | checkRight (Imp(p,q), true) =
			checkLeft(p, newState(collection)) andalso 
			checkRight (q, true)
		      | checkRight (All(_,_,p), s) = checkRight (p, s)
		      | checkRight (Some(_,_,p), s) = checkRight (p, s)
		in
		    if isgoal then
			fn x => checkRight(x, true)
		    else
			fn x => checkLeft(x,newState(collection))
		end
	end

	(*
	 * addFormulas:
	 *    given a database, encoded as modals, axioms, empty problem,
	 *    compile all the formulas and add to set of axioms
	 *)

	fun addFormulas ((collection, axioms, problem), formulas) =
	    let
		val root = Patheq.rootProblem problem
		val checkFormula = check(collection, false)

		fun add ((_,x), l) =
		    if checkFormula x then
			#1(makeLeft(x,Cterm.IntMap.empty(),nil,root,false,[root],l))
		    else
			(print "Illegal [] or = on left or -> on right in formula: \n";
			 printFormula (print) x;
			 print "\n";
			 l)
	    in
		(collection, 
		 U.rappend (foldr add nil formulas) axioms,
		 problem)
	    end

	(*
	 * makeDatabase:
	 *    compute the representation for database consisting of
	 *    FORMULAS as axioms and COLLECTION defining modal inventory
	 *)
	fun makeDatabase (collection, formulas) =
	    let 
		val problem = Patheq.newProblem collection
	    in
		addFormulas((collection, nil, problem), formulas)
	    end

	(*
	 * queryFormula:
	 *    given goal and description of environment,
	 *    try to prove goal.
	 *)
	fun urQuery (f, init) (tr0, poption, goal, (collection, axioms, problem)) =
	    let
		val back = chain (f, axioms)
		fun go s = 
		    case rewrite (s) of
			Continue s => go s 
		      | Backchain data => 
			    go (back data)
		      | Failed => NONE
		      | Solved answer => SOME answer

		val root = Patheq.rootProblem problem
	    in
		if check (collection, true) goal then
		    go ([S{goals=[R{goal=FR{formula=goal,
					    world = root,
					    translation= tr0,
					    skolemargs=nil},
				    assume = nil,
				    ancestors = [root],
				    history = [H{goal=goal,
						 translation= tr0,
						 assume=nil,
						 world=root,
						 ancestors=[root],
						 id= makeId()}]}],
		    uni = (case poption of 
			       NONE => Cterm.newSystem(problem)
			     | SOME (PF{uni,...}) => 
				   (Cterm.makeSystem
				    (Cterm.readoutVar(uni,NONE),
				     problem))),
		    restarts=nil,
		    problem=problem,
		    contributions=NONE,
		    maxworld = root,
		    mustuse = NONE,
		    source = NONE,
		    forbidden = NONE,
		    assumptions = nil,
		    depth = 0,
		    maxdepth = if Debug.debugp "iterative" then 2 else ~1,
		    data=init}])
		else
		    (print "Illegal modality in formula: \n";
		     printFormula (print) goal;
		     print "\n";
		     NONE)
	    end

	fun queryFormula (f, init) (goal, stuff) =
	    urQuery (f, init) (Cterm.IntMap.empty(), NONE, goal, stuff)

	fun followupMetavar (f, init) (pf, goal, stuff) =
	    urQuery (f, init) (Cterm.IntMap.empty(), SOME pf, goal, stuff)
	    
	fun queryFormulaTr (f, init) (tr, goal, stuff) =
	    urQuery (f, init) (tr, NONE, goal, stuff)

	fun requeryFormula (f, init) (ALTS alts, (collection, axioms, problem)) = 
	    let 
		val back = chain(f, axioms)
		fun go s = 
		    case rewrite (s) of
			Continue s => go s 
		      | Backchain data => 
			    go (back data)
		      | Failed => NONE
		      | Solved answer => SOME answer
	    in
		go alts
	    end

	fun makeNewTrans (PF{uni,...}, oldtrans, root) =
	    let
		fun gv (iv, ov, tr, default) =
		    (case Cterm.readoutVar (uni, SOME ov) of
			 Cterm.Bound t => 
			     Cterm.IntMap.insert(tr, iv, Term.TI(ov, root, t))
		       | _ => 
			     Cterm.IntMap.insert(tr, iv, default))

		fun et (iv, Term.TI(ov, _, t), tr) =
		    gv (iv, ov, tr, Term.TI(ov, root, t))
		  | et (iv, Term.VI(ov, _), tr) =
		    gv (iv, ov, tr, Term.VI(ov, root))
		  | et (_, _, tr) = tr
	    in
		Cterm.IntMap.fold et oldtrans (Term.IntMap.empty())
	    end

	fun bestAlias (PF{uni,...}, i) = Cterm.bestAlias(uni,i)

	fun mergeTrans (add, into) =
	    Cterm.IntMap.fold (fn (k,v,l) => Term.IntMap.insert(l,k,v)) add into

	fun followupMetavarTr (f, init) 
	    (tr, pf, goal, stuff as (_,_,problem)) =
	    urQuery (f, init) (makeNewTrans(pf, tr, Patheq.rootProblem problem), 
			       SOME pf, goal, stuff)

	fun setof (f, init, cxt, (collection, axioms, problem)) =
	    let
		val back = chain(f, axioms)
		val root = Patheq.rootProblem problem

		fun go (tr, alts, sofar) =
		    case rewrite alts of
			Continue s => go (tr, s, sofar)
		      | Backchain data => go (tr, back data, sofar)
		      | Failed => rev sofar
		      | Solved (pf, ALTS more) =>
			    (if Debug.debugp "lookbusy"
				 then
				     (print ".";
				      TextIO.flushOut TextIO.stdOut)
			     else ();

			     go (tr, more, pf :: sofar))
			    
		fun cxtsetof (S{problem, uni,...}, 
			      R{goal=FR{skolemargs,translation=oldtranslation,...},
				assume,ancestors,history,...}, world) 
		    (translation, goal) =
		    let
			val newtrans = mergeTrans(oldtranslation,translation)
		    in
			if check (collection, true) goal then
			    (if Debug.debugp "lookbusy" then
				 (print "consulting knowledge base about ";
				  printFormula print goal;
				  print "\n")
			     else();
				 
				 go (translation,
				     [S{goals=[R{goal = FR{formula=goal,
							   world = world,
							   translation = newtrans,
							   skolemargs = skolemargs},
						 assume = assume,
						 ancestors = ancestors,
						 history = [H{goal=goal,
							      translation= newtrans,
							      assume=assume,
							      ancestors = [root],
							      id = makeId(),
							      world=world}]}],
				     uni = uni,
				     restarts = nil,
				     problem = problem,
				     contributions=NONE,
				     maxworld = world,
				     assumptions = nil,
				     forbidden=NONE,
				     mustuse=NONE,
				     source = NONE,
				     depth = 0,
				     maxdepth = if Debug.debugp "iterative" then 2 else ~1,
				     data = init}], nil) before
				 
				 (if Debug.debugp "lookbusy" then
				      (print "\n")
				  else
				      ()))
			else
			    (print "Illegal modality in formula: \n";
			     printFormula (print) goal;
			     print "\n";
			     nil)
		    end

		fun step s =
		    case rewrite (s) of
			Continue s => step s
		      | Backchain (topgoal as R{goal=FR{world=world,...},...}, _, _, _, s) => 
			    (Patheq.keepCopy (world), cxtsetof (s, topgoal, world))
		      | _ => 
			    (print "Illegal context formula: \n";
			     printFormula (print) cxt;
			     print "\n";
			     (root, fn _ => nil))
	    in
		if check (collection, true) cxt then
		    step ([S{goals=[R{goal=FR{formula=cxt,
					      world = root,
					      translation= Cterm.IntMap.empty(),
					      skolemargs = nil},
				      assume = nil,
				      ancestors = [root],
				      history = nil}],
		    uni = Cterm.newSystem(problem),
		    restarts = nil,
		    problem = problem,
		    contributions=NONE,
		    maxworld = root,
		    assumptions = nil,
		    forbidden=NONE,
		    source = NONE,
		    mustuse=NONE,
		    depth = 0,
		    maxdepth = if Debug.debugp "iterative" then 2 else ~1,
		    data = init}])
		else
		    (print "Illegal modality in context formula: \n";
		     printFormula (print) cxt;
		     print "\n";
		     (root, fn _ => nil))
	    end
	    
	fun displayProof (show, out) (PF{data, 
					 problem, 
					 assumptions,
					 uni}) =
	    let
		val pt = (fn top => 
			  (out o (Cterm.writeprettyReifiedTerm 
				  (uni, 
				   Cterm.Patheq.writeprettyPathpt (problem, false, true), 
				   top))))
		val wp = Patheq.writeprettyPathpt (problem,false,true)
		val pp = Patheq.prettyprintPathpt out problem

		fun pa (FR{formula, world, translation,...}, w) =
		    (print "Used ";
		     prettyprintReifiedFormula out (uni,problem) (translation, formula, world);
		     print " toward ";
		     pp w;
		     print "\n")
	    in 
		print "Case history:\n";
		app pa assumptions;
		show (pt,wp,data);
		Cterm.printSystem out uni
	    end

	fun readoutMetavar (PF{uni,...}) =
	    case Cterm.readoutVar(uni,NONE) of
		Cterm.Bound t => SOME(Atomic t)
	      | Cterm.Free => NONE
	      | Cterm.Miss => NONE

	fun bindArg (f, var, world) =
	    let
		fun ba (Atomic c) = Term.bindArg(c,var,world)
		  | ba (Equals (t,_)) = Term.bindArg(t,var,world)
		  | ba (Box (_,p)) = ba p
		  | ba (And (p,_)) = ba p
		  | ba (Or nil) = Term.IntMap.empty()
		  | ba (Or (h::_)) = ba h
		  | ba (Imp (p,q)) = ba q
		  | ba (All (_,_,p)) = ba p
		  | ba (Some (_,_,p)) = ba p
	    in
		ba f
	    end

	val _ = 
	    app Debug.register 
	    [("trace", 
	      "Wait at places in search\n", 
	      false),
	     ("spew", 
	      "Print out intermediate status\n", 
	      false),
	     ("match", 
	      "Describe what matches take place\n", 
	      false),
	     ("restart",
	      "Describe action of the restart engine\n",
	      true),
	     ("empties",
	      "Warn when symbol is undefined\n",
	      false),
	     ("lookbusy",
	      "Make noise when you get a result\n",
	      false),
	     ("iterative",
	      "Search by restart depth with iterative deepening\n",
	      true)	     
	     ];
    end
