(*
 * Matthew Stone 
 * Dialup 0.1 - 3/96
 * modality-sig.sml
 *
 * Signature file for maintaining the properties of
 * a collection of necessary modalities related by
 * inclusions, as described in "Tree Construction 
 * for Reasoning about Necessity".
 *
 * For external purposes, each modality is identified
 * by an integer code.  An inventory stores a related
 * group of modalities.
 *
 * Compilation takes the transitive closure of
 * inclusions: this determines the kinds of constraints
 * that should be added for different kinds of 
 * modals, and speeds up matching path variables
 * against path constants.
 *
 * A reminder: modals differ in class according to the
 * introspection properties that are available for the
 * different modalities.  The rules are these:
 * 	K_MOD: no introspection
 *	T_MOD: box p implies p
 *	K4_MOD: box p implies box box p
 * 	S4_MOD: box p implies p, box p implies box box p
 *      F_MOD: box p iff diamond p
 *		(equivalently?: (box p->box q) -> box(p->q))
 *
 * Each inventory maintains (for convenience) an S4 modality
 * called the top whose variables match constants of any type.
 *)

signature MODALITY = 
    sig
	type inventory
	datatype 'a path = P of {modality:int, info:'a} list
	datatype mclass = 
	    K_MOD | T_MOD | K4_MOD | S4_MOD | F_MOD

	val newInventory : unit -> inventory
	val top: inventory -> int

	(* return a code for a new modality with mclass *)
	val addModality: inventory * mclass -> int

	val listElements : inventory -> int list

	(* register that [first] p implies [second] p *)
	val addInclusion: inventory * int * int -> unit

	(* build transitive closure and types *)
	val compile: inventory -> unit

	(* 
	 * given a path of variables and a path of constants 
	 * and the current positions of the variables 
	 * against the constants, return the least 
	 * well-typed match, if there is one
	 *)
	val leastMatch : ('a path * 'b path * 
			  int list * inventory -> 
			  ('a * 'b option) list option)

	(* 
	 * return the type the modal should be treated as,
	 * given its possible inclusions
	 *)
	val status : inventory * int -> mclass

	(* 
	 * modality i is clean iff 
	 * whenever [i] p -> [j] p and j is neither
	 * reflexive nor functional, then i is not
	 * reflexive and j is not transitive under i.
	 *   (clean modalities turn out unambiguous,
	 *    see "Tree Construction"...)
	 *)
	val isClean : inventory * int -> bool

	(* 
	 * modalities i and j are separate iff
	 * i is reflexive, and and
	 *   no subtype of j is a subtype of i; or
	 * i is not reflexive, and no subtype of j  
	 *   is a subtype of a transitive subtype of i.
	 *    (if all consecutive modalities on the left
	 *     turn out separate, things are easy, see
	 *     "Tree Construction"...)
	 *)
	val isSeparate : inventory * int * int -> bool

	(* 
	 * true if there is an inclusion [int] p -> [j] p
	 * where j is a functional modality
	 *)
	val isInFunctional: inventory * int -> bool
    end

