(*
 * Matthew Stone
 * Dialup 0.1 - 3/96
 * modality.sml
 *
 * Code for constructing a structure with
 * MODALITY signature, given library structure
 * implementations IntSet and IntMap.
 *
 * revised for sml-110 5/98
 *)

functor Modality (structure IntSet: MYINTSET;
		  structure IntMap: MYINTMAP) : MODALITY =
    struct

	(**************  DATATYPES *****************)

	(* for depth-first traversal of inclusions *)
	datatype vis = Unseen | Seen | Seeing

	datatype mclass = 
	    K_MOD | T_MOD | K4_MOD | S4_MOD | F_MOD

	(*
	 * A modality record includes some static
	 * information and a dynamic part.
	 * The id is the external code for the
	 * modality, the static bool parameters show
	 * which introspection rules the modality
	 * obeys, the inclusions give the explicitly
	 * specified axioms [id] p -> [other] p.
	 *
	 * The dynamic part includes a flag for
	 * depth-first traversal, the set of all
	 * OTHER modalities whose types match yours (subtypes),
	 * a tabulation of all transitive modalities
	 * whose types match yours (ltms), the guys that
         * match *them* (subtrans), and variables that
	 * show whether you match another modality
	 * with a particular introspection property.
	 *)
	datatype modality = 
	    M of {id: int,
		  functional: bool,
		  reflexive: bool,
		  transitive: bool,
		  inclusions: IntSet.intset ref,

		  visited: vis ref,
		  subtypes: IntSet.intset ref,
		  ltms: IntSet.intset ref,
		  subtrans: IntSet.intset ref,
		  href: bool ref,
		  htrans: bool ref,
		  hfunc: bool ref}

	(*
	 * A collection of related modalities is given
	 * by the number of modalities defined (size),
	 * an S4 modality (top) with [top] p -> [i] p
	 * for all other modalities i in the collection,
	 * a map (modalities) from integers to full
	 * specifications of modalities, and a flag
	 * (compiled) indicating whether the dynamic
	 * information in the collection is up-to-date.
	 *)
	datatype inventory =
	    I of {size:int ref,
		  top: int,
		  modalities:modality IntMap.intmap ref,
		  compiled: bool ref}

	(******************** OPERATIONS ***********************)

	(******* BUILDING AND COMPUTING MODAL RELATIONS ********)

	fun newModality (id, mclass) =
	    M{(* this static information is specified fully, right *)
	      id = id,
	      reflexive = (mclass = T_MOD orelse mclass = S4_MOD),
	      transitive = (mclass = K4_MOD orelse mclass = S4_MOD),
	      functional = (mclass = F_MOD),
	      inclusions = ref (IntSet.empty),

	      (* this dynamic information will ultimately change *)
	      visited = ref Unseen,
	      subtypes = ref (IntSet.empty),
	      ltms = ref (IntSet.empty),
	      subtrans = ref (IntSet.empty),
	      href = ref (mclass = T_MOD orelse mclass = S4_MOD),
	      htrans = ref (mclass = K4_MOD orelse mclass = S4_MOD),
	      hfunc = ref (mclass = F_MOD)}

	fun newInventory () =
	    I{size = ref 1,
	      top = 0,
	      modalities = ref (IntMap.insert(IntMap.empty(), 0, 
					      newModality (0, S4_MOD))),
	      compiled = ref false}

	fun top (I{top,...}) = top

	(* get all elements *)
	fun listElements (I{modalities,...}) =
	    IntMap.fold (fn (i,_,l) => i :: l) (!modalities) nil

	(* generate a new element -- include in top record *)

	fun addModality(I{size, top, modalities, compiled}, mclass) =
	    let
		val M{inclusions,...} = 
		    IntMap.find(!modalities, top) 
		    handle IntMap.NotFound =>
			raise (Fail ("add modality (modals.sml): " ^
				     "no record for top: " ^
				     (Int.toString top)))
	    in
		size := !size + 1;
		modalities := IntMap.insert(!modalities, !size, 
					    newModality (!size, mclass));
		inclusions := IntSet.add(!inclusions, !size);
		compiled := false;
		!size
	    end

	(*
	 * get the dynamic class of modality m --
	 * this specifies the kind of constraints
	 * that should be added to the tree problem
	 * for variables of type m.
	 *)
	fun status (I{modalities,...}, m) =
	    let
		val M{functional, href, htrans,...} = 
		    (IntMap.find(!modalities, m)
		     handle IntMap.NotFound => 
			 raise (Fail ("status (modality.sml): " ^ 
				      "missing modality " ^ 
				      (Int.toString m))))
	    in
		if functional then F_MOD else
		    case (!href, !htrans) of
			(true, true) => S4_MOD
		      | (false,true) => K4_MOD
		      | (true,false) => T_MOD
		      | _ => K_MOD
	    end

	(*
	 * does modality m have a functional modality
	 * as a subtype (given dynamic info). 
	 * This is needed because F goals have
	 * different restart patterns from other
	 * goals (since they lack eigenvariable
	 * conditions on possible world paths).
	 *)
	fun isInFunctional(I{modalities,...}, m) =
	    let
		val M{hfunc, href, htrans,...} = 
		    (IntMap.find(!modalities, m)
		     handle IntMap.NotFound => 
			 raise (Fail ("isInFunctional (modality.sml): " ^
				      "missing modality " ^ 
				      (Int.toString m))))
	    in
		!hfunc
	    end

	(*
	 * given dynamic records (modalities), 
	 * can a constant of modality (const)
	 * appear in the value of a path 
	 * variable over modality (modality)?
	 * Return transitive subtypes of 
	 * modality, while you're at it.
	 *
	 * Because subtypes stores OTHERS, you
	 * must also check whether it has the same
	 * type as you.
	 *)
	fun matchp modalities (modality, const) =
	    let
		val M{subtypes,ltms,...} = 
		    IntMap.find(modalities,modality)
		val result =
		    modality = const orelse IntSet.member(!subtypes,const)
	    in
		(result,!ltms)
	    end
	handle IntMap.NotFound => 
	    (raise (Fail ("matchp (modality.sml): " ^
			  "missing modality" ^ 
			  (Int.toString modality))))

	(* 
	 * modality i is clean if whenever j is a subtype of i, and 
	 * j is not reflexive or functional, then i is not 
	 * reflexive and j is not
	 * a subtype of any transitive modality under i.
	 *)
	fun isClean (inv as I{modalities,...}, i) =
	    let
		val M{href, htrans, subtypes, ltms, subtrans,...} =
		    (IntMap.find(!modalities, i)
		     handle IntMap.NotFound =>
			 raise (Fail ("isClean (modality.sml): " ^
				      "missing modality " ^ 
				      (Int.toString i))))

		(* j is unsafe if var i matches against jj *)
		fun isUnsafe j = IntSet.member(!subtrans, j)

		(* j is dangerous if not reflexive or functional *)
		fun isDangerous j = 
		    let val mc = status(inv, j) 
		    in mc = K4_MOD orelse mc = K_MOD 
		    end
		
		(* does j show that i is not clean? *)
		 val uncleanWitness =
		     if !href then isDangerous
		     else (fn j => isDangerous j andalso isUnsafe j)
	    in
		(* does i show i is not clean? *)
		(not (!htrans andalso not(!href))) andalso

		(* does any proper subtype of i show i is not clean? *)
		(case (IntSet.exists uncleanWitness (!subtypes))
		     of NONE => true
		   | SOME _ => false)
	    end

	(*
	 * modalities i and j are separate iff
	 * i is reflexive, 
	 *   no subtype of j is a subtype of i; or
	 * i is not reflexive, and no subtype of j  
	 *   is a subtype of a transitive subtype of i.
	 *)
	fun isSeparate(inv as I{modalities,...}, i, j) =
	    if j = i then
		(case status(inv, i) of
		     F_MOD => true
		   | K_MOD => true
		   | _ => false)
	    else
		let
		    (* get description of j *)
		    val M{subtypes=ref jsubtypes, href=ref jrefl,...} =
			(IntMap.find(!modalities, j)
			 handle IntMap.NotFound =>
			     raise (Fail ("isSeparate (modality.sml): " ^
					  "missing modality " ^ 
					  (Int.toString j))))

		    (* get description of i *)
		    val M{subtrans = ref trans, subtypes=ref isubtypes,
			  href = ref irefl,...} =
			(IntMap.find(!modalities, i)
			 handle IntMap.NotFound =>
			     raise (Fail ("isSeparate (modality.sml): " ^
					  "missing modality " ^ 
					  (Int.toString i))))

		    (* is x a bad modality? *)
		    val overlap = 
			if irefl then fn x => IntSet.member(isubtypes,x)
			else fn x => IntSet.member(trans,x)
		in
		    not (overlap j) andalso
		    (case (IntSet.exists overlap jsubtypes) of
			 NONE => true
		       | SOME _ => false)
		end
    
	exception Circular

	(* add axiom [i] p -> [j] p *)
	fun addInclusion(I{size, top, modalities, compiled}, i, j) =
	    let
		val M{functional,inclusions,...} = 
		    IntMap.find(!modalities, i)
	    in
		if functional then 
		    (print ("Error (in addInclusion): " ^ 
			    "Inclusion to F modality: " ^
			    (Int.toString i) ^ "\n"))
		else 
		    (inclusions := IntSet.add(!inclusions, j);
		     compiled := false)
	    end
	handle IntMap.NotFound => 
	    (print ("Error (in addInclusion): " ^
		    "Undefined modality: " ^
		    (Int.toString i) ^ "\n"))

	(*
	 * compile recomputes the dynamic parts of the 
	 * modalities currently stored, by doing a 
	 * depth-first traversal, using the update
	 * function...
	 *
	 * base is a helper function for the main
	 * loop that incorporates the dynamic information
	 * about the passed modality into the record
	 * being computed at the previous iteration up.
	 *
	 * all is straightforward except possibly 
	 * the computation of transitive modalities,
	 * which includes those from both specifications,
	 * along with those from either specification
	 * that are not less than a transitive 
	 * modality from the other specification.
	 *)
	fun base(M{subtypes = ref mysubtypes,
		   subtrans = ref mysubtrans,
		   ltms = ref mytms, href, hfunc,...},
		 (subs, refp, funcp, tms, subtrans)) =
	    let
		val newtms = IntSet.difference(mytms, subtrans)
		val goodtms = IntSet.difference(tms, mysubtrans)
	    in
		(IntSet.union(subs, mysubtypes),
		 refp orelse !href,
		 funcp orelse !hfunc,
		 IntSet.union(IntSet.intersection(mytms, tms),
			      IntSet.union(newtms, goodtms)),
		 IntSet.union(subtrans, mysubtrans))
	    end

	fun compile(I{size, top, modalities, compiled}) =
	    let
		fun update (m as M{visited=ref Seen,...}, data) =
		    base(m, data)

		  | update (M{visited=ref Seeing,...}, _) = 
		    raise Circular
		    
		  | update (m as M (r as {id, inclusions = ref incls,...}), data) =
		    let
			fun rup (x,l) =
			    update(IntMap.find(!modalities,x), l)
			    handle IntMap.NotFound => 
				raise (Fail ("rup (modality.sml) :" ^
					     "missing modality: " ^
					     (Int.toString x)))
			    
			(* loop over your children *)
			val (subs, refp, func, tms, subtrans) = 
			    (#visited r := Seeing;
			     IntSet.fold rup incls
			     (IntSet.empty, #reflexive r, #functional r, 
			      IntSet.empty, IntSet.empty))

			(* add children to subs, who won't have done so *)
			val newsubs = IntSet.union(subs, incls)

			(* replace tms and subtrans if *you're* transitive *)
			val (newtms, newsubtrans) = 
			    if #transitive r then
				(IntSet.singleton(id), 
				 IntSet.add(newsubs,id))
			    else
				(tms, subtrans)
		    in
			(* update dynamic data *)
			#ltms r := newtms;
			#htrans r := not (IntSet.isEmpty(newtms));
			#href r := refp;
			#hfunc r := func;
			#subtypes r := newsubs;
			#subtrans r := newsubtrans;
			#visited r := Seen;

			(* integrate with next loop up *)
			base(m, data)
		    end

		fun reset (_, M r) = #visited r := Unseen
		fun check (_, m) = 
		    update(m, (IntSet.empty, false, false,
			       IntSet.empty, IntSet.empty))
	    in
		IntMap.app reset (!modalities);
		IntMap.app check (!modalities);
		compiled := true
	    end

	(******************** OPERATIONS ***********************)

	(********* BUILDING AND MATCHING MODAL PATHS ***********)

	datatype 'a path =
	    P of {modality:int, info:'a} list

	(*
	 * The matching algorithm is a dynamic programming
	 * algorithm that uses the following idea:
	 * the least match of the first (n) variables
	 * against the first (k) constants is either
         * the least match of the first
	 * (n) variables against the first (k-1) constants,
	 * extended, if possible, by including the kth
	 * constant on the path associated with the nth
	 * variable---or, if this can not be done, 
	 * the least match of the first (n-1) variables
	 * against the first (k) constants, if any.
	 *
	 * A straightforward extension handles the 
	 * case where you want to start from a certain
	 * position...
	 *
	 * To extend a match by a constant, you need to
	 * know something about what shape the first
	 * match had: it could have ended with the
	 * variable empty -- in which case you're in
	 * a First state, and any subtype will match --
	 * or with a variable on a nonempty path -- in
	 * which case you're in a Second state, and
	 * only subtypes of some available
	 * transitive modality will match.  And of 
	 * course there might have been no match, 
	 * in which case you're in the None state.
	 *)

	local	
	    datatype ('a,'b) spos = 
		None of {modality:int, 
			 reflexive:bool, 
			 info:'a, 
			 atleast:int}
	      | First of {match: ('a * 'b option) list, 
			  modality:int, 
			  reflexive:bool, 
			  info:'a}
	      | Second of {match: ('a * 'b option) list, 
			   modality:int, 
			   reflexive:bool,
			   ltms:int list, 
			   info:'a}

	    fun printSpos (None _) = print "# "
	      | printSpos (First _) = print "F "
	      | printSpos (Second _) = print "S "

	    (* 
	     * init: modality IntMap.intmap ->
	     *       'a,'b path list * int list * 
	     *       ('a,'b spos list * ('a,'b option) list option) ->
	     *       'a,'b spos list * ('a,'b option) list option * int
	     *
	     * The path specifies variables and data to match,
	     * the int list specifies the current beginnings of
	     * the variables (which can only be increased),
	     * the spos list accumulates the states achieved 
	     * for matches against each variable before 
	     * considering any constants, and the final match.
	     *)
	    fun init modalities (nil, nil, (list, last)) =
		(rev list, last, 0)
	      | init modalities ({modality, info} :: more, x :: atleast, 
				 (list, last)) =
		(let
		     val M{href,...} = IntMap.find(modalities, modality)
		 in
		     init modalities 
		     (more, atleast,
		      (case (last, x) of
			   (* 
			    * if there's something so far, you could
			    * start this variable next, and, if this
			    * variable can be empty, start the next
			    * as well:
			    *)
			   (SOME match, 0) => 
			       (First{match=match,
				      modality=modality,
				      info=info,
				      reflexive= !href} :: list,
				if !href then SOME((info, NONE) :: match)
				else NONE)

			 (* otherwise, just record failure *)
			 | _ =>
			       (None{modality=modality,
				     info=info,
				     atleast=x,
				     reflexive= !href} :: list,
				NONE)))
		 end
	     handle IntMap.NotFound => 
		 raise (Fail ("init (modality.sml) : " ^
			      "undefined modality: " ^
			      (Int.toString modality))))
	      | init _ _ = 
		raise (Fail "init (modality.sml) - length mismatch")

	    (* 
	     * revise: take the input specification {this,next}
	     *   where we've just tried to extend the passed match
	     *   given data and next depth, and, in the case where we've
	     *   failed, make an adjustment:
	     *)
	    fun revise(SOME match, data, depth) 
		{this, next=n as None{modality, reflexive, atleast, info}} =

		{(* if the depth requirement is satisfied now,
		  * and this is reflexive, send the match down: *)
		 this = (if depth >= atleast then
			     (case (this, reflexive) of
				  (NONE, true) => SOME ((info, SOME data) :: match)
				| _ => this)
			 else this),

		 (* if the depth requirement is satisfied
		  * and there was a previous match, look next time: *)
		 next =	(if depth >= atleast then
			     First{match=match, modality=modality, 
				   reflexive=reflexive, info=info}
			 else n)}
	      | revise (_, _, _) i = i

	    (*
	     * update: modalities * int * 'b ->
	     *         'a,'b spos -> 
	     *         {this:match option,next:'a,'b spos}
	     *
	     * You're looking at a constant of type
	     * int with 'b data, and you want to figure out
	     * what state and match you'll get to from the current 
	     * state by matching this constant.
	     *)
	    fun update (modalities, const, data) =
		let
		    fun matchf x = matchp modalities (x, const)

		    (*
		     * if you're in the First state, you can
		     * match against any subtype, and then
		     * you'll be in the Second state, thinking
		     * you're matching any of the transtive
		     * modalities you include that in turn 
		     * include the type of the constant, or
		     * None, if none do.
		     *)
		    fun up (First{match, modality, reflexive, info}) =
			let
			    val (ok, ltms) = matchf modality
			    val nltms = (List.filter (fn x => #1(matchf x))
					 (IntSet.listItems ltms))
			in
			    {this= (if ok then SOME ((info, SOME data)::match)
				    else NONE),
			     next= (case nltms of
					nil => (None{modality=modality,
						     reflexive=reflexive,
						     atleast=0,
						     info=info})
				      | tms => (Second{match = match, 
						       modality = modality,
						       reflexive = reflexive,
						       ltms = tms,
						       info = info}))}
			end

		      (*
		       * if you're in the second state, you find out
		       * which of your possible transitive types the
		       * guy matches against, and narrow down your
		       * future state accordingly.
		       *)
		      | up (Second{match, modality, reflexive, ltms, info}) =
			let
			    val nltms = (List.filter (fn x => #1(matchf x)) 
					 ltms)
			in
			    case nltms of
				nil => {this=NONE, 
					next=(None{modality=modality,
						    reflexive=reflexive,
						    atleast=0,
						    info=info})}
			      | tms => {this=SOME ((info, SOME data)::match),
					next=(Second{match=match,
						     modality=modality,
						     reflexive=reflexive,
						     ltms=tms,
						     info=info})}
			end

		      (* if you're dead you stay dead: *)
		      | up (None i) = {this = NONE, next = None i}
		in
		    up
		end

	    (*
	     * loop: modalities * int * 'b * int ->
	     *       (states, match, sofar) -> (next states, match)
	     *
	     * for each of the current states, compute the
	     * next state and match on this position including
	     * this variable, given the old state and the match
	     * on this position for previous variables -- by first
	     * trying to update the state (to include more),
	     * then if that fails calling revise to build a 
	     * new state that extends the current match.
	     *)
	    fun loop (modalities, input, data, depth) =
		let
		    fun lp (nil, last, results) =
			(rev results,
			 case last of NONE => NONE | SOME l => SOME (rev l))
		      | lp (one::rest, last, results) =
			let
			    val {this, next} = (revise (last, data, depth+1)
						(update (modalities, input, data)
						 one))
			in
			    lp (rest, this, next::results)
			end
		in
		    lp
		end
	in
	    (* 
	     * to find the least match, call init to start,
	     * then repeatedly call loop to make the next state
	     * (and get the least match so far) until you
	     * run out of constants.
	     *)
	    fun leastMatch (P vars, P consts, atleast, I{modalities,...}) =
		let
		    fun step ({modality, info}, (state, final, depth)) =
			let
			    val (newstate, answer) =
				(loop (!modalities, modality, info, depth)
				 (state, NONE, nil))
			in
			    (newstate, answer, depth+1)
			end
		in
		    #2(foldl step (init (!modalities)
				   (vars, atleast, (nil, SOME nil))) consts)
		end
	end (* local *)
    end (* struct *)

