(*
 * Matthew Stone
 * Dialup 0.1 - 3/96
 * modspec.sml
 *
 * Implementation of the MODSPEC signature.
 *)

functor Modspec(structure Modality: MODALITY;
		structure Symbol: SYMBL;
		structure HashTable: HASH_TABLE) : MODSPEC =
    struct
	structure Modality = Modality
	structure Symbol = Symbol

	datatype absmod =
	    AM of {name: Symbol.symbol,
		   id: int ref,
		   class: Modality.mclass,
		   dim: int,
		   inclusions: int list}

	type modState = 
	    {separate:int list, rigid:bool} Array.array

	datatype levelType =
	    Local
	  | Homogeneous of Symbol.symbol
	    
	datatype collection = 
	    C of {table: (Symbol.symbol, absmod) HashTable.hash_table,
		  levels: int ref,
		  leveltypes: (levelType Vector.vector) ref,
		  tops: (absmod Vector.vector) ref,
		  data: (Modality.inventory Vector.vector) ref}

	fun depth(AM{dim,...}) = dim
	fun index(AM{id,...}) = !id
	fun isFunctional(AM{class,...}) = class = Modality.F_MOD

	exception Miss

	local
	    val topsym = Symbol.makeSymbol(".")

	    fun extend(v, d, n) =
		Vector.tabulate(d+1, 
				fn x => 
				if x = d then n
				else Vector.sub(v,x))

	    fun newi(x) = Modality.newInventory()
	    fun newt data x =
		AM{name=topsym,
		   id=ref(Modality.top(Vector.sub(data,x))),
		   class= Modality.S4_MOD, dim=x, 
		   inclusions=nil}
	in 
	    fun newCollection () =
		let
		    val data = Vector.tabulate(0, newi)
		in
		    C{table = (HashTable.mkTable 
			       (Symbol.hashSymbol, Symbol.sameSymbol)
			       (128, Miss)),
		      levels = ref 0,
		      leveltypes = ref (Vector.tabulate(0, fn x => Local)),
		      tops = ref (Vector.tabulate(0, newt data)),
		      data= ref data} 
		end

	    fun extendCollection (ltype, C{table, levels, 
					   leveltypes, tops, data}) =
		let 
		    val l = !levels
		in
		    levels := l + 1;
		    data := extend(!data, l, newi l);
		    leveltypes := extend(!leveltypes, l, ltype);
		    tops := extend(!tops, l, newt (!data) l)
		end
	end

	fun status(C{data=ref data,...}, AM{id,dim,...}) =
	    Modality.status(Vector.sub(data,dim), !id)
	fun isInFunctional(C{data=ref data,...}, AM{id,dim,...}) =
	    Modality.isInFunctional(Vector.sub(data,dim), !id)

	fun levels (C{levels=ref levels,...}) = levels
	fun top (C{tops=ref tops,...}, level) = Vector.sub(tops,level)
	fun inventory (C{data=ref data,...}, level) = Vector.sub(data,level)

	fun writeModality (AM{name,...}) = (Symbol.nameSymbol name)
	fun printModality out = out o writeModality

	fun auxAddModality(table, data,
			   a as AM{name, id, class, dim, inclusions}) =
	    let
		val key = Modality.addModality(data, class)
	    in
		app (fn i => Modality.addInclusion(data, i, key)) inclusions;
		HashTable.insert(table) (name, a);
		id := key
	    end

	fun addModality(C{table, data=ref data, ...}, a as AM{dim,...}) =
	    auxAddModality(table, Vector.sub(data,dim), a)

	local 
	    fun readStatus ("S4" :: more) =
		(Modality.S4_MOD, more)
	      | readStatus ("K4" :: more) =
		(Modality.K4_MOD, more)
	      | readStatus ("K" :: more) =
		(Modality.K_MOD, more) 
	      | readStatus ("T" :: more) =
		(Modality.T_MOD, more)
	      | readStatus ("ref" :: "trans" :: more) =
		(Modality.S4_MOD, more)
	      | readStatus ("trans" :: "ref" :: more) =
		(Modality.S4_MOD, more)
	      | readStatus ("ref" :: more) =
		(Modality.T_MOD, more)
	      | readStatus ("trans" :: more) =
		(Modality.K4_MOD, more)
	      | readStatus (more) = (Modality.K_MOD, more)

	    fun readInclusions (table, "." :: more, sofar) = (SOME sofar, more)
	      | readInclusions (table, nil, sofar) = (SOME sofar, nil)
	      | readInclusions (table, first :: more, sofar) =
		readInclusions (table, more,
				HashTable.lookup (table) (Symbol.makeSymbol first) ::
				sofar) handle Miss => (NONE, more)

	    fun readLevelType ("local" :: "." :: more) = (SOME Local, more)
	      | readLevelType ("hg" :: name :: "." :: more) =
		(SOME(Homogeneous (Symbol.makeSymbol name)),
		 more)
	      | readLevelType (str) = (NONE, str)

	in
	    fun readModality (_, nil) = (NONE, nil)
	      | readModality (c, "dim" :: more) =
		let
		    val (leveltypeopt, rest) = readLevelType(more)
		in
		    case leveltypeopt of
			NONE => (NONE, "dim" :: more)
		      | SOME t => (extendCollection(t, c);
				   (NONE,rest))
		end
	      | readModality (C{table, tops, levels, leveltypes, data},
			      name :: stuff) =
		let
		    val (class, more) = readStatus stuff
		    val (dim, next) = (!levels-1, more)
		    val (inclusions, final) = readInclusions (table, next, nil)
		in
		    case inclusions of 
			NONE => (NONE, name :: stuff) 
		      | SOME l => (SOME(AM{name=Symbol.makeSymbol name,
					   id = ref 0,
					   class=class,
					   dim=dim,
					   inclusions=map index l}),
				   final)
		end (* let *)
	end (* local *)

	local
	    fun ca (Local, table, data, n) =
		Modality.compile(data)
	      | ca (Homogeneous symbol, table, data, n) =
		(auxAddModality(table, data, 
				AM{name=symbol,
				   id = ref 0,
				   dim = n, 
				   class = Modality.K_MOD,
				   inclusions = Modality.listElements data});
		 Modality.compile(data))
		 
	in
	    fun compile (C{table, tops, levels=ref 0, leveltypes, data}) = ()
	      | compile (C{table, tops, levels, leveltypes, data}) = 
		let fun zap 0 = ca(Vector.sub(!leveltypes, 0),
				   table, 
				   Vector.sub(!data,0),
				   0)
		      | zap n = (ca(Vector.sub(!leveltypes, n),
				    table,
				    Vector.sub(!data,n),
				    n);
				 zap (n-1))
		in zap ((!levels)-1) end
	end

	fun lookup (C{table, tops, levels, leveltypes, data}, str) =
	    let
		val s = Symbol.makeSymbol str
	    in
		HashTable.find table s
	    end

	fun newState(C{levels,...}) = 
	    Array.array (!levels, {separate=nil,rigid=false})

	(*
	 * extend safely uses the following algorithm
	 * right now:
	 *
	 * At each stage, the separate list holds all
	 * the unclean modalities that might appear 
	 * to precede the new modality (depending on whether
	 * prior reflexives match the empty string).  We must
	 * make sure the new type is separate from all these
	 * types.  If it is not reflexive, we can reset the
	 * separate list for the next stage; if it is not 
	 * clean, we must (also) add it onto the next 
	 * separate list.
	 *
	 * If a problem arises, we fall into the rigid
	 * case, where we insist that all modalities that
	 * remain are K or F.  Otherwise we fail.
	 *
	 * (See the "Tree Construction" paper...)
	 *)

	local 
	    fun qs(Local, _) = SOME true
	      | qs(Homogeneous hm, name) =
		if Symbol.sameSymbol(name, hm) then
		    SOME true
		else
		    NONE
	in
	    fun querySafely(C{leveltypes=ref lt,...}, AM{dim, name,...}, _) = 
		qs(Vector.sub(lt,dim), name)
	end

	fun extendSafely(C{data,leveltypes,...}, a, AM{id=ref id, dim,...}) =
	    let
		val {separate, rigid} = Array.sub(a, dim)
		val inv = Vector.sub(!data, dim)
		val class = Modality.status(inv, id)
		open Modality
	    in
		case Vector.sub(!leveltypes, dim) of
		    Homogeneous _ => true
		  | Local => 
			if rigid then
			    if K_MOD = class orelse F_MOD = class then
				(Array.update(a, dim, {separate=nil, rigid=true});
				 true)
			    else false
			else
			    let val cleanp = Modality.isClean(inv, id)
				val overlap = 
				    List.exists (fn x=> not(isSeparate(inv,x,id)))
				    separate
				val newseparate =
				    let val n = if (T_MOD = class orelse S4_MOD = class) 
						    then separate else nil
				    in if cleanp then separate else id :: separate 
				    end
			    in
				if overlap andalso K_MOD <> class andalso
				    F_MOD <> class then
				    false
				else
				    (Array.update(a, dim, {separate=newseparate,
							   rigid=rigid orelse overlap});
				     true)
			    end (* let *)
	    end (* let *)
    end (* struct *)
