(*
 * Matthew Stone
 * Dialup 0.1 - 3/96; 5/96
 * patheq-sig.sml
 *
 * The patheq signature
 * abstracts the manipulation of modal paths for the theorem
 * prover.  This means managing constraint problems to order
 * the rule applications in the proof, managing path matches
 * to ensure that the values of variables take on legal types,
 * and establishing interactions between problems at different
 * layers in order to ensure that an increasing-domain 
 * constraint holds between modalities.
 *
 * This abstraction is associated with two data types.  The
 * pathpt data type abstracts the path taken in different
 * modal dimensions from a known point to a particular
 * possible world.
 *
 * The problem data type abstracts the constraint problems
 * at different levels.
 *)

signature PATHEQ =
    sig
	structure Modspec : MODSPEC
	structure Ctree : CONSTRAINT_TREE

	eqtype pathpt
	type problem

	(* build an unconstrained model using modalities of collection *)
	val newProblem : Modspec.collection -> problem

	(* get an abstraction representing the "real world" *)
	val rootProblem : problem -> pathpt
	    
	(* 
	 * add a path variable from modality absmod to the paths
	 * represented by ACTUAL and RESTART, according to the
	 * layout of modalities givein by the problem.  Note:
	 * ACTUAL gives the current possible world, while RESTART
	 * gives the lowest path at which knowing that a formula
	 * is true at ACTUAL could help in a proof.  (If there are
	 * no functional modalities, this is the same as ACTUAL;
	 * otherwise it's ACTUAL minus steps that could match
	 * functional modalities.)
	 *)
	val extendVar : problem * pathpt * pathpt * Modspec.absmod -> 
	    pathpt * pathpt

	(*
	 * add general path variables along each dimension to
	 * the paths represented by ACTUAL and RESTART (interpreted
	 * as above).
	 *)
	val extendVars : problem * pathpt * pathpt -> 
	    pathpt * pathpt

	(*
	 * add a path constant of modal type absmod -- 
	 * an anonymous one in the case of
	 * functional modalities -- to the path represented by
	 * pathpt.
	 *)
	val extendConst : pathpt * Modspec.absmod -> pathpt

	(* make a new path that starts at the place PATHPT ends *)
	val keepCopy : pathpt -> pathpt

	(* make a new element above PATHPT and update problem *)
	val newAncestor : problem * pathpt -> (problem * pathpt) option

	(*
	 * ensure that PATHPT one is under PATHPT two along 
	 * all the dimensions of the problem
	 *)
	val makeAncestor : problem * pathpt * pathpt -> problem option
	val makeMatchedAncestor : problem * pathpt * pathpt -> problem option

	(*
	 * ensure that PATHPT one is properly closer to the
	 * root than PATHPT two -- on the widest-scope modal
	 * dimension
	 *)
	val separateTop : problem * pathpt * pathpt -> problem option

	(*
	 * ensure that PATHPT one is at the same depth from the
	 * root as PATHPT two -- on the narrowest-scope modal
	 * dimension
	 *)
	val matchDepthBottom : problem * pathpt * pathpt -> problem option

	(*
	 * extend the problem so that the two pathpts represent the
	 * same possible world, if possible.  The first should
	 * contain a fixed start followed by just variables 
	 * (corresponding to an axiom); the second should contain
	 * a fixed start followed by just constants (corresponding
	 * to a goal).
	 *)
	val equate : problem * pathpt * pathpt -> problem option

	(*
	 * update the problem to enforce increasing domain restrictions
	 * between modal indices at different layers, if possible.
	 *)
	val updateVtables : problem -> problem option

	(* output relevant information *)
	val printProblem : (string -> unit) -> problem -> unit
	val printPathpt : (string -> unit) -> pathpt -> unit

	val writeprettyPathpt : (problem * bool * bool) -> pathpt -> string
	val splitWriteprettyPathpt : (problem * bool) -> pathpt -> string list

	val prettyprintPathpt : (string -> unit) -> problem -> pathpt -> unit

	(* compare pathpoints *)
	val samePathpt: problem * pathpt * pathpt -> bool
    end

