(*
 * Matthew Stone
 * Dialup 0.1 - 3/96 ; revised 2/98
 * patheq.sml
 *
 * Implementation of PATHEQ signature. In this implementation,
 * paths are stored as vectors of lists -- each list gives 
 * the identification of successive transitions in the path
 * (in reverse order) starting from some fixed start.  So the
 * head of each list represents the current possible world.
 *)

functor Patheq (structure Modspec : MODSPEC;
		structure Ctree : CONSTRAINT_TREE) : PATHEQ =
    struct
	structure Modspec = Modspec
	structure Ctree = Ctree

	(****************** DATATYPES *******************)

	(* Variable, Constant, Root *)
	datatype pathelt = 
	    V of int
	  | C of int
	  | Root of int

	(* number of dimensions, path in each dimension *)
	datatype pathpt = 
	    PP of (int * pathelt list Vector.vector)

	(* node where vars start, vars, node where vars end *)
	datatype equation =
	    E of {leftbase: int,
		  leftvars: pathelt list,
		  rightend: int}

	(*
	 * modal inventory,
	 * number of levels,
	 * vector giving root of each problem
	 * vector giving:
	 *    * constraint problem, 
	 *    * map assigning to each constant in problem
	 *      (list of greater indices, list of those just found)
	 *    * list of typed equations to check
	 *)
	datatype problem = 
	    P of {modals: Modspec.collection,
		  depth: int,
		  root : pathpt,
		  ecs: (Ctree.problem * 
			(int list * int list) Ctree.IntMap.intmap * 
			equation list) 
		  Vector.vector}

	(******************* FUNCTIONS *********************)

	(*
	 * To display a path, just print the ids of the
	 * final (current) world along each dimension
	 *)
	    fun printPathpt out =
		let 
		    fun printPathelt (V i) = (out ("V " ^ (Int.toString i) ^ " "))
		      | printPathelt (C i) = (out ("C " ^ (Int.toString i) ^ " "))
		      | printPathelt (Root i) = (out ("R " ^ (Int.toString i) ^ " "))

		    fun ppt (PP(i,v)) =
			let fun pi j =
			    if i = j then ()
			    else 
				((case Vector.sub(v,j) of 
				      nil => () 
				    | (x::_) => printPathelt x); 
				      pi (j+1))
			in pi 0 end
		in
		    ppt
		end

	(*
	 * mini-abstraction records information about where
	 * each constant and variable comes from, and makes
	 * sure each is assigned a unique id.
	 *
	 * if the element you create is to serve in the Ith
	 * layer of modalities, the FROM argument specified
	 * (and ultimately stored in the ASSOC map) should
	 * give the current world in the I+1th layer of
	 * modalities -- and thereby indicate the domain of
	 * definition of the new element.
	 *)
	local 
	    val ctr = ref 0
	    val assoc = ref (Ctree.IntMap.empty():
			     {modality: Modspec.absmod,
			      domain: pathelt option,
			      exact: bool} Ctree.IntMap.intmap)
	in
	    (* 
	     * make a new variable of class MODALITY,
	     * extending pathelt FROM (if specified).
	     *)
	    fun newVar (modality, from) = 
		let
		    val me = !ctr
		in
		    assoc := Ctree.IntMap.insert(!assoc, me, 
					   {modality = modality,
					    domain = from,
					    exact = false});
		    ctr := me + 2;
		    V me
		end

	    (* 
	     * make a new variable of class MODALITY,
	     * extending pathelt FROM (if specified).
	     *)
	    fun newExactVar (modality, from) = 
		let
		    val me = !ctr
		in
		    assoc := Ctree.IntMap.insert(!assoc, me, 
					   {modality = modality,
					    domain = from,
					    exact = true});
		    ctr := me + 2;
		    V me
		end

	    (* 
	     * make a new constant of class MODALITY,
	     * extending pathelt FROM (if specified).
	     *)
	    fun newConst (modality, from) = 
		let
		    val me = !ctr
		in
		    assoc := Ctree.IntMap.insert(!assoc, me, 		
					   {modality = modality,
					    domain = from,
					    exact = false});
		    ctr := me + 2;
		    C me
		end

	    (* get a new id to be the root of a problem *)
	    fun newRoot () =
		let
		    val me = !ctr
		in
		    ctr := me + 2;
		    me
		end
	    
	    (*
	     * get the modality and the pathelt associated by
	     * assoc with the node with id I.
	     *)
	    fun lookup (i) = #modality (Ctree.IntMap.find(!assoc, i))
		handle Ctree.IntMap.NotFound => 
		    raise (Fail "(patheq) lookup: not found")

	    fun peekup (i) = 
		case Ctree.IntMap.peek(!assoc, i) of
		    NONE => NONE
		  | SOME {domain,...} => domain

	end (* local *)

	(*
	 * newProblem: build a new problem with depth C.
	 * first, get nodes to serve as the root of C
	 * different problems; then create C different
	 * problems and add a constant for each's root;
	 * then build the pathpt corresponding to the
	 * overall root.
	 *)
	fun newProblem (c) = 
	    let
		val size = Modspec.levels c
		val bases = Vector.tabulate(size, fn x => newRoot())
		fun newEcs (x) =
		    ((case Ctree.addConstant(Ctree.newProblem(true),
					     Vector.sub(bases,x)) of
			  NONE => (raise (Fail "Could not start problem"))
			| SOME p => p),
		     Ctree.IntMap.empty(),
		     nil)
			  
		val ecs = Vector.tabulate(size, newEcs)
		val root = Vector.tabulate
		    (size, fn x => [Root(Vector.sub(bases,x))])
	    in
		P {modals = c,
		   depth = size,
		   root = PP(size,root),
		   ecs = ecs}
	    end

	fun rootProblem (P{root,...}) = root

	(*
	 * external operations on pathpts
	 *)
	(* make v' copy of v[size], with v'[except] = value *)
	fun justlike (size, v, except, value) = 
	    Vector.tabulate(size, fn x =>
			    if x = except then value
			    else Vector.sub(v, x))

	local 
	    (*
	     * make a FROM entry for a new transition at layer dim
	     * (recall that this must be the current world
	     * in v[dim+1], if there is one -- NONE otherwise)
	     *)
	    fun next (v, dim, size) =
		if dim = size-1 then NONE
		else 
		    case Vector.sub(v, dim+1) of
			nil => NONE
		      | h :: _ => SOME h
	in
	    (*
	     * make a new variable V of type MODALITY and add it to
	     * the first path; if MODALITY won't match a functional
	     * modality, make the second path agree with the first path
	     * on the dimension changed -- otherwise, just give
	     * the second path again.  (For computing restarts...)
	     *)
	    fun extendVar (P{modals,...}, PP(i,v1), PP(_,v2), modality) =
		let 
		    val dim = Modspec.depth modality
		    val nv = newVar (modality, next(v1, dim, i))
		in 
		    (PP(i, justlike(i, v1, dim, nv :: Vector.sub(v1,dim))),
		     if Modspec.isInFunctional(modals, modality) then
			 PP(i, v2)
		     else
			 PP(i, justlike(i, v2, dim, nv :: Vector.sub(v1,dim))))
		end

	    (*
	     * Extend v1 generally along each modal dimension;
	     * Otherwise, as in extendVar.
	     *)
	    fun extendVars (P{modals,...}, PP(i,v1), PP(_,v2)) =
		let
		    fun extend j = if j = i then nil
			else case extend(j+1) of
			    l as ((h::_)::t) => 
				(newVar(Modspec.top(modals,j), SOME h) ::
				 Vector.sub(v1,j)) :: l
			  | l => (newVar(Modspec.top(modals,j),NONE)::
				  Vector.sub(v1, j)) :: l
		    val extension = Vector.fromList(extend 0)
		in
		    (PP(i,extension),
		     PP(i,Vector.tabulate
			(i,fn x =>
			 if Modspec.isInFunctional
			     (modals, Modspec.top(modals,x)) 
			     then Vector.sub(extension,x)
			 else Vector.sub(v2,x))))
		end

	    (*
	     * Extend v1 generally along each modal dimension;
	     * Make sure higher dimensions much exactly match
	     * the constants to which they are bound.
	     *)
	    fun extendExactVars (P{modals,...}, PP(i,v1), PP(_,v2)) =
		let
		    fun extend j = if j = i then nil
			else case extend(j+1) of
			    l as ((h::_)::t) => 
				(newExactVar(Modspec.top(modals,j), SOME h) ::
				 Vector.sub(v1,j)) :: l
			  | l => (newVar(Modspec.top(modals,j),NONE)::
				  Vector.sub(v1, j)) :: l
		    val extension = Vector.fromList(extend 0)
		in
		    (PP(i,extension),
		     PP(i,Vector.tabulate
			(i,fn x =>
			 if Modspec.isInFunctional
			     (modals, Modspec.top(modals,x)) 
			     then Vector.sub(extension,x)
			 else Vector.sub(v2,x))))
		end

	    (*
	     * make a new constant of type MODALITY and add it to
	     * the path...
	     *)
	    fun extendConst (PP(i,v), modality) =
		let 
		    val dim = Modspec.depth modality
		    val nc = newConst (modality, next(v, dim, i))
		in  
		    PP(i, justlike(i, v, dim, nc :: Vector.sub(v,dim)))
		end

	    (*
	     * return a copy of the path that eliminates all its
	     * transitions, except the last -- that gives the 
	     * current world.
	     *)
	    fun keepCopy(PP(i,v)) =
		PP(i, Vector.tabulate(i, fn x => 
				      (case Vector.sub(v,x) of 
					   nil => nil 
					 | h::_ => [h])))
	end (* local *)

	exception bad

	fun meat (V x) = x
	  | meat (C x) = x
	  | meat (Root x) = x 

	(*
	 * Functions for constructing and relating pathpts 
	 * representing the indices where terms are defined,
	 * for unification with an increasing domain 
	 * constraint.
	 *)
	local
	    (*
	     * given ecs data, new path and old path for
	     * a particular level of the tree, update the
	     * ecs constraint problem to include a variable
	     * for the end of the new path, that will be
	     * closer to the root than the end of the old
	     * path.
	     *)
	    fun newAncestorLevel ((constraints, table, equations), 
				  newElt :: _, descendant :: _) = 
		let
		    val i = meat newElt
		    val j = meat descendant

		    val new = case Ctree.addElement(constraints, i) of
			NONE => NONE
		      | SOME p => case Ctree.addElement(p, i+1) of
			    NONE => NONE
			  | SOME p' =>
				Ctree.addConstraints(p',Ctree.le(j,j+1,i,i+1))
		in
		    case new of 
			NONE => raise bad
		      | SOME new => (new, table, equations)
		end
	      | newAncestorLevel (_, _, nil) = raise bad
	      | newAncestorLevel (_, nil, _) = raise bad
	       
	    (*
	     * given ecs data and two paths, ensure that
	     * the end of the first path has the specified
	     * relation to the end of the second.
	     *)
	    fun makeRLevel relation
		((constraints, vltable, equations),
		 first :: _, second :: _) =
		let
		    val i = meat first
		    val j = meat second
		in 
		    (case Ctree.addConstraints
			 (constraints, relation(j,j+1,i,i+1))
			 of NONE => raise bad
		       | SOME p => (p, vltable, equations))
		end
	      | makeRLevel _ (_, nil, _) = raise bad
	      | makeRLevel _ (_, _, nil) = raise bad

	    val makeAncestorLevel = makeRLevel Ctree.le
	    val makePAncestorLevel = makeRLevel Ctree.lt
	    val makeSameDepthLevel = makeRLevel Ctree.eqdp		
	in
	    (*
	     * build a new pathpt into the passed problem that
	     * will lie closer to the root than the argument
	     * pathpt along each dimension of the problem, if possible
	     *)
	    fun newAncestor(P{modals, depth, root, ecs}, PP(_,v)) =
		let
		    fun mkvars d = if d = depth then nil
			else case mkvars (d+1) of
			    nil => [ [ newVar(Modspec.top(modals,d), NONE) ] ]
			  | l as ([h]::r) => 
				[ newVar(Modspec.top(modals,d), SOME h) ] :: l
			  | _ => raise bad

		    val news = Vector.fromList(mkvars 0)
		in
		    (SOME(P{modals=modals,
			    depth=depth,
			    root=root,
			    ecs= Vector.tabulate(depth, fn x => newAncestorLevel
						 (Vector.sub(ecs,x),
						  Vector.sub(news,x),
						  Vector.sub(v,x)))},
		    PP(depth, news)))
		end
		handle bad => NONE

	    (*
	     * Update the passed problem to ensure that the first
	     * pathpt lies closer to the root than the second
	     * pathpt along each dimension of the problem,
	     * if possible
	     *)
	    fun makeAncestor(P{modals, depth, root, ecs}, PP(_,anc), PP(_,desc)) =
		((SOME(P{modals=modals,
			 depth=depth,
			 root=root,
			 ecs=Vector.tabulate(depth, fn x => makeAncestorLevel
					     (Vector.sub(ecs, x),
					      Vector.sub(anc, x),
					      Vector.sub(desc, x)))})))
		handle bad => NONE

	    (*
	     * Update the passed problem to ensure that the first
	     * pathpt lies closer to the root than the second
	     * pathpt along each dimension of the problem,
	     * if possible, and that they are equal at the 
	     * widest scope modal dimension.
	     *)
	    fun makeMatchedAncestor(P{modals, depth, root, ecs}, PP(_,anc), PP(_,desc)) =
		((SOME(P{modals=modals,
			 depth=depth,
			 root=root,
			 ecs=Vector.tabulate(depth, 
					     fn x => 
					     let val mid = makeAncestorLevel
						 (Vector.sub(ecs, x),
						  Vector.sub(anc, x),
						  Vector.sub(desc, x))
					     in if x = depth - 1 then
						 makeAncestorLevel
						 (mid, Vector.sub(desc,x), Vector.sub(anc,x))
						else mid
					     end)})))
		handle bad => NONE

	    (*
	     * ensure that PATHPT one is properly closer to the
	     * root than PATHPT two -- on the widest-scope modal
	     * dimension
	     *)
	    fun separateTop(P{modals, depth, root, ecs}, 
			    PP(_,anc), PP(_, desc)) =
		((SOME(P{modals=modals,
			 depth=depth,
			 root=root,
			 ecs=justlike(depth, ecs, depth - 1,
				      makePAncestorLevel
				      (Vector.sub(ecs,depth - 1),
				       Vector.sub(anc,depth - 1),
				       Vector.sub(desc,depth - 1)))}))
		handle bad => NONE)

	    (*
	     * ensure that PATHPT one is at the same depth from the
	     * root as PATHPT two -- on the narrowest-scope modal
	     * dimension
	     *)
	    fun matchDepthBottom(P{modals, depth, root, ecs},
				 PP(_,first), PP(_, second)) =
		((SOME(P{modals=modals,
			 depth=depth,
			 root=root,
			 ecs=justlike(depth, ecs, 0,
				      makeSameDepthLevel
				      (Vector.sub(ecs,0),
				       Vector.sub(first,0),
				       Vector.sub(second,0)))}))
		handle bad => NONE)

	end (* local *)
    
	(*
	 * printing
	 *)
	local
	    fun printDetails out (constraints,_,equations) =
		let
		    fun printThing (firstp, NONE) = out "*"
		      | printThing (true, SOME(guy:int)) = out (Int.toString guy)
		      | printThing (false, SOME guy) = out (", " ^ (Int.toString guy))

		    fun printClose (true, true) = out "*)."
		      | printClose (false, true) = out ")."
		      | printClose (true, false) = out "*), "
		      | printClose (false, false) = out "), "

		    fun printPath((V code) :: morev, l) =
			let
			    fun printValuesPath (firstp, nil) = 
				printClose(firstp, true)
			      | printValuesPath(firstp, l as (depth, const) :: morep) =
				if not (Ctree.inCell(constraints, code,
						     code+1, depth)) then
				    (printClose(firstp, false);
				     printPath(morev, l))
				else 
				    (printThing(firstp, const);
				     printValuesPath(false, morep))
			in
			    out ("(" ^ (Int.toString code) ^ " = ");
			    printValuesPath (true, l)
			end
		      | printPath((C code) :: morev, l) =
			(out ((Int.toString code) ^ ", ");
			 printPath(morev,l))
		      | printPath((Root code) :: morev, l) =
			(out "*, ";
			 printPath(morev,l))
		      | printPath(nil, l) = ()

		    fun printEquation (E{leftbase, leftvars, rightend}, n:int) =
			let
			    val path = Ctree.getPath(constraints,
						     leftbase, leftbase+1,
						     rightend, rightend+1)
			in
			    out ("Equation " ^ (Int.toString n) ^ ": ");
			    out ("[" ^ (Int.toString leftbase) ^ "] ");
			    printPath(leftvars, path);
			    out "\n";
			    n+1
			end
		in
		    foldl printEquation 0 equations;
		    Ctree.printProblem out constraints;
		    ()
		end

	    fun writeprettyPath (depth, constraints, root, endpoint, bigname) =
		let
		    fun nameThing (NONE) = "*"
		      | nameThing (SOME(guy:int)) = (Int.toString guy)

		    fun printpathguy nil = ""
		      | printpathguy ((_, const) :: rest) = 
			(("_" ^ (nameThing const)) ^
			 (printpathguy rest))

		    val path = Ctree.getPath(constraints,
					     root, root+1,
					     endpoint, endpoint+1)
		in
		    (Char.toString(chr(ord(if bigname then #"A" else #"a")+depth))) ^ 
		    (printpathguy path)
		end
	in
	    fun printProblem out (P{depth, ecs,...}) =
		let
		    fun pp i = if i = depth then () else 
			(printDetails out (Vector.sub(ecs, i));
			 pp (i+1))
		in
		    pp 0
		end

	    fun splitWriteprettyPathpt (P{depth,root=PP(_,root),ecs,...},bigname) 
		(PP(_,place)) =
		let
		    fun dodim n =
			let
			    val r = meat (hd (Vector.sub(root,n)))
			in
			    writeprettyPath (n,
					     #1 (Vector.sub(ecs,n)),
					     r,
					     (case (Vector.sub(place,n)) of
						  nil => r
						| x::_ => meat x),
						  bigname)
			end
		in
		    map dodim (List.tabulate(depth,fn x => x))
		end

	    fun writeprettyPathpt (p,bigname,commas) pp =
		let
		    fun flat nil = ""
		      | flat [one] = one
		      | flat (two::more) = two ^ (if commas then ", " else " ") ^ (flat more)
		in
		    flat (splitWriteprettyPathpt (p,bigname) pp)
		end

	    fun prettyprintPathpt out system =
		out o (writeprettyPathpt (system, false, false))

	end (* local *)

	(*
	 * buildPaths takes current information:
	 *             problem, vars (list yet to deal with),
	 *             constants (list yet to deal with),  
	 *             index (number of constants seen),
	 *             lindex (index where last var ended=
	 *                     index where this var may start),
	 *        and lists storing previous results:
	 *             vpath= list of records giving the typings 
	 *                  and codes for variables seen so far
	 *             cpath= list of records giving the typings
	 *                  and codes for constants seen so far
	 *             atleast= list giving the depths where
	 *                  each variable's value may start
	 *        and returns a final vpath, cpath, and atleast
	 *            suitable to serve as an argument to 
	 *            Modality.leastMatch.
	 *)

	fun makeInfo (m, i, l) = 
	    {modality=Modspec.index(lookup m), info=i} :: l

	fun buildPaths (_, nil, nil, _, _, vpath, cpath, atleast) =
	    (Modspec.Modality.P (rev vpath),
	     Modspec.Modality.P (rev cpath),
	     rev atleast)

	  (* after all constants, any variable's match is over *)
	  | buildPaths (problem, (V code) :: morev, nil, 
			index, lindex, vpath, cpath, atleast) =
	    buildPaths (problem, morev, nil, index, index, 
			makeInfo(code,code,vpath), cpath, 
			lindex :: atleast)

	  (* if variable splits, match is over; else it extends *)
	  | buildPaths (problem,
			(V code) :: morev, (depth, SOME guy) :: morep,
			index, lindex, vpath, cpath, atleast) =
	    if not (Ctree.inCell(problem, code, code+1, depth)) then
		buildPaths(problem,
			   morev, (depth, SOME guy) :: morep, index, index,
			   makeInfo(code, code, vpath), cpath, 
			   lindex :: atleast)
	    else
		buildPaths(problem,
			   (V code) :: morev, morep, index + 1, lindex, 
			   vpath, makeInfo(guy, (depth, guy), cpath), 
			   atleast)

	  (* if there's no constant, we're in trouble *)
	  | buildPaths (problem, _, (depth, NONE) :: _, _, _, _, _, _) = 
	    (print ("Build paths failed with depth " ^ (Int.toString depth) ^
		    " having no constant value\n");
	     raise bad)

	  (* if there's a mismatch, we're in trouble *)
	  | buildPaths (problem, nil, _, _, _, _, _, _) = 
	    (print ("Build paths failed with variables gone early\n");
	     raise bad)

	  (* here must not have had (V code) on the left: bad *)
	  | buildPaths (problem, _, _, _, _, _, _, _) = 
	    (print ("Build paths failed for constant or root pathelt\n");
	     raise bad)

	(*
	 * relax: readjust the CONSTRAINTS to satisfy the path
	 * constraints specified by EQUATIONS.
	 *
	 * inner function ZAP cycles through the list of equations
	 * for each:
	 *    obtaining the sequence of constants that match
	 *       against each variable
	 *    computing the least correct match above this,
	 *       (if there is one, failing otherwise)
	 *    and adjusting the positions of variables in the
	 *       constraint problem to reflect these matches
	 *       (if possible, failing otherwise)
	 * continuing until either all the equations are satisfied
	 * -- and thus no change is observed on a full cycle -- or
	 * until failure occurs.
	 *
	 * See "Tree Construction" for arguments about why this
	 * is correct and why it terminates.
	 *)

	fun relax inventory (constraints, vltable, equations) =
	    let
		fun zap (nil, constraints) = (SOME constraints, false)
		  | zap (E{leftbase,leftvars,rightend} :: more, constraints) =
		    let
			val (vpath, cpath, atleast) = 
			    buildPaths(constraints, leftvars, 
				       Ctree.getPath(constraints, 
						     leftbase, leftbase+1,
						     rightend, rightend+1),
				       0, 0, nil, nil, nil)

			fun tr(a, SOME(d, g)) = SOME(a, a+1, d, g)
			  | tr _ = NONE
		    in
			case (Modspec.Modality.leastMatch
			      (vpath, cpath, atleast, inventory)) of
			    NONE => 
				(if Debug.debugp "modularity" then
				     print "No matching sequence\n"
				 else
				     ();
				 raise bad)
			  | SOME pairings => 
				case Ctree.moveUp(constraints,
						  List.mapPartial tr pairings) of
				    (_, false) =>
					zap(more, constraints)
				  | change => change
		    end

		fun go (SOME problem, false) = SOME problem
		  | go (SOME problem, true) = go(zap(equations, problem))
		  | go (NONE, _) = NONE
	    in
		case (go (SOME constraints, true)) of
		    NONE => raise bad
		  | SOME newconstraints => (newconstraints, vltable, equations)
	    end		       

	(*
	 * add equation: given modals (overall spec) and inventory
	 *               (spec at this layer), update ecs record
	 *               (constraints, vltable, equations) for this
	 *               layer to ensure that path (leftbase, leftvars)
	 *               matches against the path FROM LEFTBASE TO
	 *               the final constant in rightconsts.  Vars and
	 *               Consts must be new to the problem.
	 *)

	(* shortcut a frequent and easy case *)
	fun addEquation(modals, inventory, 
			(constraints, vltable, equations),
			leftbase, nil, rightbase, nil) =
	    let 
		val i = meat leftbase
		val j = meat rightbase
	    in
		case Ctree.addConstraints(constraints, 
					  Ctree.eq(i,i+1,j,j+1)) of
		    NONE => (raise bad)
		  | SOME p => (p, vltable, equations)
	    end
		     
	  | addEquation (modals, inventory, 
			 (constraints, vltable, equations),
			 leftbase, leftvars, rightbase, rightconsts) =
	    let
		open Modspec.Modality

		(* the constraint to add for variables depends on type *)
		fun varconstraint(modal) = 
		    case Modspec.status(modals, modal) of
			S4_MOD => (Ctree.le)
		      | T_MOD => (Ctree.kidsame)
		      | K4_MOD => (Ctree.lt)
		      | K_MOD => (Ctree.kid)
		      | F_MOD => (Ctree.kid)

		(* for F mods, constants are anonymous, otherwise distinct *)
		fun makeconst(problem, modal, code) =
		    (case Modspec.status(modals, modal) of
			 F_MOD => (Ctree.addAnon
				   (problem, Modspec.index(modal), code))
		       | _ => Ctree.addConstant(problem, code))

		(* add elements for the variables in the left path *)
		fun addleft(nil, parentcode, problem) = (parentcode, problem)
		  | addleft(V code :: more, parentcode, problem) =
		    let
			val const = (varconstraint (lookup code) 
				     (code, code+1, 
				      parentcode, parentcode+1))
		    in
			case Ctree.addElement(problem, code) of
			    NONE => 
				(print "Constraint failure 1\n";
				 raise bad)
			  | SOME p' => case Ctree.addElement(p', code+1) of
				NONE => 
				(print "Constraint failure 2\n";
				 raise bad)
			      | SOME p'' => case (Ctree.addConstraints
						  (p'', const)) of 
				    NONE => 
					(if Debug.debugp "untree" then
					     print "No equality: variables stretch too deep\n"
					 else ();
					 raise bad)
				  | SOME p''' => addleft(more, code, p''')
		    end
		  | addleft _ =
		    (print "Illegal element in left list\n";
		     raise bad)
		    
		(* add elements for the constants in the right path *)
		fun addright(nil, parentcode, problem) = (parentcode, problem)
		  | addright(C code :: more, parentcode, problem) =
		    (case makeconst(problem, (lookup code), code) of
			 NONE => 
			     (print "Constraint failure 4\n";
			      raise bad)
		       | SOME p' => 
			     case (Ctree.addConstraints
				   (p', Ctree.kid(code, code+1, 
						  parentcode, parentcode+1))) of
				 NONE => 
				     (print "Constraint failure 5\n";
				      raise bad)
			       | SOME p'' => addright(more, code, p''))
		  | addright _ = 
		    (print "Illegal element in right list\n";
		     raise bad)

		val leftstart = meat leftbase
		val rightstart = meat rightbase
		val (rightie, p') = addright(rightconsts, rightstart, constraints)
		val (leftie, p'') = addleft(leftvars, leftstart, p')

		(* add the equality *)
		val newProblem = 
		    (case (Ctree.addConstraints
			   (p'', Ctree.eq(leftie, leftie+1,
					  rightie, rightie+1))) of
			 NONE => 
			     (if Debug.debugp "untree" then
				  print "Couldn't add final equality\n"
			      else ();
			      raise bad)
		       | SOME p => 
			     (p, vltable, 
			      (equations @ 
			       [E{leftbase=leftstart, 
				  leftvars=leftvars, 
				  rightend=rightie}])))
	    in
		relax inventory newProblem
	    end

	(* 
	 * equate: update problem by incorporating 
	 * equations identifying endpoints of each of the
	 * passed modal paths at each of the modal
	 * layers of the problem.
	 *)
	fun equate (P{modals, depth, root, ecs},
		    PP(_, left), PP(_, right)) =
	    let
		fun eqn i = 
		    (case (rev (Vector.sub(left, i)), 
			   rev (Vector.sub(right,i))) of
			 (leftbase :: leftvars, rightbase :: rightvars) =>
			     addEquation(modals, Modspec.inventory(modals, i),
					 Vector.sub(ecs, i), 
					 leftbase, leftvars, 
					 rightbase, rightvars)
		       |  _ => raise bad)
	    in
		SOME (P{modals=modals,
			depth=depth,
			root=root,
			ecs=Vector.tabulate(depth, eqn)})
		handle bad => (if Debug.debugp "untree" then
				   print "Result is bad!\n"
			       else ();
			       NONE)
	    end

	(*
	 * test whether the two pathpts happen (right now)
	 * to identify the same pathpt in the constraint
	 * system along each modal dimension.
	 *)

	fun samePathpt (P{modals, depth, root, ecs}, PP(_,p1), PP(_,p2)) =
	    let
		fun same i =
		    i = depth orelse
		    let val (tree,_,_) = Vector.sub(ecs,i)
		    in case (Vector.sub(p1,i),Vector.sub(p2,i)) of
			(nil, nil) => true
		      | (c1::_,c2::_) => 
			    let
				val r = Ctree.sameCell(tree,meat c1,(meat c1)+1,
						       meat c2,(meat c2)+1) 
			    in 
				r andalso same(i+1)
			    end
		      | _ => false
		    end
	    in
		same 0
	    end

	(*
	 * updateVtables.  
	 * At each level but the last, get all the variables
	 * that extend each constant, and make sure that the
	 * indices of the variables at the NEXT level ALSO
	 * extend the indices of the constants there.
	 *)
	fun updateVtables(p as P{depth=0,...}) = SOME p
	  | updateVtables(p as P{depth=1,...}) = SOME p
	  | updateVtables(P{modals, depth, root, ecs}) =
	    let
		(* if g is real, make j closer than g to the root *)
		fun mkcs j (g, l) =
		    case peekup g of
			NONE => l
		      | SOME stop => 
			    let val i = meat stop
			    in Ctree.le(i,i+1,j,j+1) @ l 
			    end

		(* if i is real, make it closer to root than newguys *)
		fun makeconstraints(i, (_, newguys), constraints) =
		    case peekup i of
			NONE => constraints
		      | SOME start => (foldr (mkcs (meat start))
				       constraints newguys)

		(* 
		 * generate and add constraints on this level from 
		 * the last one, then, if necessary, recompute this
		 * level for use in the next iteration.
		 *)
		fun updateTables (d, prevtable, sofar) =
		    let val (p, t, e) = Vector.sub(ecs, d) 
			val constraints = (Ctree.IntMap.fold makeconstraints
					   prevtable nil)
		    in case Ctree.addConstraints(p, constraints) of
			NONE => NONE
		      | SOME p' => 
			    let val fpte as (fp, ft, fe) = 
				relax (Modspec.inventory(modals,d)) (p', t, e)
			    in if d=depth-1 then 
				SOME(Vector.fromList(rev (fpte :: sofar)))
			       else 
				   let val ntable = Ctree.updateVLessStats(fp, ft)
				   in updateTables(d+1, ntable, (fp, ntable, fe)::sofar)
				   end
			    end handle bad => NONE
		    end
	    in
		let val (p1, t1, e1) = Vector.sub(ecs,0)
		    val ntable = Ctree.updateVLessStats(p1, t1)
		in case updateTables(1, ntable, [(p1, ntable, e1)]) of
		    NONE => NONE
		  | SOME newecs => SOME (P{modals=modals,
					   depth=depth,
					   root=root,
					   ecs=newecs})
		end
	    end    

	val _ = 
	    app Debug.register
	    [ ("modularity", 
	       "Report path checking failures for modal matches\n", 
	       false),
	      ("untree", 
	       "Report tree construction failures for modal matches\n", 
	       false) ]
	     
    end
