
signature MATCHBANK =
    sig
	structure Term : TERM
	structure Patheq : PATHEQ
	sharing Term.Patheq = Patheq

	type 'a matchelt
	datatype 'a match = 
	    M of {trans: Term.init Term.IntMap.intmap,
		  data: 'a}
	    
	val makeMatchelt: Term.term * 'a -> 'a matchelt
	val match: Term.term -> 'a matchelt -> 'a match option

	val readMatchelt: ((string list -> 'a option * string list) ->
			   string list -> 'a matchelt option * string list)
	val printMatchelt: (('a -> unit) * (string -> unit) ->
			    'a matchelt -> unit)
    end

functor Matchbank(structure Term : TERM;
		  structure Patheq : PATHEQ;
		  sharing Term.Patheq = Patheq) : MATCHBANK =
    struct
	structure Term = Term
	structure Patheq = Patheq

	datatype 'a matchelt = 
	    E of {head: Term.term,
		  fvs: Term.IntSet.intset,
		  data: 'a}

	datatype 'a match =
	    M of {trans: Term.init Term.IntMap.intmap,
		  data: 'a}

	fun makeMatchelt (term, data) =
	    E{head = term,
	      fvs = Term.freeVarsTerm (term, Term.IntSet.empty),
	      data = data}

	fun readMatchelt re s =
	    (case Term.readTerm s of
		 (SOME head, ":" :: rest) =>
		     (case re rest of
			  (SOME data, more) =>
			      (SOME (makeMatchelt(head, data)), more)
			| _ => (NONE, s))
	       | _ => (NONE, s))

	fun printMatchelt (display, out) (E{head,data,...}) =
	    (Term.printTerm out head;
	     out ": ";
	     display data)

	local 
	    val problem = Patheq.newProblem(Patheq.Modspec.newCollection());
	    val root = Patheq.rootProblem problem;
	    val zip = Term.IntMap.empty() : Term.init Term.IntMap.intmap;
	    val init = Term.newSystem(problem)
		
	    fun makeInTrans (fvs) =
		let
		    fun et (i, tr) =
			let
			    val var = Term.newVar()
			in
			    Term.IntMap.insert(tr, i, Term.VI(var,root))
			end
		in
		    Term.IntSet.fold et fvs zip
		end
	    
	    fun makeOutTrans (system, fvs) =
		let
		    fun et (inv, Term.VI(out, _), tr) =
			(case Term.readoutVar (system, SOME out) of
			     Term.Miss => tr
			   | Term.Free => 
				 Term.IntMap.insert(tr, inv,
						    Term.VI(out, root))
			   | Term.Bound t => 
				 Term.IntMap.insert(tr, inv, 
						    Term.TI(out, root, t)))
		      | et (_, _, tr) = tr
		in
		    Term.IntMap.fold et fvs zip
		end
	in
	    fun match to (E{head, fvs, data}) =
		if not (Term.sameHead (to, head) andalso 
			Term.quickMatch(to, zip, head, zip, init)) then
		    NONE
		else
		    let
			val trans = makeInTrans (fvs)
			val (newsystem, _) =
			    Term.unify (Term.updateSystem
					 (to, Term.IntMap.empty(), root,
					  head, trans, root, problem, init))
		    in
			SOME (M {trans = makeOutTrans (newsystem, trans),
				 data = data})
		    end
		handle Term.Clash => NONE
		     | Term.Cycle => NONE
		     | Term.Scope => NONE
	end		
    end
