(*
 * Matthew Stone
 * Dialup 0.1 - 3/96
 * tree-sig.sml
 *
 * Signature file for building possible-worlds models for
 * modal logic, as described in "Tree Construction for
 * Reasoning about Necessity".  
 *
 * "elements" -- represented as integers -- correspond to
 * leaves in the tree.  Possible worlds correspond to the
 * least common ancestors of pairs of leaves.
 *
 * You can specify constraints of three forms:
 *    (k1,k2) is a descendant of (p1,p2) -- meaning that
 *        (p1,p2) is closer to the root than (k1,k2)
 *    (nk1,nk2) is not a descendant of (np1,np2) -- which
 *        means that (np1,np2) is not closer to the root
 *    (h1,h2) ccommands (g1,g2) -- meaning that (g1,g2) is
 *        a descendant of the parent of (h1,h2) -- 
 *        borrowing from the linguistic description of
 *        ccommand, h stands for head, g for governed.
 *
 * When such constraints are added, the tree will be 
 * realigned so that these constraints are satisfied,
 * if possible.
 *
 * The translation of proofs calls for relating possible
 * worlds corresponding to constants by the notdesc
 * relation.  An optimization allows you to sidestep
 * the explicit declaration of these notdesc constraints,
 * by declariing (i,i+1) to be a new constant.  For
 * functional modalities, constants may be identical,
 * so (i,i+1) can be declared as an anonymous constant.
 *
 * To allow for multiple levels of modalities (with an
 * increasing domain constraint between levels), you
 * can compute the set of variables further from the
 * root than each constant (and the set of variables 
 * which have gotten this property since the last time
 * you checked).  In the increasing domain setup, the
 * modal indices of the variables at the next level
 * must extend the modal indices of smaller constants.
 *
 * Invariant: elements should be added to a problem
 * in *increasing* order, in order to have correct
 * behavior.
 *)

signature CONSTRAINT_TREE =
    sig
	structure IntMap : MYINTMAP

	datatype constraint =
	    Descendant of {p1:int, p2:int, k1:int, k2:int}
	  | Notdesc of {np1:int, np2:int, nk1:int, nk2:int}
	  | Ccommand of {h1:int, h2:int, g1:int, g2:int}
	  | NotShallower of {ns1:int, ns2:int, nd1:int, nd2:int}

	type problem 

	(* 
	 * we keep statistics on the number of operations
	 * performed 
	 *)
	val resetStats : unit -> unit
	val printStats : (string -> unit) -> unit

	val newProblem : bool -> problem
	val printProblem : (string -> unit) -> problem -> unit
	val addConstraints : problem * constraint list -> 
	    problem option
	val printConstraint : (string -> unit) -> constraint -> unit

	val addElement : problem * int -> problem option
	val addConstant : problem * int -> problem option
	val addAnon : problem * int * int -> problem option

	(*
	 * test the relationship between things:
	 *  in cell tests whether (i,j) has depth d or more
	 *  same cell tests whether (i,j) and (k,l) 
	 *    identify the same node in the tree
	 *)
	val inCell : problem * int * int * int -> bool
	val sameCell : problem * int * int * int * int -> bool

	(*
	 * report the path in the tree from (1,2) at the
	 * root end to (3,4) at the leaf end;
	 * the path is given as a list of pairs whose 
	 * first element is the depth in the tree and
	 * whose second element is the constant label
	 * of the appropriate node, if any.
	 *
	 * allows an equation checker to determine whether 
	 * paths in the tree match their type specifications.
	 *)
	val getPath : problem * int * int * int * int -> 
	    (int * int option) list

	val getConst : problem * int * int -> int option

	(*
	 * the list contains (variable, constant) pairs 
	 * indicating the least type-consistent match for
	 * a path of variables.  the function updates
	 * the problem, if necessary (and possible), the
	 * boolean returns whether any change was made.
	 *)
	val moveUp : problem * (int * int * int * int) list -> 
	    problem option * bool

	(*
	 * compute and update a map associating with each
	 * constant i, a list of non-constant elements j
	 * such that the leaf j is a descendanat of (i,i+1)
	 * and a list of those such j recognized for the
	 * first time on this iteration.
	 *)
	val makeVLessStats : problem -> (int list * int list) IntMap.intmap
	val updateVLessStats : problem * (int list * int list) IntMap.intmap ->
	    (int list * int list) IntMap.intmap

	(*
	 * higher level interface for creating constraints:
	 * these functions assemble the collection of
	 * constraints required to guarantee some property.
	 *)
	val notle : int * int * int * int -> constraint list
	val le : int * int * int * int -> constraint list
	val lt : int * int * int * int -> constraint list
	val eq : int * int * int * int -> constraint list
	val kid : int * int * int * int -> constraint list
	val kidsame : int * int * int * int -> constraint list
	val notsh : int * int * int * int -> constraint list
	val eqdp : int * int * int * int -> constraint list
    end
