(*
 * Matthew Stone
 * Dialup 0.1 - 3/96 - revised for sml110 5/98
 * tree.sml
 *
 * Implementation of the CONSTRAINT_TREE signature.
 * (Everything is represented with applicative
 * data structures!)
 *
 * The basic algorithm is roughly as follows:
 *  Maintain the tree as a table of LEVELS.  Each
 *  level is made up of CELLS corresponding to 
 *  nodes in the tree (at the depth corresponding
 *  to the level).  These cells are merged to
 *  satisfy constraints.
 * (See 'tree construction' paper for details and
 *  justifications.)
 *
 * This file implements an optimization in the handling
 * of constants.  Constants are kept distinct by 
 * imposing an order between pairs of constants (i,i')
 * and (j,j') with i<j if i is introduced earlier than j
 * in proof search.  Corresponding to i<j is a constraint
 * (i,i') cannot be a descendant of (j,j').
 * (There are n^2 such constraints, which would have
 * to be maintained but which wouldn't often fire.)
 *
 * The implementation of the notless constraint is that
 * if i,i',j and j' are in the same cell, j and j' must
 * go together at the next level.  So there is at most
 * one constant for each cell, and this constant is the
 * least constant in the cell.  
 *
 * So the optimization is to keep track of the least
 * constant at each node explicitly.  When cells are 
 * merged, we only have to demote the greater least
 * constant of the two maintained.
 *
 * An added complication arises in the case of 
 * 'functional' modalities: modalities where constants
 * are not necessarily distinct (because each world
 * has a unique other accessible to it).  These are
 * called 'anonymous' constants here---since in other
 * respects they still behave like constants. 
 * Anonymous constants are represented as a cluster
 * in each cell's record of (least) distinct keys, 
 * and will only be demoted as a group by a real
 * constant (or a smaller, incompatible anonymous cell).
 *)

functor ConstraintTree (structure IntMap : MYINTMAP;
			structure IntSet : MYINTSET;
			structure Fifo : INTFIFO) : CONSTRAINT_TREE =
    struct
	structure IntMap = IntMap

	(***********************************************************
	 *
	 * DATATYPES
	 *
	 ***********************************************************)

	(*
	 * The constraints managed by the algorithm take
	 * one of three forms:
	 * Descendant(p1,p2,k1,k2):
	 *     (p1,p2) is on a path from the root to (k1,k2)
	 * Notdesc(np1,np2,nk1,nk2):
	 *     (np1,np2) is NOT on a path from the root to (nk1,nk2)
	 * Ccommand(h1,h2,g1,g2):
	 *     The parent of (h1,h2) is on a path from the root to
	 *     (g1,g2)
	 * NotShallower(ns1,ns2,nd1,nd2):
	 *     (ns1,ns2) lies no closer to the root of the tree
	 *     [measuring depth only] than does (nd1,nd2).
	 *)

	datatype constraint =
	    Descendant of {p1:int, p2:int, k1:int, k2:int}
	  | Notdesc of {np1:int, np2:int, nk1:int, nk2:int}
	  | Ccommand of {h1:int, h2:int, g1:int, g2:int}
	  | NotShallower of {ns1:int, ns2:int, nd1:int, nd2:int}

	(*
	 * ldk: 'least distinct key'
	 * used to keep track of a canonical name of a cell
	 * (see below).  The least distinct key gives the
	 * first constant contained in a cell.  Because later
	 * constants will be pushed down away from the root
	 * to enforce distinctness, the ldk of a cell is the
	 * only constant i for which the cell could be (i,i+1).
	 * This is complicated by anonymous constants: many
	 * that share some type could match.  
	 * ANON(type, least, all) has LEAST as the index of
	 * the least constant --- of any sort --- TYPE is the
	 * type of this constant, and ALL is the inventory of
	 * all anonymous constants with type TYPE in the cell.
	 *)

	datatype ldk = 
	    DK of int 
	  | ANON of int * int * int list

	(*
	 * cell: record of a node in the tree,
	 *       with dynamic info to allow cells to merge.
	 * index: a key integer identifying the CELL
	 * distinctkeys: the set of leaves contained in the
	 *    subtree rooted at the CELL that have been marked
	 *    as constants.
	 * varkeys: the rest of the leaves contained in the
	 *    subtree rooted at the CELL.
	 * leastdk: the smallest of the constants contained
	 *    in the CELL, if any, recorded as described 
	 *    above.
	 * rulesize: the number of constraints which may have
	 *    to trigger when the CELL is merged with another.
	 * rules: the constraints that may have to apply 
	 *    when the CELL is merged with another.
	 *)

	datatype cell =
	    Cell of {index:int, 
		     distinctkeys: int list,
		     varkeys: int list,
		     leastdk: ldk option,
		     rulesize:int,
		     rules:constraint list}

	(*
	 * level: record of the nodes at a set DEPTH in the tree.
	 * nexcell: index to identify next cell created
	 * asgn: map from leaves to the cell (index) that
	 *   contains them
	 * cells: map from indices to cell structures
	 *)

	datatype level =
	    Level of {depth:int,
		      nextcell: int,
		      asgn: int IntMap.intmap,
		      cells: cell IntMap.intmap}

	(*
	 * constraint_table
	 *  contains CONSTRAINTS, a map from leaf codes to
	 *  the constraints that may fire when a cell 
	 *  containing that leaf is merged with another.
	 *)

	datatype constraint_table =
	    Table of {constraints:constraint list IntMap.intmap}

	(*
	 * qe: queue element
	 *   an instruction to merge the cell containing K1
	 *   and the cell containing K2 at DEPTH 
	 *)

	datatype qe =
	    Merge of {depth:int, k1:int, k2:int}

	(*
	 * problem:
	 *    the complete record of an effort to build the tree
	 * size: an estimate of the size of the resulting tree
	 *    (for logic problems, this is the number of distinct
	 *    constants)
	 * cvtype: whether all constraints will refer to some
	 *    constant pair or variable pair.
	 * distinctkeys: the first leaf of every pair (i,i') denoting
	 *    a constant that must stay distinct
	 * anonkeys: the first leaf of every pair (i,i') denoting
	 *    an anonymous constant (mapped to the constant's type)
	 * table: the constraints maintained in the problem, 
	 *    organized by the leaves to which they refer.
	 * levels: a record of the tree, organized by depths.
	 *)

	datatype problem =
	    P of {size:int,
		  cvtype: bool,
		  distinctkeys: IntSet.intset,
		  anonkeys: int IntMap.intmap,
		  table:constraint_table,
		  levels:level IntMap.intmap}

	(***********************************************************
	 *
	 * DEBUGGING, HELPER AND DIAGNOSTIC FUNCTIONS
	 *
	 ***********************************************************)

	(*
	 * inconsistency: make sure ASGN and CELL agree on 
	 *  what elements CELL contains.
	 *)

	fun inconsistency(asgn, Cell{index, varkeys, distinctkeys,...}) =
	    let
		fun checkto (x,y,r) = r orelse
		    if y = index then
			not ((List.exists (fn z => z=x) varkeys) orelse
			     (List.exists (fn z => z=x) distinctkeys))
		    else
			false
		fun checkfrom x =
		    not (IntMap.peek(asgn, x) = SOME index)
	    in
		if (IntMap.fold checkto asgn false) orelse
		    (List.exists checkfrom distinctkeys) orelse
		    (List.exists checkfrom varkeys) then
		    (print ("asgn does not match cell with index " ^
			    (Int.toString index) ^ "\n");
		     print "Elts are: ";
		     app (fn x => print ((Int.toString x) ^ " ")) (varkeys @ distinctkeys);
		     print "\nAsgn has: ";
		     (IntMap.app (fn (x,y) => 
				  (if y = index then
				       (print (Int.toString x);
					print " ")
				   else ()))
		      asgn);
		     raise Fail "")
		else
		    ()
	    end

	(*
	 * this submodule maintains global counters of tree operations
	 * performed -- to get a sense of how much work is involved
	 * in a problem.
	 *)

	local 
	    val merges = ref 0
	    val enqueues = ref 0
	    val constraints = ref 0
	in
	    fun incmerge () = (merges := !merges + 1)

	    fun inccstr () = (constraints := !constraints + 1)

	    fun enqueue (q, e as Merge {depth,...}) =
		(enqueues := !enqueues + 1;
		 Fifo.enqueue (q, e, depth))

	    fun resetStats () = 
		(merges := 0;
		 enqueues := 0;
		 constraints := 0)

	    fun printStats (out) =
		(out ("Tree constraint statistics:\nMerges: " ^
		      (Int.toString (!merges)) ^
		      "\nEnqueues: " ^
		      (Int.toString (!enqueues)) ^
		      "\nConstraints: " ^
		      (Int.toString (!constraints)) ^
		      "\n"))

	end (* local *)
	       

	fun printConstraint out (Descendant {p1, p2, k1, k2}) =
	    out ("Descendant: parent(" ^ (Int.toString p1) ^ "," ^
		 (Int.toString p2) ^") child(" ^ (Int.toString k1) ^ "," ^
		 (Int.toString k2) ^ ")\n")
	  | printConstraint out (Notdesc {np1, np2, nk1, nk2}) =
	    out ("nondescendant: nonparent(" ^ (Int.toString np1) ^ "," ^
		 (Int.toString np2) ^") nonchild(" ^ (Int.toString nk1) ^ "," ^
		 (Int.toString nk2) ^ ")\n")
	  | printConstraint out (Ccommand {h1, h2, g1, g2}) = 
	    out ("ccommand: head(" ^ (Int.toString h1) ^ "," ^
		 (Int.toString h2) ^") governed(" ^ (Int.toString g1) ^ "," ^
		 (Int.toString g2) ^ ")\n")
	  | printConstraint out (NotShallower {ns1, ns2, nd1, nd2}) =
	    out ("notshallower: nonshallower(" ^ (Int.toString ns1) ^ "," ^
		 (Int.toString ns2) ^") nondeeper(" ^ (Int.toString nd1) ^ "," ^
		 (Int.toString nd2) ^ ")\n")

	fun printProblem out = 
	    let
		fun printCell (_,Cell{varkeys,distinctkeys,...}) =
		    (out "( ";
		     app (fn x => out ((Int.toString x) ^ " ")) distinctkeys;
		     app (fn x => out ((Int.toString x) ^ " ")) varkeys;
		     out ")  ")

		fun printLevel (_,Level{depth,cells,...}) =
		    (out ((Int.toString depth) ^ ": ");
		     IntMap.app printCell cells;
		     out "\n")
	    in 
		fn (P{levels,table=Table{constraints},...}) => 
		(IntMap.app printLevel levels)
	    end

	(*
	 * isAlive:
	 *   Following ASSU (and for now, the tree construction paper),
	 *   the only way to recognize that a problem has no solution
	 *   is to merge at too deep a level.  Here, we do a better
	 *   test, to be called when a queue empties out at a particular
	 *   level in the tree.  Thereafter, no interesting new merges
	 *   will be posted at the previous level, so if cells at this
	 *   level duplicate cells at their parents' level, whatever
	 *   constraints applied to yield this duplication can be 
	 *   applied all the way down the tree --- so there can be no
	 *   solution.
	 *)

	local
	    (* make sure every cell divides *)

	    fun test (Level{cells,...}, Level{asgn,...}) =
		let 
		    fun different cell x = not(cell = IntMap.peek(asgn,x))

		    fun testcell (Cell{distinctkeys=nil,varkeys=nil,...}) = true
		      | testcell (Cell{distinctkeys=[_],varkeys=nil,...}) = true
		      | testcell (Cell{distinctkeys=nil,varkeys=[_],...}) = true
		      | testcell (Cell{distinctkeys=h::t,varkeys=e,...}) =
			(case IntMap.peek(asgn, h) of
			     NONE => true
			   | cell => 
				 (List.exists (different cell) t) orelse
				 (List.exists (different cell) e))
		      | testcell (Cell{distinctkeys=e,varkeys=h::t,...}) =
			(case IntMap.peek(asgn, h) of
			     NONE => true
			   | cell => 
				 (List.exists (different cell) t) orelse
				 (List.exists (different cell) e))
		in
		    IntMap.fold (fn (_,x,b) => b andalso testcell x) cells true
		end
	in
	    fun isAlive (P{levels,...}, testdepth, queue) =
		case Fifo.howEmpty (queue) of
		    Fifo.Empty => (testdepth, true)
		  | Fifo.EmptyUpTo d => 
			((if d <= testdepth+2 then (testdepth, true)
			  else (d-2,
				test(IntMap.find(levels, d-3), 
				     IntMap.find(levels, d-2))))
			      handle IntMap.NotFound =>
				  (raise (Fail ("isAlive (ftree): depth " ^
						(Int.toString (d-3)) ^
						" not found"))))
	end (* local *)

	(*
	 * levels (and cells) are created lazily -- until
	 * you want to put things together in them, they 
	 * don't even exist.  
	 * newLevel realizes a level, and getLevel 
	 * arranges to call this transparently when
	 * needed.  newProblem is here because it needs
	 * level zero (corresponding to the root of the tree).
	 *)

	local
	    fun newLevel (P{size, cvtype, distinctkeys,
			    anonkeys, table, levels}, 
			  depth) =
		let
		    val newL = Level{nextcell=0, depth=depth,
				     asgn = IntMap.empty(),
				     cells = IntMap.empty()}
		in
		    (P{size=size, cvtype = cvtype, distinctkeys=distinctkeys, 
		       anonkeys=anonkeys, table=table, 
		       levels = IntMap.insert(levels, depth, newL)},
		     newL)
		end
	in
	    fun newProblem (cvtype) =
		#1 (newLevel (P{size=if cvtype then 1 else 0,
				cvtype=cvtype,
				distinctkeys=IntSet.empty,
				anonkeys=IntMap.empty(),
				table = Table{constraints=IntMap.empty()},
				levels = IntMap.empty()},
			      0))

	    fun getLevel (p as P{levels,...}, depth) =
		case IntMap.peek (levels, depth) of
		    NONE => newLevel(p, depth)
		  | SOME thing => (p, thing)
	end

	(*
	 * makeCell -- makes a cell with index INDEX, 
	 *    to contain element ELT.  
	 * The constraint table is used to figure what
	 *    rules to have in the cell representation,
	 * The DISTINCT- and ANONKEYS are used to figure
	 *    out how to initialize the cells distinctness
	 *    fields.
	 *)
	fun makeCell (index, elt, Table{constraints}, 
		      distinctkeys, anonkeys) =
	    let
		val rules =
		    case IntMap.peek(constraints, elt) of
			NONE => nil
		      | SOME list => list

		val thisdk = 
		    if IntSet.member(distinctkeys, elt) then
			SOME (DK elt)
		    else case IntMap.peek(anonkeys, elt) of
			SOME sort => SOME (ANON(sort, elt, [elt]))
		      | NONE => NONE
	    in
		Cell {index = index,
		      rulesize = length rules,
		      rules = rules,
		      leastdk = thisdk,
		      distinctkeys = (case thisdk of SOME _ => [ elt ] | _ => nil),
			  varkeys = (case thisdk of SOME _ => nil | _ => [ elt ])}
	    end

	(*
	 * together: are A and B located in the same cell,
	 * according to ASGN.
	 *)
	fun together (asgn, a, b) =
	    (case IntMap.peek(asgn,a) of
		 NONE => false
	       | t => t = IntMap.peek(asgn,b))

	exception Inconsistent

	(*
	 * check: (ASGN, DEPTH, MAXDEPTH) ->
	 *        (CONSTRAINT, QUEUE)
	 * update the queue to include the appropriate
	 * merger, if necessary, triggered by CONSTRAINT
	 * under the passed assignment.
	 *)
	fun check (asgn, depth, maxdepth) =
	    let 
		fun join (j, k1, k2, queue) =
		    if not (j < maxdepth) then
			(raise Inconsistent)
		    else
			if (j > 0) then
			    enqueue(queue, Merge{depth=j, k1=k1, k2=k2})
			else
			    queue

		(* if p1 & p2 are in the same cell, all must be *)
		fun try (Descendant{p1,p2,k1,k2}, queue) =
		    if together(asgn,p1,p2) then
			SOME (join(depth, k1, p1,
				   join(depth, k2, p1, queue)))
		    else
			NONE
			
		  (* if all now in the same cell, np1 & np2 must be at next *)
		  | try (Notdesc{np1,np2,nk1,nk2}, queue) =
		    if  together(asgn,np1,np2) andalso
			together(asgn,nk1,nk2) andalso
			together(asgn,np1,nk1)  then
			SOME (join(depth+1, np1, np2, queue))
		    else
			NONE

		  (* if nd1 and nd2 now together, ns1 and ns2 must be *)
		  | try (NotShallower{ns1,ns2,nd1,nd2}, queue) =
		    if  together(asgn, ns1, ns2) then
			SOME(join(depth, nd1, nd2, queue))
		    else
			NONE

		  (* if h1 & h2 now together, all must be at previous *)
		  | try (Ccommand {h1,h2,g1,g2}, queue) =
		    if together (asgn,h1,h2) then
			SOME (join(depth-1,g1,h1,
				   join(depth-1,g2,h1,queue)))
		    else
			NONE
	    in
		try
	    end
	    
	(*
	 * eraseCell -- remove the cell at INDEX in LEVEL,
	 * provided the third arg, a boolean indicating
	 * whether the cell was added to the level in the
	 * first place, is TRUE.
	 *)

	fun eraseCell (Level{nextcell, asgn, cells, depth}, index, true) =
	    (Level{nextcell = nextcell,
		   depth = depth,
		   asgn = asgn,
		   cells = #1 (IntMap.remove(cells, index))})

	  | eraseCell (level, index, false) = level

	(* 
	 * record in the problem structure the fact that
	 * the passed level should now have the given 
	 * assignment NEWASGN (mapping leaves to cells) and
	 * should have NEWCELL as its cell at INDEX
	 *)

	fun updateLevel(p as P{table,size,cvtype,distinctkeys,anonkeys,levels},
			Level{nextcell, asgn, cells, depth},
			newasgn, index, newCell) =
	    let val newCells = IntMap.insert(cells, index, newCell)
		val _ =  ()
	    in  
		incmerge();
		(* inconsistency(newasgn, newCell); *)
		P{table = table,
		  size = size,
		  cvtype = cvtype,
		  distinctkeys = distinctkeys,
		  anonkeys = anonkeys,
		  levels = IntMap.insert(levels, depth,
					 Level{nextcell = nextcell + 1,
					       depth = depth,
					       asgn = newasgn,
					       cells = newCells})}
	    end

	local

	    (*
	     * cmpDk compares two records of the least distinct
	     * keys of cells.  It constructs the distinct key
	     * needed for the merger of the two cells, and
	     * enqueues the old least distinct key, if necessary.
	     *)

	    fun lower (dk1, depth, queue) =
		(foldr (fn (i,q) => 
		       (enqueue (q, Merge{depth=depth, k1=i, k2=i+1})))
		 queue
		 dk1)

	    fun cmpDk (DK v1, DK v2, depth, queue) = 
		if v1 > v2 then
		    (SOME(DK v2), lower([v1], depth, queue))
		else
		    (SOME(DK v1), lower([v2], depth, queue))
	      | cmpDk (DK v1, a as ANON(_,v2,guys), depth, queue) = 
		if v1 > v2 then
		    (SOME a, lower([v1], depth, queue))
		else
		    (SOME(DK v1), lower(guys, depth, queue))
	      | cmpDk (a as ANON(_, v1,guys), DK v2, depth, queue) =
		if v1 > v2 then
		    (SOME (DK v1), lower(guys, depth, queue))
		else 
		    (SOME a, lower([v1], depth, queue))
	      | cmpDk (a1 as ANON(t1, v1, k1), 
		       a2 as ANON(t2, v2, k2), depth, queue) = 
		if t1 <> t2 then
		    if v1 > v2 then
			(SOME a2, lower(k1, depth, queue))
		    else
			(SOME a1, lower(k2, depth, queue))
		else 
		    (SOME(ANON(t1, if v1 > v2 then v2 else v1, k1 @ k2)),
		     queue)
		
	    fun bumpDown (depth, maxdepth, NONE, leastdk, queue) =
		(leastdk, queue)
	      | bumpDown (depth, maxdepth, leastdk, NONE, queue) =
		(leastdk, queue)
	      | bumpDown (depth, maxdepth, SOME dk1, SOME dk2, queue) =
		if not (depth+1 < maxdepth) then
		    (raise Inconsistent)
		else cmpDk (dk1, dk2, depth+1, queue)

	    (* update ASGN so that GUYS have NINDEX *)
	    fun reassign nindex guys asgn = 
		foldr (fn (x,a) => IntMap.insert(a,x,nindex)) asgn guys

	    (*
	     * doMerge (problem, level, asgn, depth, maxdepth, queue)
	     *         (c1, c1real, c2, c1real)
	     * merge cells C1 and C2 (located at LEVEL in PROBLEM).
	     * First, you want to be sure that you process the one
	     *   with the smaller constraint list (to take advantage
	     *   of unioning NlogN complexity), which is done by the
	     *   initial test.
	     * Then, you need to build a new assignment by 
	     *   reassigning keys from the little one to the big one.
	     *   If the big one is virtual (c1real is false), then 
	     *   you need to record its keys also.
	     * Next, you trigger all of the constraints in the smaller
	     *   cell, using CHECK.
	     * Finally, you bump down the constants, build the 
	     *   new record of (least) distinct keys, and construct
	     *   the overall new cell you want.
	     *)
	    fun doMerge (p, level, asgn, depth, maxdepth, queue) =
		let
		    fun dom (c1 as Cell big, c1real, c2 as Cell little, c2real) =
			if #rulesize big < #rulesize little then
			    dom(c2, c2real, c1, c1real)
			else
			    let
				val nindex = #index big

				val _ = ()

				val nasgn = 
				    (reassign nindex (#distinctkeys little)
				     (reassign nindex (#varkeys little)
				      (if c1real then 
					  asgn 
				       else
					   (reassign nindex (#distinctkeys big)
					    (reassign nindex (#varkeys big) asgn)))))

				fun mergerules (nil, queue, atend) = (queue, atend)
				  | mergerules (r::rules, queue, atend) =
				    case check(nasgn, depth, maxdepth)(r, queue) of
					NONE =>
					    mergerules(rules, queue, r::atend)
				      | SOME nq =>
					    mergerules(rules, nq, atend)

				val (rulequeue, newrules) =
				    mergerules(#rules little, queue, #rules big)
				val (newldk, newqueue) =
				    bumpDown (depth, maxdepth, 
					      #leastdk big, #leastdk little,
					      rulequeue)
				val newdks = (#distinctkeys little) @ (#distinctkeys big) 
			    in
				(updateLevel
				 (p, 
				  eraseCell(level, #index little, c2real),
				  nasgn, nindex,
				  Cell{index = nindex,
				       distinctkeys = newdks,
				       leastdk = newldk,
				       varkeys = (#varkeys little) @ (#varkeys big),
				       rulesize = length newrules,
				       rules = newrules}),
				 newqueue)
			    end
		in
		    dom
		end
	in
	    (*
	     * process: problem * int * queue ->
	     *          problem * int * queue
	     *
	     * Get the next entry, and construct (or retrieve) the
	     * appropriate level and merge operation for the merge.
	     * Based on whether cells exist for the two points
	     * being merged (and whether they are already the same)
	     * apply the appropriate merge.
	     * Finally, if merge goes through and ISALIVE says
	     * keep going, process the new queue, else die.
	     *)

	    fun process (p, testdepth, queue) =
		let
		    val (nqueue, Merge{depth, k1, k2}) = Fifo.dequeue(queue)
		    val (np as P{table, distinctkeys=dk, anonkeys=ak, size,...}, 
			 l as Level{asgn,nextcell,cells,...}) = getLevel(p, depth)

		    (* build closure of curried function *)
		    val merge = doMerge(np, l, asgn, depth, size, queue)

		    fun safefind (k,c) =
			IntMap.find(cells, c) 
			handle IntMap.NotFound =>
			    raise (Fail ("Process (ftree): " ^
					 "No entry for cell " ^
					 (Int.toString c) ^ " at depth " ^
					 (Int.toString depth) ^ " associated with var " ^
					 (Int.toString k)))

		    val (fp, fqueue) =
			case (IntMap.peek(asgn, k1), IntMap.peek(asgn, k2)) of
			    (NONE, NONE) =>
				merge(makeCell (nextcell, k1, table, dk, ak), false,
				      makeCell (nextcell+1, k2, table, dk, ak), false)
			  | (SOME c, NONE) =>
				merge(safefind (k1, c), true,
				      makeCell (nextcell, k2, table, dk, ak), false)
			  | (NONE, SOME c) =>
				merge(makeCell (nextcell, k1, table, dk, ak), false,
				      safefind (k2, c), true)
			  | (SOME c1, SOME c2) =>
				if c1 = c2 then
				    (np, nqueue)
				else
				    merge(safefind (k1, c1), true,
					  safefind (k2, c2), true)
		in
		    case isAlive(fp, testdepth, fqueue) of
			(fdepth, true) =>
			    process (fp, fdepth, fqueue)
		      | _ => 
			    (NONE)
		end 
	    handle Fifo.Dequeue => SOME p
		 | Inconsistent => 
		       (NONE)
	end

	(***********************************************************
	 *
	 * USER LEVEL FUNCTIONS
	 *
	 ***********************************************************)

	(* are A and B in the same cell at DEPTH in P? *)
	fun inCell (P{levels,...}, a, b, depth) =
	    (case IntMap.peek(levels, depth) of
		 NONE => false
	       | SOME (Level{asgn,...}) => together(asgn, a, b))

	(* is (K1,K2) a descendant of (P1,P2) in P? *)
	fun isDescendant(P{levels,...}, p1, p2, k1, k2) =
	    let
		fun try d = 
		    case IntMap.peek(levels,d) of
			NONE => true
		      | SOME (Level{asgn,...}) =>
			    if together(asgn, p1, p2) then
				together(asgn, p1, k1) andalso
				together(asgn, p1, k2) andalso
				try (d+1)
			    else
				true
	    in
		try 1
	    end

	(* does (A,B) denote the same node in P as (C,D)? *)
	fun sameCell (P{levels,...},a,b,c,d) =
	    let
		fun samenow (_, Level{asgn,...}, SOME answer) = SOME answer
		  | samenow (j, Level{asgn,...}, NONE) =
		    let
			val one = IntMap.peek(asgn,a)
			val two = IntMap.peek(asgn,b)
			val three = IntMap.peek(asgn,c)
			val four = IntMap.peek(asgn,d)
		    in
			if (not (one = two) orelse one = NONE) then
			    if (not (three = four) orelse three = NONE) then
				SOME true
			    else
				SOME false
			else 
			    if one = three andalso one = four then 
				NONE
			    else
				SOME false 
		    end
	    in
		case IntMap.revfold samenow levels NONE of
		    NONE => true
		  | SOME answer => answer
	    end

	(*
	 * Add Constraints.
	 *
	 * This is a somewhat complicated operation, since it
	 * requires two steps.  First, the constraint must be
	 * added to the table of constraints for all the 
	 * relevant variables.  This is accomplished by 
	 * addTableEntry and addTableEntries.
	 *
	 * Second, the constraint must be added, and if
	 * necessary triggered, on each of the live cells
	 * which would have accessed the constraint - if
	 * it had been around when the cell was created.
	 * This is done by addProduction and checkNeeded.
	 *
	 * By a lemma from the tree paper, the descendant
	 * relation (i,i') < (j,j') when (i,i') represents
	 * a constant grows monotonically.  So if we check
	 * such statements and they're true, we don't have
	 * to do the constraint munging for them...
	 *)
	local
	    (* note that GUYs merges may trigger CONSTRAINT action *) 
	    fun addTableEntry(Table{constraints}, guy, constraint) =
		(Table{constraints=
		       IntMap.insert(constraints,
				     guy,
				     constraint :: IntMap.find(constraints, guy))}
		 handle IntMap.NotFound =>
		     raise (Fail ("addTableEntry (ftree) : constraints missing guy " ^
				  (Int.toString guy))))

	    (*
	     * addTableEntries:
	     *  - descendant triggers when parents merge 
	     *  - notdesc triggers when ALL merge (any may be alone in smallest)...
	     *  - ccommand triggers when head merges.
	     *)
	    fun addTableEntries(c as Descendant{p1,p2,...}, table) =
		(inccstr ();
		 addTableEntry(addTableEntry(table, p1, c), p2, c))
	      | addTableEntries(c as NotShallower{nd1,nd2,...}, table) =
		(inccstr ();
		 addTableEntry(addTableEntry(table, nd1, c), nd2, c))
	      | addTableEntries(c as Notdesc{np1,np2,nk1,nk2}, table) =
		(inccstr ();
		 addTableEntry
		 (addTableEntry
		  (addTableEntry
		   (addTableEntry (table, np1, c), np2, c), nk1, c), nk2, c))
	      | addTableEntries(c as Ccommand{h1, h2, g1, g2}, table) =
		(inccstr ();
		 addTableEntry(addTableEntry(table, h1, c), h2, c))

	    (*
	     * it's slightly involved to add a production onto
	     * a live cell -- just because a lot of junk ends
	     * up copied.
	     *)
	    fun addProduction (level as Level{nextcell, depth, asgn, cells}, 
			       guy, constraint) =
		case IntMap.peek(asgn, guy) of
		    NONE => level
		  | SOME (cellcode) =>
			let 
			    val Cell{index, varkeys, rulesize, rules, 
				     distinctkeys, leastdk} = 
				IntMap.find(cells, cellcode)
			in
			    Level{nextcell=nextcell,
				  depth=depth,
				  asgn=asgn,
				  cells=IntMap.insert(cells,index,
						      Cell{index=index,
							   varkeys=varkeys,
							   distinctkeys=distinctkeys,
							   leastdk=leastdk,
							   rulesize=rulesize+1,
							   rules=constraint::rules})}
			end
		    handle IntMap.NotFound => 
			raise (Fail ("addProduction (ftree): cells missing cellcode " ^
				     (Int.toString cellcode) ^ " for guy " ^
				     (Int.toString guy)))

	    (* 
	     * these conventions are the same as for addTableEntries
	     *)
	    fun addProductions(level, c as Descendant{p1,p2,...}) =
		addProduction(addProduction(level, p1, c), p2, c)
	      | addProductions(level, c as NotShallower{nd1,nd2,...}) =
		addProduction(addProduction(level, nd1, c), nd2, c)
	      | addProductions(level, c as Notdesc{np1,np2,nk1,nk2}) =
		addProduction
		(addProduction
		 (addProduction
		  (addProduction(level, np1, c), np2, c), nk1, c), nk2, c)
	      | addProductions(level, c as Ccommand{h1,h2,...}) =
		addProduction(addProduction(level, h1, c), h2, c)

	    (*
	     * for each constraint, if it triggers on the assignment,
	     * you're done, otherwise you should add it to the appropriate
	     * cell --- by CHECKONE.
	     * the rest is bookkeeping.
	     *)
	    fun checkNew (levels, size, constraints) =
		let
		    fun checkOne (constraint, 
				  (level as Level{asgn,depth,...}, queue)) =
			case check (asgn, depth, size) (constraint, queue) of
			    NONE =>
				(addProductions(level, constraint), queue)
			  | SOME newqueue =>
				(level, newqueue)

		    val thequeue = ref (Fifo.empty())

		    fun transformation level =
			let
			    val (result, newqueue) =
				foldr checkOne (level, !thequeue) constraints 
			in
			    thequeue := newqueue;
			    result
			end
		in
		    (IntMap.transform transformation levels, !thequeue)
		end

	    (* 
	     * isNeeded filters out
	     * those relations that grow monotonically:
	     * they can be checked now and disposed of 
	     *)
	    fun isNeeded (p as P{distinctkeys,...})
			 (Descendant{p1,p2,k1,k2}) =
		if ((p1 = k1 andalso p2 = k2) orelse
		    (p1 = k2 andalso p2 = k1) orelse 
		    (IntSet.member(distinctkeys, p1) orelse
		     IntSet.member(distinctkeys, p2)) andalso
		    isDescendant(p, p1, p2, k1, k2)) then
		    false
		else
		    true
	      | isNeeded _ _ = true
	in
	    fun addConstraints(p as P{table, size, cvtype, 
				      distinctkeys, anonkeys, levels}, 
			       constraints) =
		let
		    val newconstraints = List.filter (isNeeded p) constraints
		    val (newlevels, queue) = 
			checkNew(levels, size, newconstraints)
		in
		    process(P{table = foldr addTableEntries table newconstraints,
			      size = size,
			      cvtype = cvtype,
			      distinctkeys = distinctkeys,
			      anonkeys = anonkeys,
			      levels = newlevels},
			    0,
			    queue)
		end handle Inconsistent => NONE
	end

	(*
	 * adding an element.
	 *)

	local 
	    (*
	     * getCell:
	     * starts things rolling by adding the guy to
	     * cell zero at level zero.  This may mean 
	     * creating the cell (for the first element);
	     * but usually involves adding it to the existing
	     * cell zero, updating the least dk if necessary,
	     * and, if necessary, triggering the first merge
	     * that will percolate a new distinct constant to
	     * its natural level...
	     *)
	    fun getCell (nextcell, cells, elt, isdistinct, anontype,
			 table, newdks, newanons) =
		if nextcell = 0 then
		    (makeCell(0, elt, table, newdks, newanons), Fifo.empty())
		else
		    let
			val Cell{varkeys, distinctkeys, leastdk, rulesize, rules,...} =
			    IntMap.find(cells, 0)
			val (newdk, queue) = 
			    if not (isdistinct) then
				(leastdk, Fifo.empty())
			    else 
				if leastdk = NONE then
				    case anontype of
					SOME t => (SOME (ANON(t,elt,[elt])), Fifo.empty())
				      | NONE => (SOME (DK(elt)), Fifo.empty())
				else
				    (leastdk, 
				     enqueue(Fifo.empty(),
					     Merge{depth=1,k1=elt,k2=elt+1}))
		    in
			(Cell{index=0,
			      varkeys=if isdistinct then varkeys else elt :: varkeys,
			      leastdk=newdk,
			      rulesize = rulesize,
			      rules=rules,
			      distinctkeys = if isdistinct then elt::distinctkeys
					     else distinctkeys},
			 queue)
		    end
		handle IntMap.NotFound => 
		    raise (Fail "getCell (ftree) : no zero cell")

	    (*
	     * addelt:
	     *   generic adder function, that builds the
	     *   cell, updates inventory information on the
	     *   problem, and then process any merges that
	     *   have resulted.
	     *)
	    fun addelt(elt, isdistinct, anontype,
		       P{table=table as Table{constraints}, 
			 distinctkeys, anonkeys, size, cvtype, levels}) =
		let
		    val newdks = (case (isdistinct, anontype) of
				      (true, NONE) => IntSet.add(distinctkeys,elt)
				    | _ => distinctkeys)
		    val newanons = (case (isdistinct, anontype) of
					(true, SOME k) => IntMap.insert(anonkeys,elt,k)
				      | _ => anonkeys)

		    val Level{depth, nextcell, asgn, cells} = 
			IntMap.find(levels, 0)

		    val (newcell, queue) = 
			getCell(nextcell, cells, elt, isdistinct, anontype,
				table, newdks, newanons)

		    val newcells = 
			IntMap.insert(cells, 0, newcell)
		in
		    process(P{table = Table {constraints = 
					     IntMap.insert(constraints, 
							   elt, nil)},
			      size = (if not(cvtype) orelse isdistinct then size + 1 
				      else size),
			      distinctkeys = newdks,
			      anonkeys = newanons,
			      cvtype = cvtype,
			      levels = 
			      IntMap.insert(levels, 0,
					    Level{depth=depth, nextcell = 1,
						  asgn=IntMap.insert(asgn, elt, 0),
						  cells = newcells})},
		    0,
		    queue)
		end
	    handle IntMap.NotFound => 
		raise (Fail "addelt (ftree) : no level 0")
	in
	    (* three different kinds *)

	    fun addElement(p, elt) = 
		(addelt(elt, false, NONE, p))

	    fun addConstant(p, elt) = 
		(case addelt(elt+1, false, NONE, p) of
		     NONE => NONE
		   | SOME np => addelt(elt, true, NONE, np))

	    fun addAnon(p, typ, elt) =
		(case addelt(elt+1, false, NONE, p) of
		     NONE => NONE
		   | SOME np => addelt(elt, true, SOME typ, np))
	end

	(*
	 * getPath/moveUp
	 *    These functions support typechecking of modal
	 *    paths (now performed by the external patheq
	 *    module).
	 *
	 * getPath: problem * int * int * int * int ->
	 * 	    (int * int option) list
	 * (1,2) = (t1, t2) represents the root end of
	 *    a path through the tree;
	 * (3,4) - (b1, b2) represents the leaf end of
	 *    the path.
	 * the job of this function is to return the 
	 * path between them as a list of (depth, constant?)
	 * pairs---giving the level in the tree and the
	 * ldk label of the cell in the path.  (Note that
	 * (t1, t2) is not included in this path.)
	 *)

	local
	    (*
	     * if A and B are together, given
	     * ASGN and CELLS, return SOME (LDK)
	     * where LDK gives the (constant?)
	     * coding for the cell where they 
	     * both are---else return NONE.
	     *)
	    fun samecelldk (asgn, cells, a, b) =
		let
		    val clo = IntMap.peek(asgn, a)
		in
		    case clo of NONE => NONE
		  | SOME index =>
			if clo = IntMap.peek(asgn,b) then
			    case IntMap.find(cells,index) of
				Cell{leastdk,...} => 
				    case leastdk of
					NONE => SOME(NONE)
				      | SOME (DK i) => SOME(SOME i)
				      | SOME (ANON(_,i,_)) => SOME(SOME i)
			else
			    NONE
		end
	    handle IntMap.NotFound =>
		raise (Fail "samecelldk (ftree) : cells missing index")

	    (*
	     * find the root of the path---the node
	     * labeled (T1,T2).
	     *)
	    fun findroot (levels, t1, t2, start) =
		let 
		    fun step (depth) =
			let
			    val Level{asgn,cells,...} = 
				IntMap.find(levels, depth)
			in
			    case samecelldk (asgn, cells, t1, t2) of
				NONE => depth
			      | SOME next => step(depth+1)
			end handle IntMap.NotFound => depth
		in
		    step start
		end
	in
	    fun getPath (P{levels,...}, t1, t2, b1, b2) =
		let
		    (*
		     * complete the path---add on pairs given
		     * by samecelldk, as long as there are any
		     *)
		    fun buildpath (depth,sofar) =
			let
			    val Level{asgn,cells,...} = 
				IntMap.find(levels, depth)
			in
			    case samecelldk (asgn, cells, b1, b2) of
				NONE => rev sofar
			      | SOME next => 
				    buildpath(depth+1, (depth, next) :: sofar)
			end	handle IntMap.NotFound => rev sofar
		in
		    buildpath(findroot (levels, t1, t2, 0), nil)
		end

	    fun getConst (P{levels,...}, t1, t2) =
		let
		    val depth = findroot (levels,t1,t2,0)

		    val Level{asgn,cells,...} =
			IntMap.find(levels,depth-1)
		in
		    case samecelldk(asgn, cells, t1, t2) of
			NONE => NONE
		      | SOME dk => dk
		end
	end

	(*
	 * moveUp: problem * (int * int * int * int) list ->
	 *     problem option * bool
	 * 
	 * Each element of the list is of the form (i,i',d,k)
	 * and says that the node (i,i') should be merged
	 * down to depth d, where it should be put into
	 * the same cell as constant k.
	 *
	 * Some of these may duplicate the locations at which
	 * node (i,i') already is, so these instructions are
	 * checked before being posted.  If none are posted,
	 * we return the current problem and note that there
	 * were no changes; if some are, we process the queue
	 * and return that there were changes...
	 *)

	fun moveUp (p as P{levels,...}, optlist) =
	    let
		fun addp (e, depth, k, queue) =
		    if inCell(p, e, k, depth) then
			queue
		    else
			enqueue(queue,
				Merge{depth=depth, k1=e, k2=k})

		fun mu ((t1, t2, depth, k), queue) =
		    foldl (fn (l,q) => 
			     addp(t1, l, k, addp(t2, l, k, queue)))
		    queue
		    (List.tabulate(depth, fn x => x+1))

		val newqueue = foldr mu (Fifo.empty()) optlist
	    in
		if Fifo.isEmpty newqueue then
		    (SOME p, false)
		else
		    (process(p, 0, newqueue), true)
	    end

	(*
	 * updateVLessStats / makeVLessStats
	 *    to check an increasing domain constraint on
	 *    higher modal levels, we want to compile a record
	 *    of which constants appear on the values of which
	 *    variables.  These functions compute that.
	 * The loop structure takes advantage of the 
	 * representation of cells, with the ldk distinguished.
	 * All and only the vars in this cell contain the ldk's value.
	 * (With ANONs there's the possibility of having more than
	 * one ldk, and the attendant possibility of having a bogus
	 * element in the ldk list.)  
	 * Both functions return intmaps from constants to
	 *   (all vars, new vars) pairs.
	 *)

	local
	    (* this fast diff function fits a merge strategy *)
	    fun diff (l, nil) = l
	      | diff (h1::t1, h2::t2) = 
		if h1 = h2 then diff(t1, t2)
		else h1 :: (diff(t1, h2::t2))
	      | diff (nil, _) = 
		(print "Bad case (tree.sml): diff failed.\n";
		 nil)

	    (* now LD's record is not as given in LESSMAP but as VARKEYS *)
	    fun updateVs (varkeys, ld, lessmap) =
		(case IntMap.peek(lessmap, ld) of
		     NONE => IntMap.insert(lessmap, ld, (varkeys, varkeys))
		   | SOME (all,_) => IntMap.insert(lessmap, ld, 
						   (varkeys, diff(varkeys,all))))

	    (* make a record for LD only if LD splits -- so names this cell *)
	    fun safeupdateVs (varkeys,together) (ld, lessmap) =
		if together (ld, ld+1) then lessmap
		else updateVs(varkeys, ld, lessmap)

	    (* update the record for CELL's LD's *)
	    fun updateCell _ (_, Cell{leastdk=NONE,...}, lessmap) = lessmap
	      | updateCell _ (_, Cell{varkeys, leastdk=SOME (DK ld),...}, lessmap) =
		updateVs (varkeys, ld, lessmap)
	      | updateCell together 
		(_, Cell{varkeys, leastdk=SOME (ANON(_,_,lds)),...}, lessmap) =
		foldr (safeupdateVs (varkeys, together)) lessmap lds

	    (* update all LEVEL's cells, with split check for next level *)
	    fun updateLevel levels (_, Level{cells,depth,...}, lessmap) =
		IntMap.fold (updateCell (case IntMap.peek(levels, depth+1) of
					     SOME(Level{asgn,...}) =>
						 (fn (x,y) => together(asgn,x,y))
					   | NONE => fn _ => false))
		cells lessmap
	in
	    fun updateVLessStats(P{levels,...},  lessmap) =
		IntMap.fold (updateLevel levels) levels lessmap

	    fun makeVLessStats(p) =
		updateVLessStats(p, IntMap.empty())
	end

	(*
	 * convenience functions abbreviating interesting
	 * complex constraints:
	 *   notle - notdesc
	 *   le - descendant
	 *   lt - proper descendant
	 *   eq - identity
	 *   kid - child (immediate descendant)
	 *   kidsame - child or identity
	 *   notsh - not shallower
	 *   edqp - same depth
	 *)

	fun notle(a,b,c,d) = [ Notdesc{nk1=a,nk2=b,np1=c,np2=d} ]

	fun le(a,b,c,d) = [ Descendant{k1=a,k2=b,p1=c,p2=d} ]

	fun lt(a,b,c,d) = [ Descendant{k1=a,k2=b,p1=c,p2=d},  
			   Notdesc{np1=a,np2=b,nk1=c, nk2=d} ]

	fun eq(a,b,c,d) = [ Descendant{k1=a,k2=b,p1=c,p2=d},  
			   Descendant{p1=a,p2=b,k1=c,k2=d} ]

	fun kid(a,b,c,d) = lt(a,b,c,d) @ [ Ccommand{h1=a, h2=b, g1=c, g2=d} ]
	fun kidsame(a,b,c,d)  = [Ccommand{h1=a, h2=b, g1=c, g2=d},
				 Descendant{k1=a, k2=b, p1=c, p2=d} ]

	fun notsh(a,b,c,d) = [ NotShallower{ns1=a, ns2=b, nd1=c, nd2=d} ]
	fun eqdp(a,b,c,d) = [ NotShallower{ns1=a, ns2=b, nd1=c, nd2=d},
			     NotShallower{nd1=a, nd2=b, ns1=c, ns2=d} ]
    end
