   
/*
    socket.c - socket convenience functions
    Drew Olbrich, 1992
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <errno.h>
#include <signal.h>
#include <ctype.h>
#include <fcntl.h>

#include "socket.h"

#define SELECT_FD_INC   8   /* increment for select FD table */

int socket_errno = 0;

/*
    Opens a TCP socket connection for a server process.  The file 
    descriptor for the socket is returned.  -1 is returned
    if a connection cannot be established.
  
    To choose a random unused port, pass zero as the port number.
    Call 'socket_port()' to find out which port was used.
*/

int socket_server_open(int port)
{
    struct sockaddr_in addr;
    int fd;
    
    fd = socket(AF_INET, SOCK_STREAM, PF_UNSPEC);
    if (fd == -1) {
        socket_errno = SOCKET_ERR_SOCKET;
        return -1;
    }
    
    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);
    addr.sin_addr.s_addr = 0;
    
    if (bind(fd, (struct sockaddr *) &addr, sizeof(addr)) == -1) {
        socket_errno = SOCKET_ERR_BIND;
        return -1;
    }
    
    if (listen(fd, SOMAXCONN) == -1) {
        socket_errno = SOCKET_ERR_LISTEN;
        return -1;
    }
    
    return fd;
}

/*
    Opens a TCP socket connection for a client process.  The file 
    descriptor for the socket is returned.  -1 is returned
    if a connection cannot be established.
*/

int socket_client_open(char *host, int port)
{
    struct sockaddr_in addr;
    struct hostent *hp;
    int fd;
    
    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);
    
    hp = gethostbyname(host);
    if (hp == NULL) {
        socket_errno = SOCKET_ERR_HOST;
        return -1;
    }
    
    bcopy(hp->h_addr, (char *) &addr.sin_addr, sizeof(addr.sin_addr));
    
    fd = socket(AF_INET, SOCK_STREAM, PF_UNSPEC);
    if (fd == -1) {
        socket_errno = SOCKET_ERR_SOCKET;
        return -1;
    }
    
    if (connect(fd, (struct sockaddr *) &addr, sizeof(addr)) == -1) {
	(void)close(fd);
        socket_errno = SOCKET_ERR_CONNECT;
        return -1;
    }
    
    return fd;
}

/*
    Accept a new socket connection to a server.  Returns the
    socket number, or -1 on failure.
  
    A read event on a server socket indicates a new connection
    is being requested.  'socket_accept()' should then be
    called to complete the connection.
*/

int socket_accept(int fd)
{
    int fd_new;
    int addr_len;
    struct sockaddr_in addr;
    
    addr_len = sizeof(addr);
    fd_new = accept(fd, (struct sockaddr *) &addr, &addr_len);
    if (fd_new == -1) {
        socket_errno = SOCKET_ERR_ACCEPT;
        return -1;
    }
    
    return fd_new;
}

/*
    Returns the port number associated with a file descriptor.
*/

int socket_port(int fd)
{
    int addr_len;
    struct sockaddr_in addr;
    
    addr_len = sizeof(addr);
    if (getsockname(fd, (struct sockaddr *) &addr, &addr_len) == -1) {
        socket_errno = SOCKET_ERR_GETSOCKNAME;
        return -1;
    }
    
    return ntohs(addr.sin_port);
}

/*
    If 'value' is nonzero, the specified socket will have
    packetization of transmitted data disabled.  All data passing
    through the socket will be sent immediately.  By default,
    a short delay occurs before data is sent, so that successive
    bytes may be concatenated into a single packet.
*/

int socket_nodelay(int fd, int value)
{
    if (setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, 
		   (char *) &value, sizeof(int)) < 0) {
        socket_errno = SOCKET_ERR_SETSOCKOPT;
        return -1;
    }
    
    return 0;
}

/*
    Return a new select structure.  Use a select structure
    to keep track of a list of file descriptors that you
    want to block on.  See "socket_select_loop()" below.
*/

SOCKET_SELECT *socket_select_new()
{
    SOCKET_SELECT *s;
    
    s = (SOCKET_SELECT *) malloc(sizeof(SOCKET_SELECT));
    if (s == NULL) {
        socket_errno = SOCKET_ERR_MALLOC;
        return NULL;
    }
    
    s->count = 0;
    s->max = 0;
    s->sfd = NULL;
    FD_ZERO(&s->fds_read);
    FD_ZERO(&s->fds_write);
    FD_ZERO(&s->fds_except);
    s->timeout.tv_sec = -1;
    s->timeout.tv_usec = -1;
    
    return s;
}

/*
    Free up a select structure.
*/

void socket_select_free(SOCKET_SELECT *s)
{
    free(s->sfd);
    free(s);
}

/*
    Set the timeout value for a select structure.
*/

void socket_select_timeout(SOCKET_SELECT *s, int sec, int usec)
{
    s->timeout.tv_sec = sec;
    s->timeout.tv_usec = usec;
}

/*
    Add a file descriptor to a select structure.  This file 
    descriptor will be checked on calls to 'socket_select_loop()'.
    
    -1 is returned if the file descriptor is already in the table.
    
    'socket_select_callback()' should be called to set up
    callbacks for the file descriptor.
*/

int socket_select_add(SOCKET_SELECT *s, int fd)
{
    int index;
    int i;
    
    index = -1;
    for (i = 0; i < s->count; i++) {
        if (s->sfd[i].fd == -1)
            index = i;
            
        if (s->sfd[i].fd == fd) {
            socket_errno = SOCKET_ERR_EXISTS;
            return -1;
        }
    }
    
    if (index == -1) {
        index = s->count;
        
        if (s->max == 0) {
            s->max = SELECT_FD_INC;
            s->sfd = (SOCKET_SELECT_FD *) 
                malloc(sizeof(SOCKET_SELECT_FD)*s->max);
        } else if (s->count == s->max) { 
            s->max += SELECT_FD_INC;
            s->sfd = (SOCKET_SELECT_FD *) realloc(s->sfd, 
                sizeof(SOCKET_SELECT_FD)*s->max);
        }
    
        s->count++;
    }
    
    s->sfd[index].fd = fd;
    s->sfd[index].callback_read = NULL;
    s->sfd[index].callback_write = NULL;
    s->sfd[index].callback_except = NULL;
    
    return 0;
}

/*
    Remove a file descriptor from a select structure.  
    Returns -1 if the file descriptor isn't in the list.
*/

int socket_select_remove(SOCKET_SELECT *s, int fd)
{
    int i;
    
    for (i = 0; i < s->count; i++)
        if (s->sfd[i].fd == fd) {
            s->sfd[i].fd = -1;
            
            FD_CLR(fd, &s->fds_read);
            FD_CLR(fd, &s->fds_write);
            FD_CLR(fd, &s->fds_except);
            
            return 0;
        }
    
    socket_errno = SOCKET_ERR_NOMATCH;
    return -1;
}

/*
    Set the callbacks for a select structure.  Each file
    descriptor has its own set of callbacks.
*/

int socket_select_callback(SOCKET_SELECT *s, int fd, int cb, void (*fn)())
{
    int i;
    SOCKET_SELECT_FD *sfd;
    
    for (i = 0; i < s->count; i++) {
        sfd = &s->sfd[i];
            
        if (sfd->fd == fd) {
            switch (cb) {
            case SOCKET_CB_READ :
                sfd->callback_read = fn;
                if (fn != NULL)
                    FD_SET(fd, &s->fds_read);
                else
                    FD_CLR(fd, &s->fds_read);
                break;
                
            case SOCKET_CB_WRITE :
                sfd->callback_write = fn;
                if (fn != NULL)
                    FD_SET(fd, &s->fds_write);
                else
                    FD_CLR(fd, &s->fds_write);
                break;
                
            case SOCKET_CB_EXCEPT :
                sfd->callback_except = fn;
                if (fn != NULL)
                    FD_SET(fd, &s->fds_except);
                else
                    FD_CLR(fd, &s->fds_except);
                break;
                
                default :
                    socket_errno = SOCKET_ERR_INVALIDCB;
                return -1;
                break;
            }
            
            return 0;
        }
    }
    
    socket_errno = SOCKET_ERR_NOMATCH;
    return -1;
}

/*
    This function waits for socket activity to occur.
  
    If a socket read, write, or exception event occurs, the
    corresponding callback (set with 'socket_select_callback()')
    is invoked, with the active file descriptor as an argument.
    
    Zero is returned if a timeout (set with 'socket_select_timeout()')
    occurs.  
    
    If an error occurs, -1 is returned.
*/

int socket_select_loop(SOCKET_SELECT *s)
{
    fd_set read_fds;
    fd_set write_fds;
    fd_set except_fds;
    struct timeval timeout;
    int count;
    int i;
    SOCKET_SELECT_FD *sfd;
    
    for (;;) {
        read_fds = s->fds_read;
        write_fds = s->fds_write;
        except_fds = s->fds_except;
        
        if (s->timeout.tv_sec < 0 || s->timeout.tv_usec < 0)
            count = select(FD_SETSIZE, &read_fds, &write_fds, 
                           &except_fds, NULL);
        else {
            timeout = s->timeout;
            count = select(FD_SETSIZE, &read_fds, &write_fds, 
                           &except_fds, &timeout);
        }
        
        /* timeout */
        if (count == 0)
            return 0;
            
        if (count == -1) {
            socket_errno = SOCKET_ERR_SELECT;
            return -1;
        }
        
        for (i = 0; i < s->count; i++) {
            sfd = &s->sfd[i];
            
            /* we check the condition 'sfd->fd != -1' repeatedly
               because it might change during callback invocation */
            
            if (sfd->fd != -1 && sfd->callback_read != NULL 
                && FD_ISSET(sfd->fd, &read_fds))
                sfd->callback_read(sfd->fd);
            
            if (sfd->fd != -1 && sfd->callback_write != NULL
                && FD_ISSET(sfd->fd, &write_fds))
                sfd->callback_write(sfd->fd);
            
            if (sfd->fd != -1 && sfd->callback_except != NULL
                && FD_ISSET(sfd->fd, &except_fds))
                sfd->callback_except(sfd->fd);
        }
    }
}

/*
    Create a TCP server socket and place a reference to it
    in the specified file.
    
    Returns the file decriptor of the socket or -1 on failure.
    
    This function will fail if another server referencing the
    same file name is already active.
*/

int socket_server_open_fn(char *fn)
{
    char temp_fn[256];
    char host[256];
    FILE *file;
    int count;
    int fd_server;
    int fd;
    mode_t old_mask;
    int port;
    
    sprintf(temp_fn, "%s.temp.temp", fn);
    
    file = fopen(fn, "r");
    if (file != NULL) {
        count = fscanf(file, "%s %d\n", host, &port);
        fclose(file);
        
        if (count != 2) {
            socket_errno = SOCKET_ERR_READ;
            return -1;
        }
        
        fd = socket_client_open(host, port);
        if (fd != -1) {
            close(fd);
            
            socket_errno = SOCKET_ERR_CLIENT;
            return -1;
        }
    }
    
    fd_server = socket_server_open(0);
    if (fd_server == -1) {
        socket_errno = SOCKET_ERR_SERVER;
        return -1;
    }
    
    port = socket_port(fd_server);
    if (port == -1) {
        socket_errno = SOCKET_ERR_PORT;
        return -1;
    }
    
    if (gethostname(host, sizeof(host) - 1) == -1) {
        socket_errno = SOCKET_ERR_HOST;
        return -1;
    }
    
    old_mask = umask(0000);
    fd = open(temp_fn, O_RDWR | O_CREAT, 0666);
    if (fd == -1) {
        socket_errno = SOCKET_ERR_OPEN;
        return -1;
    }
    
    umask(old_mask);
    
    file = fdopen(fd, "w");
    if (file == NULL) {
        socket_errno = SOCKET_ERR_OPEN;
        return -1;
    }
    
    fprintf(file, "%s %d\n", host, port);
    fclose(file);
    
    if (rename(temp_fn, fn) == -1) {
        socket_errno = SOCKET_ERR_RENAME;
        return -1;
    }
    
    return fd_server;
}

/*
    Open a TCP client socket connection using information found
    in a file.
  
    Returns the file decriptor of the socket or -1 on failure.
*/

int socket_client_open_fn(char *fn)
{
    FILE *file;
    char host[256];
    int port;
    int count;
    int fd;
    
    file = fopen(fn, "r");
    if (file == NULL) {
        socket_errno = SOCKET_ERR_OPEN;
        return -1;
    }
    
    count = fscanf(file, "%s %d\n", host, &port);
    fclose(file);
    
    if (count != 2) {
        socket_errno = SOCKET_ERR_READ;
        return -1;
    }
    
    fd = socket_client_open(host, port);
    if (fd == -1) {
        socket_errno = SOCKET_ERR_CLIENT;
        return -1;
    }
    
    return fd;
}

#ifdef UDP

/*
    Opens a UDP socket.  If 'port' is zero, a new unique port
    number is created, and can be later checked with a call to
    'socket_port()'.
*/

int socket_datagram_open(int port)
{
    int fd;
    struct sockaddr_in addr;
    
    fd = socket(AF_INET, SOCK_STREAM, PF_UNSPEC);
    if (fd == -1) {
        socket_errno = SOCKET_ERR_SOCKET;
        return -1;
    }
    
    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);
    addr.sin_addr.s_addr = 0;
    
    if (bind(fd, &addr, sizeof(struct sockaddr_in)) == -1) {
        socket_errno = SOCKET_ERR_BIND;
        return -1;
    }
    
    return fd;
}

#endif

