(*
 * Matthew Stone
 * Spud 0.3 - 1/99
 *
 * Signature of some actions to supply on an
 * interactive pipe to spud.
 *
 *)

functor SpudMainStuff (structure Main : MAIN
		       structure Derive : DERIVE
		       structure Morphdb : MORPHDB

		       sharing Morphdb.Morph.Features = Derive.Itag.Tag.Features

		       val init: string * string list -> unit;
		       val input: unit -> string;
		       val output: string -> unit;
		       val finish: unit -> OS.Process.status) : MAIN_ACTS =
    struct
	structure Main = Main

	val init = init
	val input = input
	val output = output
	val finish = finish
	
	exception readError

	datatype grammar = 
	    GR of { modspecs : Derive.Word.Formula.Patheq.Modspec.collection,
		    laws : Derive.Word.Formula.database,
		    words : Derive.Word.lexicon,
		    trees : Derive.Itag.taglex,
		    forms : Morphdb.morphdb }

	val theGrammar = ref NONE : grammar option ref
	val theFacts = ref NONE : Derive.Word.Formula.database option ref
	val theContext = ref NONE : Derive.Word.Formula.database option ref

	fun reset nil =
	    (case !theGrammar of
		 NONE => 
		     (theFacts := NONE;
		      theContext := NONE;
		      true)
	       | SOME (GR{laws,...}) => 
		     (theFacts := SOME laws;
		      theContext := SOME laws;
		      true))
	  | reset _ =
		 (print "Wrong number of args to reset.\n";
		  true)

	local 
	    fun readModspecs (file) =
		let
		    open Derive.Itag.Formula.Patheq.Modspec
		    val new = newCollection ()
		    fun reader (str, ()) =
			(case readModality (new, str) of
			     (SOME m, _) =>
				 (addModality (new, m);
				  ((), true))
			   | _ => ((), false))
		in
		    (case Main.readLoop (file, reader, ()) of
			 NONE => raise readError
		       | SOME () => (compile(new); new))
		end
				 
	    fun addAxioms (file, collection, database) =
		let
		    open Derive.Word.Formula
		    val rf = readFormula collection
		    fun reader (nil, db) = (db, false)
		      | reader (str, db) =
			(case rf str of
			     (SOME f, nil) =>
				 (addFormulas (db, [("anonymous", f)]), true)
			   | (SOME f, [ "." ]) =>
				 (addFormulas (db, [("anonymous", f)]), true)
			   | (SOME f, trail) =>
				 (print ("WARNING file " ^ file ^ 
					 ": unparsed trailing material:\n");
				  app (fn x => print (x ^ " ")) trail;
				  print "\n";
				  (addFormulas (db, [("anonymous", f)]), true))
			   | _ => (db, false))
		in
		    (case Main.readLoop (file, reader, database) of
			 NONE => 
			     (print "couldn't read database\n";
			      raise readError)
		       | SOME db => db)
		end

	    fun readMorphs (file) =
		let
		    open Morphdb
		    val rm = Morph.readEntry
		    val db = newdb()
		    fun reader (nil, ()) = ((), false)
		      | reader (str, ()) =
			(case rm str of
			     (SOME m, _) =>
				 (addEntry db m, true)
			   | _ => ((), false))

		in
		    (case Main.readLoop (file, reader, ()) of
			 NONE =>
			     (print "couldn't read morphs\n";
			      raise readError)
		       | SOME () => 
			     db)
		end

	    fun readSpecSet (tokens, collection, database) =
		let 
		    val specpfx =
			(case Derive.Word.Formula.readFormula collection tokens of
			     (SOME specpfx, nil) => specpfx
			   | (SOME specpfx, trail) =>
				 (print "WARNING: Unparsed trailing info in specificity test formula:\n";
				  app (fn x => (print (x ^ " "))) trail;
				  print "\n";
				  specpfx)
			   | _ => 
				 (print "ERROR: Couldn't parse specificity test formula \n";
				  raise readError))

		    val (_, sset) = Derive.Word.Formula.setof(fn _ => nil, (), specpfx, database)
		in
		    sset
		end
		    
	    fun readWords (file, collection, sset) = 
		let
		    open Derive.Word
		    val rw = readWord collection
		    fun reader (nil, lex) = (lex, false)
		      | reader (str, lex) =
			case rw str of
			    (SOME w, _) =>
				(addWord(lex, w), true)
			  | _ => (lex, false)
		in
		    (case Main.readLoop(file, reader, newLexicon()) of
			 NONE => 
			     (print "couldn't read lexicon\n";
			      raise readError)
		       | SOME lex => 
			     specOrder sset lex)
		end

	    fun readTrees (file, collection, sset) =
		let
		    open Derive.Itag
		    val rt = readEntry collection
		    fun reader (nil, trees) = (trees, false)
		      | reader (str, trees) =
			case rt str of
			    (SOME t, _) => (addEntry(trees, t), true)
			  | _ => (trees, false)
		in
		    (case Main.readLoop (file, reader, newTaglex()) of
			 NONE => 
			     (print "couldn't read trees\n";
			      raise readError)
		       | SOME trees => 
			     specOrder sset trees)
		end
	    
	    fun fileconcat (filename, "." :: more) = 
		filesnarf ("." :: filename, more)
	      | fileconcat (filename, "/" :: more) = 
		filesnarf ("/" :: filename, more)
	      | fileconcat (filename, rest) = 
		(String.concat (rev filename), rest)

	    and filesnarf (filename, "." :: more) =
		filesnarf ("." :: filename, more)
	      | filesnarf (filename, "/" :: more) = 
		filesnarf ("/" :: filename, more)
	      | filesnarf (filename, a :: more) =
		fileconcat (a :: filename, more)
	      | filesnarf (filename, nil) = 
		fileconcat(filename, nil)
		
	in
	    fun readGrammar (tokens) = 
		let
		    val (filename, spectokens) = filesnarf(nil, tokens)
		    val collection = readModspecs(filename ^ ".mdl")
		    val database = addAxioms (filename ^ ".max", collection,
					      Derive.Word.Formula.makeDatabase(collection,nil))
		    val sset = readSpecSet (spectokens, collection, database)
		    val morphs = readMorphs (filename ^ ".mph")
		    val words = readWords (filename ^ ".lex", collection, sset)
		    val trees = readTrees (filename ^ ".tr", collection, sset)
		in
		    theGrammar := 
		    SOME (GR {modspecs = collection,
			      laws = database,
			      words = words,
			      trees = trees,
			      forms = morphs});
		    theFacts := SOME database;
		    theContext := SOME database;
		    true
		end handle readError => (print "No grammar loaded\n";
					 true)

	    fun readFacts (tokens) =
		let
		    val (file, rest) = filesnarf(nil, tokens)
		in
		    case (!theGrammar, !theFacts, rest) of
			(SOME (GR{modspecs,...}), SOME facts, nil) =>
			    (theFacts := SOME (addAxioms(file, modspecs, facts));
			     theContext := !theFacts;
			     true)
		      | (_, _, _ :: _) => 
			    (print "Error: bad args to factfile.\n";
			     true)
		      | _ => (print "Error: No grammar loaded\n";
			      true)
		end			
	end

	val shared = ref NONE : Derive.Word.Formula.formula option ref
	val private = ref NONE : Derive.Word.Formula.formula option ref
	val pattern = ref NONE : Derive.Constraint.Call.call option ref
	val features = ref NONE : Derive.Itag.Tag.Features.features option ref
	val communicate = ref nil : Derive.Word.Formula.formula list ref

	fun clearComm nil =
	    (communicate := nil; true)
	  | clearComm _ =
	    (print "Error: wrong number of arguments\n";
	     true)

	local
	    fun rf x = 
		case !theGrammar of
		    NONE => (print "No grammar loaded\n";
			     (NONE, x))
		  | SOME (GR{modspecs,...}) =>
			Derive.Word.Formula.readFormula modspecs x

	    fun af f =
		case !theFacts of
		    NONE => ()
		  | SOME db => 
			(theFacts := 
			 SOME (Derive.Word.Formula.addFormulas (db, [("anonymous", f)]));
			 theContext := !theFacts)

	    fun ac f = 
		case !theContext of
		    NONE => ()
		  | SOME db => 
			(theContext := 
			 SOME (Derive.Word.Formula.addFormulas (db, [("anonymous", f)])))

	    fun aca f =
		case (!theContext, !shared, !private) of
		    (SOME db, SOME sh, SOME pr) => 
			communicate := Derive.expandAbbrev(db, pr, sh, f) @ 
			!communicate
		  | _ => ()

	in
	    val setShared =
		Main.ftest1(rf,
			    fn x => shared := SOME x,
			    fn x => x)

	    val setPrivate =
		Main.ftest1(rf,
			    fn x => private := SOME x,
			    fn x => x)

	    val addCommunicate =
		Main.ftest1(rf,
			    fn x => communicate := x :: !communicate,
			    fn x => x)

	    val addCommAll =
		Main.ftest1(rf,
			    aca,
			    fn x => x)

	    val setPattern = 
		Main.ftest1(Derive.Constraint.Call.readCall 
			    (Derive.Constraint.Call.newPPVar ()),
			    fn x => pattern := SOME x,
			    fn x => x)

	    val setFeatures = 
		Main.ftest1(Derive.Itag.Tag.Features.readFeatures,
			    fn x => features := SOME x,
			    fn x => x)

	    val addFact =
		Main.ftest1(rf, af, fn x => x)

	    val addContext =
		Main.ftest1(rf, ac, fn x => x)

	end

	local 
	    open Derive

	    fun doDescribe (call) =
		case (!theGrammar, !theContext, !shared, !private, !pattern, !features, !communicate) of
		    (SOME (GR{modspecs, laws, words, trees, forms}), SOME context,
		     SOME currShared, SOME currPrivate, SOME currPattern, SOME currFeatures, 
		     currCommunicate) => 
		    let
			val problem =
			    assembleGenerand "gen" 
			    [NAME "stdin", FMLA currPrivate, FMLA currShared, 
			     CALL call, CALL currPattern, FEATURES currFeatures,
			     GOALS currCommunicate] 
			val doword = Morphdb.findForm forms
		    in
			case problem of
			    NONE => (print "couldn't build problem\n") 
			  | SOME p => 
				let val result = generateList (words, trees, context) p
				in outputResult (doword, output) result;
				    communicate := remainingGoals result
				end
		    end
		   | _ => (print "Setup incomplete\n")
	in
	    val describe =
		Main.ftest1(Derive.Constraint.Call.readCall
			    (Derive.Constraint.Call.newPPVar ()),
			    fn x => doDescribe x,
			    fn x => x)
	end

	val actions =
	    [Main.A{keyword = "quit",
		    description = "Exit interpreter.\n",
		    operation = fn _ => false,
		    call = "quit"},

	     Main.A{keyword = "grammar",
		    description = "Load grammar with stem <stem> and spec formula <formula>.\n",
		    operation = readGrammar,
		    call = "grammar <stem> <formula>"},

	     Main.A{keyword = "reset",
		    description = "Remove contingent facts and context.\n",
		    operation = reset,
		    call = "reset"},

	     Main.A{keyword = "clear",
		    description = "Forget about any outstanding communicative goals.\n",
		    operation = clearComm,
		    call = "clear"},

	     Main.A{keyword = "fact",
		    description = "Add new fact <formula> and reset context.\n",
		    operation = addFact,
		    call = "fact <formula>"},

	     Main.A{keyword = "factfile",
		    description = "Add facts from <file> and reset context.\n",
		    operation = readFacts,
		    call = "factfile <file>"},

	     Main.A{keyword = "context",
		    description = "Add new formula <formula> to context.\n",
		    operation = addContext,
		    call = "context <formula>"},

	     Main.A{keyword = "features",
		    description = "Set syntactic features for next gen run to <features>.\n",
		    operation = setFeatures,
		    call = "features <features>"},

	     Main.A{keyword = "pattern",
		    description = "Set call pattern for next gen run to <call>.\n",
		    operation = setPattern,
		    call = "pattern <call>"},

	     Main.A{keyword = "shared",
		    description = "Set shared context for next gen run to <formula>.\n",
		    operation = setShared,
		    call = "shared <formula>"},

	     Main.A{keyword = "private",
		    description = "Set private context for next gen run to <formula>.\n",
		    operation = setPrivate,
		    call = "private <formula>"},
	     
	     Main.A{keyword = "communicate",
		    description = "Add <formula> to communicate on next gen run.\n",
		    operation = addCommunicate,
		    call = "communicate <formula>"},
	     
	     Main.A{keyword = "communicateAll",
		    description = "Expand <formula> to obtain facts to communicate on next gen run.\n",
		    operation = addCommAll,
		    call = "communicateAll <formula>"},
	     
	     Main.A{keyword = "describe",
		    description = "Generate a sentence for call <call>.\n",
		    operation = describe,
		    call = "describe <formula>"}
	     
	     ]
    end
