(*
 * Matthew Stone
 * Dialup 0.1 - 3/96
 *
 * Implementation of a user-level interpreter
 * for the Dialup modal logic programming language.
 *)

functor MainStuff (structure Main: MAIN;
		   structure Matchbank: MATCHBANK;
		   structure Formulas: FDB;
		   structure Assume: ASSUME;

		   structure Derive : DERIVE;
		       
		   sharing Derive.Constraint.Formula = Formulas;
		   sharing Formulas.Term = Assume.Term;
		   sharing Matchbank.Term = Formulas.Term;

		   val init: string * string list -> unit;
		   val input: unit -> string;
		   val output: string -> unit;
		   val finish: unit -> OS.Process.status) : MAIN_ACTS =
    struct
	structure Main = Main
	structure Constraint = Derive.Constraint
	structure Word = Derive.Word
	structure Itag = Derive.Itag

	val init = init
	val input = input
	val output = output
	val finish = finish

	(* 
	 * Maintain an active spec of modalities, and
	 * an active database of rules...
	 *)
	
	local 
	    open Formulas
	    open Patheq.Modspec
	in
	    val activeModspecs = ref (newCollection ())
	    val activeModspecsStale = ref false

	    val activeDatabase = ref (makeDatabase(!activeModspecs, nil))
	    val activeQabbrevs = ref (nil : formula Matchbank.matchelt list)
	    val activeSchemas = ref (nil : (formula * formula) Matchbank.matchelt list)
	    val activeWords = ref (Word.newLexicon ())
	    val activeTrees = ref (Itag.newTaglex ())

	    (*
	     * We also maintain a current constraint-satisfaction problem,
	     * to test the constraint-solving package.
	     *)
     
	    datatype cinfo =
		CI of {constraints: Constraint.problem,
		       ginit: Constraint.Call.init,
		       ginfo: Constraint.Call.info,
		       sinit: Constraint.Call.init,
		       sinfo: Constraint.Call.info}

	    val activeConstraints = ref (NONE: cinfo option)

	    (*
	     (CI{constraints = Constraint.newProblem (),
	     ginit = Constraint.Call.mkInit (),
	     ginfo = Constraint.Call.mkInfo (),
	     sinit = Constraint.Call.mkInit () ,
	     sinfo = Constraint.Call.mkInfo ()})
	     *)

	    fun resetConstraints () = 
		activeConstraints := NONE
	    (*
	     CI{constraints = Constraint.newProblem (),
	     ginit = Constraint.Call.mkInit(),
	     ginfo = Constraint.Call.mkInfo(),
	     sinit = Constraint.Call.mkInit(),
	     sinfo = Constraint.Call.mkInfo()}
	     *)

	    local
		exception screwup
	    in
	    fun addConstraint (str) = 
		let
		    val (assert, m1) = readFormula (!activeModspecs) str
		    val (presup, m2) = readFormula (!activeModspecs) m1
		    val (call, m3) = 
			Constraint.Call.readCall
			(Constraint.Call.newPPVar ()) m2
		    val (cf, m4) = 
			readFormula (!activeModspecs) m3
			
		    val CI{constraints, ginit, ginfo, sinit, sinfo} = 
			(case !activeConstraints of
			     NONE => raise screwup
			   | SOME x => x)
		in
		    print "in (main) addConstraint\n";
		    (case (assert, presup, call, cf, m4) of
			 (SOME assert, SOME presup, SOME call, SOME cf, nil) =>
			     let
				 val (privateworld, privateset) =
				     setof (fn _ => nil, (), assert,
					    !activeDatabase)
				 val public =
				     setof (fn _ => nil, (), presup,
					    !activeDatabase)

				 val (gcopycall, newginit) = 
				     Constraint.Call.newcopyCall(call, ginit)
				 val (scopycall, newsinit) =
				     Constraint.Call.copycallAfter
				     (call, sinit, newginit)
				 val tinit = 
				     Constraint.Call.getInit
				     (sinfo, scopycall, privateworld)

				 val (newsinfo) =
				     case (Constraint.Call.setWithInit
					   (sinfo,
					    newsinit,
					    makeNewTrans
					    (hd (privateset (tinit, cf)),
					     tinit, privateworld),
					    NONE)) of
					 NONE => raise screwup
				       | SOME thing => thing

				 val newproblem =
				     Constraint.addConstraint 
				     (constraints, cf, 
				      Constraint.Call.getInit
				      (ginfo, gcopycall, #1 public), ginfo,
				      Constraint.Call.getInit
				      (newsinfo, scopycall, #1 public), 
				      newsinfo)
			     in
				 print "Added: ";
				 Constraint.Call.prettyprintCall newsinfo call;
				 print "\n";
				 Constraint.printProblem(newproblem);
				 
				 activeConstraints :=
				 SOME(CI{constraints = newproblem,
					 ginit = newginit,
					 ginfo = ginfo,
					 sinit = newsinit,
					 sinfo = newsinfo});
				 true
			     end
		       | _ => 
			     (print "addconstraint: failed to parse args\n";
			      true))
		end
	    handle screwup => 
		(print "Screwup on constraints\n";
		 true)
		 | Hd => (print "No solutions for constraint\n";
			  true)
	    end

	    fun resetDatabase () =
		(activeDatabase := makeDatabase(!activeModspecs, nil);
		 activeModspecsStale := false;
		 activeQabbrevs := nil;
		 activeSchemas := nil;
		 activeWords := Word.newLexicon ();
		 activeTrees := Itag.newTaglex ())

	    fun resetModspecs [] =
		(activeModspecs := newCollection ();
		 resetDatabase();
		 true)
	      | resetModspecs _ =
		(print "Error: no arguments expected.\n";
		 true)

	    fun readModspecs nil = true
	      | readModspecs (files) = 
		let
		    val new = newCollection ()
		fun reader (str, ()) =
		    (case readModality (new, str) of
			 (SOME m, _) =>
			     (addModality (new, m);
			      ((), true))
		       | _ => ((), false))
		in
		    (case Main.readLoopLoop (files, reader, ()) of
			 NONE => true
		       | SOME () => (activeModspecs := new;
				     compile(new);
				     resetDatabase();
				     true))
		end

	    fun readQabbrevs (files) =
		let
		    val rmf = Matchbank.readMatchelt (readFormula (!activeModspecs))
		    fun reader (nil, db) = (db, false)
		      | reader (str, db) =
			(case rmf str of 
			     (SOME e, _) =>
				 (e :: db, true)
			   | _ => (db, false))
		in
		    (case Main.readLoopLoop(files, reader, nil) of
			 NONE => true
		       | SOME db => (activeQabbrevs := db;
				     true))
		end

	    fun readSchemas (files) =
		let
		    val rf = readFormula (!activeModspecs)

		    fun r2f s =
			(case rf s of
			     (SOME f1, ":" :: more) =>
				 (case rf more of
				      (SOME f2, endstr) => 
					  (SOME (f1,f2), endstr)
				    | _ => (NONE, s))
			   | _ => (NONE, s))
			     
		    val rmf = Matchbank.readMatchelt r2f

		    fun reader (nil, db) = (db, false)
		      | reader (str, db) =
			(case rmf str of 
			     (SOME e, _) =>
				 (e :: db, true)
			   | _ => (db, false))
		in
		    (case Main.readLoopLoop(files, reader, nil) of
			 NONE => true
		       | SOME db => (activeSchemas := db;
				     true))
		end

	    fun readDatabase (files) =
		let
		    val rf = readFormula (!activeModspecs)

		    fun reader (nil, db) = (db, false)
		      | reader (str :: "::" :: rest, db) =
			(case rf (rest) of
			     (SOME f, _) =>
				 (addFormulas(db, [(str,f)]), true)
			   | _ => (db, false))
		      | reader (str, db) =
			(case rf (str) of
			     (SOME f, _) =>
				 (addFormulas(db, [("anonymous",f)]), true)
			   | _ => (db, false))
		in
		    (case Main.readLoopLoop (files, reader, 
					     (if !activeModspecsStale then 
						  makeDatabase(!activeModspecs, nil)
					      else !activeDatabase)) of
			 NONE => true
		       | SOME db => (activeDatabase := db;
				     activeModspecsStale := false;
				     true))
		end 

	    fun readWords (files) =
		let
		    val rw = Word.readWord (!activeModspecs)

		    fun reader (nil, db) = (db, false)
		      | reader (str, db) = 
			case rw str of
			    (SOME e, _) => (Word.addWord(db, e), true)
			  | _ => (db, false)
		in
		    (case Main.readLoopLoop(files, reader, Word.newLexicon()) of
			 NONE => true
		       | SOME db => (activeWords := db;
				     true))
		end

	    fun readTrees (files) =
		let
		    val re = Itag.readEntry (!activeModspecs)

		    fun reader (nil, db) = (db, false)
		      | reader (str, db) = 
			case re str of
			    (SOME e, _) => (Itag.addEntry(db, e), true)
			  | _ => (db, false)
		in
		    (case Main.readLoopLoop(files, reader, Itag.newTaglex()) of
			 NONE => true
		       | SOME db => (activeTrees := db;
				     true))
		end

	    fun matchWords (m0) =
		let
		    val (assert, m1) = readFormula (!activeModspecs) m0
		    val (presup, m2) = readFormula (!activeModspecs) m1
		    val (call, m3) = Word.Call.readCall (Word.Call.newPPVar ()) m2
		    val (sf, m4) = Word.Features.readFeatures m3
		in
		    print "in (main) matchWords\n";
		    case (assert, presup, call, sf, m4) of
			(SOME assert, SOME presup, SOME call, SOME sf, nil) =>
			    let
				val private = 
				    setof (fn _ => nil, (), assert, 
						   !activeDatabase)
				val public = 
				    setof (fn _ => nil, (), presup,
						   !activeDatabase)

				val info = Word.Call.mkInfo ()

				val choices = 
				    Word.Tag.choices(Word.Tag.start(sf, call))

				val matches =
				    Word.matchAll (public, private)
				    (hd choices, info) (!activeWords)

				val calls = 
				    Word.getCalls matches
			    in
				print "Matches:\n";
				app Word.printMatch matches;
				print "\nCalls:\n";
				app (fn (x,y) => Word.Call.prettyprintCall y x) calls;
				true
			    end
		       | _ => 
			     (print "match, failed to parse arguments.\n";
			      true)
		end

	    fun matchTrees (m0) =
		let
		    val (presup, m2) = readFormula (!activeModspecs) m0
		    val (call, m4) = Word.Call.readCall (Word.Call.newPPVar ()) m2
		in
		    print "in (main) matchTrees\n";
		    case (presup, call, m4) of
			(SOME presup, SOME call, nil) =>
			    let
				val public = 
				    setof (fn _ => nil, (), presup,
						   !activeDatabase)

				val info = Itag.Call.mkInfo ()
			    in
				case (Itag.matchAll public 
				      [(call, info)] (!activeTrees)) of
				    [m] => 
					(Itag.printMatch m;
					 true)
				  | _ => 
					(print "No match.\n";
					 true)
			    end
		      | _ => 
			    (print "match, failed to parse arguments.\n";
			     true)
		end

	    local
		val currentTree = ref NONE : Derive.deriveState option ref
	    in
		fun resetTree (m0) =
		    let
			val (f, m1) = Itag.Tag.Features.readFeatures m0
			val (c1o, m2) = Itag.Call.readCall (Itag.Call.newPPVar ()) m1
			val (c2o, m3) = Itag.Call.readCall (Itag.Call.newPPVar ()) m2
			val (cg, m4) = FieldStr.readListOf (readAtom) m3
		    in
			case (f, c1o, c2o, cg, m4) of
			    (SOME features, SOME c1, SOME c2, SOME cg, nil) =>
				(currentTree := NONE;
				 (* SOME (Derive.start(features, c1, c2, cg)); *)
				 true)
			  | _ => 
				(print "reset tree: failed to parse arguments.\n";
				 true)
		    end

		fun extend m0 =
		    let
			val (assert, m1) = readFormula (!activeModspecs) m0
			val (presup, m2) = readFormula (!activeModspecs) m1
		    in
			(case (!currentTree, assert, presup, m2) of
			    (SOME tree, SOME assert, SOME presup, nil) => 
				let
				    val private =
					setof (fn _ => nil, (), assert,
					       !activeDatabase)
				    val public =
					setof (fn _ => nil, (), presup,
					       !activeDatabase)
				in
				    Main.selectList (Derive.extensions 
						     (!activeWords, !activeTrees,
						      public, presup, private) tree,
						     Derive.printState,
						     fn _ => (),
						     fn x => (currentTree := SOME x;
							      true))  nil
				end
			  | _ =>
				(print "Extend: bad situation\n";
				 true))
		    end

		fun printState _ =
		    (case !currentTree of
			 NONE => (print "PrintState: no state\n";
				  true)
		       | SOME tree => 
			     (Derive.printState tree;
			      true))
	    end

	    datatype last =
		Nothing
	      | Query of unit more
	      | AQs of (unit more * formula Matchbank.match list)
	      | Plan of (int * Term.Patheq.Modspec.absmod * Term.Patheq.Modspec.absmod * 
			 Assume.assumption list more)

	    val alternatives = ref Nothing : last ref

	    local 
		val queryt  = queryFormulaTr (fn _ => nil, ()) 
	    in
		fun checkAbbrevs nil = Nothing
		  | checkAbbrevs (Matchbank.M{trans,data} :: matches) = 
		    case queryt (trans, data, !activeDatabase) of
			NONE => checkAbbrevs matches
		      | SOME (pf, more) =>
			    (displayProof (fn _ => (), output) pf;
			     AQs (more, matches))
	    end

	    fun qabbrev x =
		let
		    val ways = (List.mapPartial (Matchbank.match x) (!activeQabbrevs))
		in
		    alternatives := Nothing;
		    alternatives := checkAbbrevs ways;
		    case (!alternatives) of
			Nothing => false 
		      | _ => true
		end

	    fun query x =
		(print "Query is: ";
		 printFormula (print) x;
		 print ":\n";
		 alternatives := Nothing;
		 case (queryFormula (fn _ => nil, ()) (x, !activeDatabase)) of
		     NONE => false
		   | SOME (pf, more) => 
			 (displayProof (fn _ => (), output) pf;
			  alternatives := Query more;
			  true))

	    fun queryall dump x =
		let
		    fun n NONE = ()
		      | n (SOME (pf, more)) = 
			(if dump then
			     (displayProof (fn _ => (), output) pf;
			      print "----\n")
			 else
			     (print ".";
			      TextIO.flushOut TextIO.stdOut);
			     
			 n (requeryFormula (fn _ => nil, ()) (more, !activeDatabase)))
		in
		    print "Query is: ";
		    printFormula (print) x;
		    print ":\n";
		    alternatives := Nothing;
		    n (queryFormula (fn _ => nil, ()) (x, !activeDatabase));
		    print "\n"
		end

	    fun queryallbut (x, y) =
		let
		    fun check pf =
			(case followupMetavar (fn _ => nil, ()) (pf, y, !activeDatabase) of
			     SOME (pf, more) => ()
			   | NONE => (displayProof (fn _ => (), output) pf;
				      print "----\n"))

		    fun n NONE = ()
		      | n (SOME (pf, more)) = 
			(check pf;
			 n (requeryFormula (fn _ => nil, ()) (more, !activeDatabase)))
		in
		    print "Finding all: ";
		    printFormula (print) x;
		    print "\nwithout: ";
		    printFormula (print) y;
		    print "\n";
		    alternatives := Nothing;
		    n (queryFormula (fn _ => nil, ()) (x, !activeDatabase))
		end

	    fun queryallbutm (Matchbank.M{trans, data=(x, y)}) =
		let
		    fun check pf =
			(case followupMetavarTr (fn _ => nil, ()) 
			     (trans, pf, y, !activeDatabase) of
			     SOME (pf, more) => ()
			   | NONE => (displayProof (fn _ => (), output) pf;
				      print "----\n"))

		    fun n NONE = ()
		      | n (SOME (pf, more)) = 
			(check pf;
			 n (requeryFormula (fn _ => nil, ()) (more, !activeDatabase)))
		in
		    print "Finding all: ";
		    printFormula (print) x;
		    print "\nwithout: ";
		    printFormula (print) y;
		    print "\n";
		    alternatives := Nothing;
		    n (queryFormulaTr (fn _ => nil, ()) (trans, x, !activeDatabase))
		end

	    fun qschema x =
		(case List.mapPartial (Matchbank.match x) (!activeSchemas) of
		     nil => print "No matching schema\n"
		   | (h :: _) => queryallbutm h)

	    fun plan (n, km, tm) x =
		(print "Query is: ";
		 printFormula (print) x;
		 print ":\n";
		 alternatives := Nothing;
		 case (queryFormula (Assume.matchAssumptions(n,km,tm), nil) 
		       (x, !activeDatabase)) of
		     NONE => false
		   | SOME (pf, more) => 
			 (displayProof (Assume.prettyprintAssumptions output, output) pf;
			  alternatives := Plan(n,km,tm,more);
			  true))

		local
		    fun process Nothing =
			 (print "No search currently in progress\n"; Nothing)
		      | process (Query more) =
			 (case requeryFormula (fn _ => nil, ()) (more, !activeDatabase) of
			      NONE => Nothing
			    | SOME (pf, more) => 
				  (displayProof (fn _ => (), output) pf;
				   Query more))
		      | process (AQs (more, try)) =
			(case process (Query more) of
			     Query m' => AQs (m', try) 
			   | _ => checkAbbrevs try)
		      | process (Plan (n, km, tm, more)) =
			(case requeryFormula (Assume.matchAssumptions(n,km,tm), nil) 
			     (more, !activeDatabase) of
			     NONE => Nothing
			   | SOME (pf, more) => 
				 (displayProof (Assume.prettyprintAssumptions output, output) pf;
				  Plan (n,km,tm,more)))
		in
		    fun requery () =
			(alternatives := process (!alternatives);
			 case !alternatives of
			     Nothing => false 
			   | _ => true)
		end

	    fun stats () =
		(if Debug.debugp "statistics" then
		     Formulas.Patheq.Ctree.printStats(print)
		 else ();
		 Formulas.Patheq.Ctree.resetStats())

	end (* local *)
		
	fun debug ([string]) = (Debug.debug string; true)
	  | debug _ = (print "Debug: bad argument\n"; true)

	fun nodebug ([string]) = (Debug.nodebug string; true)
	  | nodebug _ = (print "Nodebug: bad argument\n"; true)

	fun planFn (knowledgestr :: nextstr :: timestr :: depth :: query) =
	    (let
		 val knowledgem = Formulas.Patheq.Modspec.lookup
		     (!activeModspecs,knowledgestr)
		 val timem = Formulas.Patheq.Modspec.lookup
		     (!activeModspecs,timestr)
		 fun bq (0, q) = q
		   | bq (n, q) = bq (n-1, knowledgestr :: nextstr :: q)
	     in
		 case (knowledgem, timem, Int.fromString depth) of
		     (SOME km, SOME tm, SOME ndepth) =>
			 Main.ftest1(fn x => (Formulas.readFormula (!activeModspecs) x),
				     plan (ndepth, km, tm),
				     fn x => (if x then () else output "no\n"; stats()))
			 (bq (ndepth, query))
				      | _ => (print ("Could not identify knowledge (" ^
						     knowledgestr ^
						     ") or time(" ^
						     timestr ^ ") or depth(" ^
						     depth ^ ")\n");
					      true)
	     end)
	  | planFn _  = 
	    (print ("Plan: too few arguments.\n");
	     true)

	(*
	 * Actions
	 *)
     
	val actions = 
	    [Main.A{keyword = "quit",
		    description = "Exit interpreter.\n",
		    operation = fn _ => false,
		    call = "quit"},

	     Main.A{keyword = "query",
		    description = "Do proof search on <formula>.\n",
		    operation = Main.ftest1(fn x => (Formulas.readFormula 
						     (!activeModspecs) x),
					    query,
					    fn x => (if x then () else output "No.\n"; stats())),
		    call = "query <formula>"},

	     Main.A{keyword = "aq",
		    description = "Prove query matching <term>.\n",
		    operation = Main.ftest1(Formulas.Term.readTerm,
					    qabbrev,
					    fn x => (if x then () else output "No.\n";
							 stats ())),
		    call = "aq <term>"},

	     Main.A{keyword = "sm",
		    description = "Describe (schema) <term>.\n",
		    operation = Main.ftest1(Formulas.Term.readTerm,
					    qschema,
					    fn _ => (output "Done.\n";
						     stats ())),
		    call = "sm <term>"},

	     Main.A{keyword = "qall",
		    description = "Print all guys of on <formula>.\n",
		    operation = Main.ftest1(fn x => (Formulas.readFormula 
						     (!activeModspecs) x),
					    queryall true,
					    fn _ => (output "Done.\n";
						     stats())),
		    call = "qall <formula>"},

	     Main.A{keyword = "run",
		    description = "Exhaust proof search on <formula>.\n",
		    operation = Main.ftest1(fn x => (Formulas.readFormula 
						     (!activeModspecs) x),
					    queryall false,
					    fn _ => (output "Done.\n";
						     stats())),
		    call = "run <formula>"},

	     Main.A{keyword = "allbut",
		    description = ("Exhaust proof search on <formula>\n" ^
				   "Exhibiting $-instances that don't match <formula>\n"),
		    operation = Main.ftest2(fn x => (Formulas.readFormula 
						     (!activeModspecs) x),
					    fn x => (Formulas.readFormula 
						     (!activeModspecs) x),
					    queryallbut,
					    fn _ => (output "Done.\n";
						     stats())),
		    call = "qall <formula>"},

	     Main.A{keyword = "next",
		    description = "Compute the next solution to the last query.\n",
		    operation = Main.ftest0(requery,
					    fn x => (if x then () else output "No.\n"; stats())),
		    call = "next"},

	     Main.A{keyword = "plan",
		    description = "Plan with <knowledge> <next> <later> <steps> <goal>.\n",
		    operation = planFn,
		    call = "plan <knowledge> <next> <later> <steps> <goal>"},

	     Main.A{keyword = "show",
		    description = "Read and print <formula>.\n",
		    operation = Main.ftest1(fn x => (Formulas.readFormula
						     (!activeModspecs) x),
					    Formulas.printFormula (output),
					    fn _ => ()),
		    call = "show <formula>"},

	     Main.S{keyword = "match",
		    description = "Match parts of the lexicon.\n",
		    subactions =
		    [Main.A{keyword = "words",
			    description = 
			    ("Find words describing <call> with <cat> \n" ^
			     "at <assert> and <presup>.\n"),
			    call = "match words <assert> <presup> <call> <cat>",
			    operation = matchWords},
		    
		    Main.A{keyword = "trees",
			    description = 
			    ("Find trees describing <call> with <cat> \n" ^
			     "at <presup>.\n"),
			    call = "match trees <presup> <call> <cat>",
			    operation = matchTrees}
		    ]},

	     Main.A{keyword = "constrain",
		    description = ("Act as if word with presup <cf> and call <call>\n"
				   ^ "is interpreted at <assert> and <presup>.\n"),
		    call = "constrain <assert> <presup> <call> <cf>",
		    operation = addConstraint},

	     Main.A{keyword = "more",
		    description = "Add another word.\n",
		    call = "more <private> <public>",
		    operation = extend},

	     Main.S{keyword = "reset",
		    description = "Reset various different data structures.\n",
		    subactions =
		    [Main.A{keyword = "modalities",
			    description = ("Reset inventory of necessity operators.\n" ^
					   "Arg <n> gives number of families of boxes.\n"),
			    operation = resetModspecs,
			    call = "reset modalities"},

		     Main.A{keyword = "db",
			    description = "Reset inventory of true facts.\n",
			    operation = fn _ => (resetDatabase (); true),
			    call = "reset db"},

		     Main.A{keyword = "derv",
			    description = "Reset derivation in progress.\n",
			    operation = resetTree,
			    call = "reset db <features> <speccall> <gencall> <goals>"}
		     ]},

	     Main.S{keyword = "debug",
		    description = "Set the database to notice certain things.\n",
		    subactions =
		    [ Main.A{keyword = "flag",
			     description = "Set debug flag <flag>.\n",
			     call = "debug flag <flag>",
			     operation = debug}
		     ]},

	     Main.S{keyword = "nodebug",
		    description = "Set the database not to notice certain things.\n",
		    subactions =
		    [ Main.A{keyword = "flag",
			     description = "Turn off debug flag <flag>.\n",
			     call = "nodebug flag <flag>",
			     operation = nodebug}
		     ]},

	     Main.S{keyword = "load",
		    description = "Load datafiles of various sorts.\n",
		    subactions =
		    [Main.A{keyword = "modalities",
			    description = ("Use files in <filelist> for specs of boxes.\n"),
			    call = "load modalities <filelist>",
			    operation = readModspecs},

		     Main.A{keyword = "facts",
			    description = "Use files in <filelist> for true facts.\n",
			    call = "load facts <filelist>",
			    operation = readDatabase},

		     Main.A{keyword = "schemas",
			    description = "Use files in <filelist> for talk schemas.\n",
			    call = "load schemas <filelist>",
			    operation = readSchemas},

		     Main.A{keyword = "abbrevs",
			    description = "Use files in <filelist> for query abbrevs.\n",
			    call = "load abbrevs <filelist>",
			    operation = readQabbrevs},

		     Main.A{keyword = "words",
			    description = "Use files in <filelist> for lexicon.\n",
			    call = "load words <filelist>",
			    operation = readWords},
			   
		     Main.A{keyword = "trees",
			    description = "Use files in <filelist> for tag lexicon.\n",
			    call = "load trees <filelist>",
			    operation = readTrees}
			   
		     ]}
	     ]

	val _ = 
	    app Debug.register
	    [("statistics",
	      "Display statistics on tree constraint algorithm\n",
	      false)
	     ];
	    

    end (* struct *)


