(*
 * Matthew Stone
 * Dialup 0.1 - 3/96
 * main-sig.sml
 *
 * Signature file for quickly specifying high-level interpreters.
 *)

signature MAIN =
    sig
	(*
	 * We use an abstraction of IO so that this 
	 * module will work with either files or
	 * sockets.
	 *)
	structure Read : IOA

	(*
	 * You specify the interpreter as a set of commands,
	 * or actions.  
	 * 'A' actions take the form <keyword> <args>
	 *   the ARGS are passed as a string list to the
	 *   OPERATION function, which does whatever 
	 *   computation is appropriate and then returns
	 *   TRUE unless the user has decided to quit
	 *   the program.  CALL and DESCRIPTION are used
	 *   to print out help information on the command.
	 * 'S' actions handle all the cases for <keyword>
	 *   The rest of the string is recursively 
	 *   processed using the definitions given in
	 *   the SUBACTIONS list.  Again DESCRIPTION is
	 *   used to print out help information.
	 *)
	datatype action = 
	    A of {keyword: string,
		  operation: string list -> bool,
		  call: string,
		  description: string}
	  | S of {keyword: string,
		  subactions: action list,
		  description: string}

	(*
	 * function that prints out help information
	 * for the actions in <ACTION LIST>, according
	 * to the queried command <STRING LIST>
	 *)
	val help: action list -> string list -> bool

	(*
	 * executes the appropriate command from
	 * <ACTION LIST> with appropriate parameters,
	 * when the user gives <STRING LIST>
	 *)
	val process: action list -> string list -> bool

	(* Polymorphic functions for frequently performed tasks *)

	(*
	 * readerLoop: given a Read.reader
	 *      an update function that incorporates a string list
	 *         into its argument (returning also indication of
	 *         success), and
	 *      an initial value
	 * incorporate all of the clauses in the named file into
	 * the object---does not catch exceptions;
	 * returns also the number of additions processed.
	 *)
	val readerLoop: (Read.reader * 
		       (string list * 'a -> 'a * bool) * 
		       'a -> 'a * int )

	(*
	 * do reader Loop but from the file named STRING
	 *) 
	val readLoop: (string * 
		       (string list * 'a -> 'a * bool) * 
		       'a -> 'a option )

	(*
	 * readLoopLoop: like readLoop but for a list of files.
	 *)
	val readLoopLoop: (string list *
			   (string list * 'a -> 'a * bool) * 
			   'a -> 'a option)

	(*
	 * Polymorphic functions to test the output of 
	 *    operations.
	 * ftest1: read an alpha, compute beta from alpha,
	 *         print beta.
	 * ftesto1: read an alpha, compute beta option from alpha,
	 *          if NONE, print STRING, otherwise print beta.
	 * ftest2: read an alpha, then a beta.
	 *         compute gamma from them, and print it out.
	 * selectList: given an alpha list, 
	 *         a display function,
	 *         a kill function,
	 *         a function to call on selected
	 *     show the elements of the list and get the user
	 *     to choose one.  call the selection function on
	 *     that, call the kill function on the others.
	 * fselect2: read in an alpha and a beta, then
	 *     generate a list of gammas from the inputs.
	 *     Select one of the gammas.
	 *)
	val ftest0: (unit -> 'b) * ('b -> unit) -> string list -> bool

	val ftest1: ((string list -> 'a option * string list) *
		     ('a -> 'b) *
		     ('b -> unit)) -> string list -> bool
	val ftesto1: ((string list -> 'a option * string list) *
		      ('a -> 'b option) *
		      string *
		      ('b -> unit)) -> string list -> bool
	val ftest2: ((string list -> 'a option * string list) *
		     (string list -> 'b option * string list) *
		     ('a * 'b -> 'c) *
		     ('c -> unit)) -> string list -> bool
	val selectList: ('a list *
			 ('a -> unit) *
			 ('a -> unit) *
			 ('a -> bool)) -> string list -> bool
	val fselect2: ((string list -> 'a option * string list) *
		       (string list -> 'b option * string list) *
		       ('a * 'b -> 'c list) *
		       ('c -> unit) *
		       ('c -> unit) *
		       ('c -> bool)) -> string list -> bool

	(*
	 * an intepreter loop over ACTION list 
	 * with an initialization 
	 * function, suitable for curried application and
	 * being exportFn'd.
	 *)
	val main: (action list * ('a -> unit) * (unit -> OS.Process.status) *
		   (unit -> string) * (string -> unit)) -> 
	    'a -> OS.Process.status
    end
