(*
 * Matthew Stone
 * Dialup 0.1 - 3/96
 *
 * Implementation of main signature
 * (utilities for user-level command interpreter).
 *)

functor Main(structure Lex: LEX;
	     structure Read: IOA;
	     structure Clock: CLOCK): MAIN =
    struct

	structure Read = Read

	(*
	 * You specify the interpreter as a set of commands,
	 * or actions.  
	 * 'A' actions take the form <keyword> <args>
	 *   the ARGS are passed as a string list to the
	 *   OPERATION function, which does whatever 
	 *   computation is appropriate and then returns
	 *   TRUE unless the user has decided to quit
	 *   the program.  CALL and DESCRIPTION are used
	 *   to print out help information on the command.
	 * 'S' actions handle all the cases for <keyword>
	 *   The rest of the string is recursively 
	 *   processed using the definitions given in
	 *   the SUBACTIONS list.  Again DESCRIPTION is
	 *   used to print out help information.
	 *)
	datatype action = 
	    A of {keyword: string,
		  operation: string list -> bool,
		  call: string,
		  description: string}
	  | S of {keyword: string,
		  subactions: action list,
		  description: string}

	local 
	    (*
	     * select: string list * action list -> 
	     *         (action*args) option
	     * find the function from action list that
	     * matches the call string list
	     *)
	    fun select (nil,_) =
		(print "Incomplete command specified.\n";
		 NONE)
	      | select (head::_, nil) =
		(print "No such command: ";
		 print head;
		 print ".\n";
		 NONE)
	      | select (head::rest, 
			(action as A{keyword,...}) :: more) =
		if head = keyword then
		    SOME (action, rest)
		else
		    select (head::rest, more)
	      | select ([head], 
			(action as S{keyword,subactions,...}) :: more) =
		if head = keyword then
		    SOME (action, nil)
		else
		    select (head::nil, more) 
	      | select (head::rest, 
			(action as S{keyword,subactions,...}) :: more) =
		if head = keyword then
		    select (rest, subactions)
		else
		    select (head::rest, more)

	    (* pk: print keyword of aciton *)
	    fun pk (A{keyword,...}) =
		(print keyword;
		 print " ")
	      | pk (S{keyword,...}) =
		(print keyword;
		 print " ")

	    (* describe: print help on an action *)
	    fun describe (A{keyword, call, description,...}) =
		(print "Usage: ";
		 print call;
		 print "\nFor: ";
		 print description;
		 true)
	      | describe (S{keyword, subactions, description,...}) =
		(print keyword;
		 print " family of operations: \n   ";
		 app pk subactions;
		 print "\nFor: ";
		 print description;
		 true)
	in
	    (* help: print out help about action in actions
	     * matching string list *)
	    fun help actions nil = (app (ignore o describe) actions; true)
	      | help actions ["."] = (app (ignore o describe) actions; true)
	      | help actions (str) =
		(case select (str, actions) of
		     NONE => true
		   | SOME (a, _) => describe a)

	    (* process: execute the action in actiosn designated 
	     * the passed string list *)
	    fun process _ nil = true
	      | process actions ("help" :: str) = 
		help actions str
	      | process actions (str) =
		(case select (str, actions) of
		     NONE => true
		   | SOME (A{operation,...}, arg) => (operation arg)
		   | SOME (S_, arg) => (print "Incomplete command\n";
					true))
	end (* local *)
			 
	(*
	 * Different reading operations share the following basic loop.
	 *)

	fun readerLoop (isr, readf, init) =
	    let
		fun next (old, ct) =
		    case Read.getclause isr of
			"" => (old, ct) 
		      | s => 
			    let val (new, success) = 
				readf (Lex.segment s, old)
			    in next (new, ct + (if success then 1 else 0))
			    end
	    in
		next (init, 0)
	    end

	fun readLoop (file, reader, init) =
	    (let
		 val is = (TextIO.openIn file)
		 val isr = (Read.wrapInstream is)
 
		 val (result, ct) = readerLoop(isr, reader, init)
	     in
		 TextIO.closeIn (is);
		 print (Int.toString ct);
		 print " objects read\n";
		 SOME(result)
	     end (* let *)
	 handle IO.Io {name,...} => 
	     (print "Error: ";
	      print name;
	      print "\n";
	      NONE))

	fun readLoopLoop (nil, reader, sofar) = SOME sofar
	  | readLoopLoop ("." :: more, reader, sofar) =
	    readconcat("." :: more, reader, sofar)
	  | readLoopLoop (file :: "." :: more, reader, sofar) =
	    readconcat ((file ^ ".") :: more, reader, sofar)
	  | readLoopLoop (file::more, reader, sofar) =
	    (case readLoop(file, reader, sofar) of
		 NONE => NONE
	       | SOME next => 
		     readLoopLoop (more, reader, next))

	and readconcat (file :: "." :: more, reader, sofar) =
	    readconcat ((file ^ ".") :: more, reader, sofar)
	  | readconcat (file :: "/" :: more, reader, sofar) =
	    readconcat ((file ^ "/") :: more, reader, sofar)
	  | readconcat (file :: stuff :: more, reader, sofar) =
	    readLoopLoop((file ^ stuff) :: more, reader, sofar)
	  | readconcat (rest, reader, sofar) =
	    readLoopLoop(rest, reader, sofar)

	(*
	 * Different function-testing operations share the following
	 * basic loops
	 *)

	fun ftest1 (input, f, output) (str) =
	    let
		val (argo, rest) = input (str)
	    in
		(case (argo, rest) of
		     (SOME arg, nil) => (output (f(arg));
					 true)
		   | _ => (print "Reading error\n";
			   true))
	    end

	fun ftest0 (f, output) = ftest1 (fn x => (SOME(),x), f, output) 

	fun ftesto1 (input, f, pstr:string, output) (str) =
	    let
		val (argo, rest) = input (str)
	    in
		(case (argo, rest) of
		     (SOME arg, nil) => 
			 (case f(arg) of 
			      NONE => (print pstr;
				       print " error\n";
				       true)
			    | SOME res => 
				       (output res;
					true))
		   | _ => (print "Reading error\n";
			   true))
	    end

	fun ftest2 (input1, input2, f, output) (str) =
	    let
		val (arg1o, mid) = input1 str
		val (arg2o, rest) = input2 mid
	    in
		(case (arg1o, arg2o, rest) of
		     (SOME arg1, SOME arg2, nil) =>
			 (output(f(arg1,arg2));
			  true)
		   | _ => (print "Reading error\n";
			   true))
	    end

	local
	    fun output show (n:int, nil) = ()
	      | output show (n, h::t) = 
		(print "Option ";
		 print (Int.toString n);
		 print ":\n";
		 show h;
		 output show (n+1, t))
	    
	    fun select (max, list, free) =
		(print "Enter your choice: ";
		 TextIO.flushOut TextIO.stdOut;
		 (case Int.fromString (TextIO.inputLine TextIO.stdIn) of
		      NONE =>
			  (print "Could not read value.\n";
			   print "Enter value between 0 and ";
			   print (Int.toString max);
			   print "\n";
			   select (max,list,free))
		    | SOME ans => 
			  let
			      fun ith (nil, _) = raise Subscript
				| ith (t::r, 0) = 
				  (app free r;
				   t)
				| ith (t::r, n) =
				  (free t;
				   ith (r, n-1))
			  in
			      if ans >= 0 andalso ans < max then
				  ith(list,ans)
			      else
				  (print "No such value.\nEnter value between 0 and ";
				   print (Int.toString max);
				   print "\n";
				   select (max,list,free))
			  end))

	in
	    fun selectList(list, show, free, final) (str) =
		(output show (0,list);
		 case list of 
		     nil => (print "No options.\n";
			     true)
		   | _ => final(select(length list, list, free)))

	    fun fselect2(input1, input2, f, show, free, final) (str) =
		let
		    val (arg1o, mid) = input1 str
		    val (arg2o, rest) = input2 mid
		in
		    case (arg1o, arg2o, rest) of
			(SOME arg1, SOME arg2, nil) =>
			    let
				val list = f(arg1,arg2)
			    in
				output show (0, list);
				case list of 
				    nil => (print "No options.\n"; true)
				  | _ => final(select(length list, list, free))
			    end
		      | _ => (print "Reading error\n";
			      true)
		end (* let *)
	end (* local *)

	fun main (actions, init, finish, input, output) arg = 
	    let
		fun continue () = 
		    (case input () of 
			 "" => false
		       | s => 
			     (process (A{keyword = "time",
					 description = "Time <command>\n",
					 call = "time <command>",
					 operation = Clock.clockapp(process actions) 
					 "operation performed"} :: actions)
			      (Lex.segment s)))

		fun loop () =
		    (output "\n";
		     if Clock.interruptible(continue,true) then
			 loop()
		     else
			 finish())
	    in
		init arg;
		loop()
	    end  (* let *)
    end (* struct *)

